@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1997-2016, University of Amsterdam
+    Copyright (c)  1997-2020, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -43,6 +44,7 @@
     prolog:message//1,              % entire message
     prolog:error_message//1,        % 1-st argument of error term
     prolog:message_context//1,      % Context of error messages
+    prolog:deprecated//1,	    % Deprecated features
     prolog:message_location//1,     % (File) location of error messages
     prolog:message_line_element/2.  % Extend printing
 :- discontiguous
@@ -51,6 +53,8 @@
 :- public
     translate_message//1.
 
+:- create_prolog_flag(message_context, [thread], []).
+
 %!  translate_message(+Term)// is det.
 %
 %   Translate a message Term into message lines. The produced lines
@@ -85,10 +89,15 @@ translate_message2(Term) -->
     prolog:message(Term).
 translate_message2(Term) -->
     prolog_message(Term).
-translate_message2(error(resource_error(stack), Name)) -->
-    [ 'Out of ~w stack'-[Name] ].
-translate_message2(error(resource_error(Missing), _)) -->
-    [ 'Not enough resources: ~w'-[Missing] ].
+translate_message2(error(resource_error(stack), Context)) -->
+    !,
+    out_of_stack(Context).
+translate_message2(error(resource_error(tripwire(Wire, Context)), _)) -->
+    !,
+    tripwire_message(Wire, Context).
+translate_message2(error(existence_error(reset, Ball), SWI)) -->
+    swi_location(SWI),
+    tabling_existence_error(Ball, SWI).
 translate_message2(error(ISO, SWI)) -->
     swi_location(SWI),
     term_message(ISO),
@@ -118,6 +127,8 @@ term_message(Term) -->
 term_message(Term) -->
     [ 'Unknown error term: ~p'-[Term] ].
 
+iso_message(resource_error(Missing)) -->
+    [ 'Not enough resources: ~w'-[Missing] ].
 iso_message(type_error(evaluable, Actual)) -->
     { callable(Actual) },
     [ 'Arithmetic: `~p'' is not a function'-[Actual] ].
@@ -141,8 +152,8 @@ iso_message(permission_error(Action, Type, Object)) -->
 iso_message(evaluation_error(Which)) -->
     [ 'Arithmetic: evaluation error: `~p'''-[Which] ].
 iso_message(existence_error(procedure, Proc)) -->
-    [ 'Undefined procedure: ~q'-[Proc] ],
-    undefined_proc_msg(Proc).
+    [ 'Unknown procedure: ~q'-[Proc] ],
+    unknown_proc_msg(Proc).
 iso_message(existence_error(answer_variable, Var)) -->
     [ '$~w was not bound by a previous query'-[Var] ].
 iso_message(existence_error(Type, Object)) -->
@@ -193,29 +204,31 @@ permission_error(output, binary_stream, Stream) -->
     [ 'No permission to write characters to binary stream `~p'''-[Stream] ].
 permission_error(open, source_sink, alias(Alias)) -->
     [ 'No permission to reuse alias "~p": already taken'-[Alias] ].
+permission_error(tnot, non_tabled_procedure, Pred) -->
+    [ 'The argument of tnot/1 is not tabled: ~p'-[Pred] ].
 permission_error(Action, Type, Object) -->
     [ 'No permission to ~w ~w `~p'''-[Action, Type, Object] ].
 
 
-undefined_proc_msg(_:(^)/2) -->
+unknown_proc_msg(_:(^)/2) -->
     !,
-    undefined_proc_msg((^)/2).
-undefined_proc_msg((^)/2) -->
+    unknown_proc_msg((^)/2).
+unknown_proc_msg((^)/2) -->
     !,
     [nl, '  ^/2 can only appear as the 2nd argument of setof/3 and bagof/3'].
-undefined_proc_msg((:-)/2) -->
+unknown_proc_msg((:-)/2) -->
     !,
     [nl, '  Rules must be loaded from a file'],
     faq('ToplevelMode').
-undefined_proc_msg((:-)/1) -->
+unknown_proc_msg((:-)/1) -->
     !,
     [nl, '  Directives must be loaded from a file'],
     faq('ToplevelMode').
-undefined_proc_msg((?-)/1) -->
+unknown_proc_msg((?-)/1) -->
     !,
     [nl, '  ?- is the Prolog prompt'],
     faq('ToplevelMode').
-undefined_proc_msg(Proc) -->
+unknown_proc_msg(Proc) -->
     { dwim_predicates(Proc, Dwims) },
     (   {Dwims \== []}
     ->  [nl, '  However, there are definitions for:', nl],
@@ -224,7 +237,7 @@ undefined_proc_msg(Proc) -->
     ).
 
 faq(Page) -->
-    [nl, '  See FAQ at http://www.swi-prolog.org/FAQ/', Page, '.txt' ].
+    [nl, '  See FAQ at https://www.swi-prolog.org/FAQ/', Page, '.txt' ].
 
 type_error_comment(_Expected, Actual) -->
     { type_of(Actual, Type),
@@ -289,7 +302,7 @@ syntax_error(cannot_start_term) -->
 syntax_error(punct(Punct, End)) -->
     [ 'Unexpected `~w\' before `~w\''-[Punct, End] ].
 syntax_error(undefined_char_escape(C)) -->
-    [ 'Undefined character escape in quoted atom or string: `\\~w\''-[C] ].
+    [ 'Unknown character escape in quoted atom or string: `\\~w\''-[C] ].
 syntax_error(void_not_allowed) -->
     [ 'Empty argument list "()"' ].
 syntax_error(Message) -->
@@ -305,6 +318,25 @@ domain(range(Low,High)) -->
 domain(Domain) -->
     ['`~w\''-[Domain] ].
 
+%!  tabling_existence_error(+Ball, +Context)//
+%
+%   Called on invalid shift/1  calls.  Track   those  that  result  from
+%   tabling errors.
+
+tabling_existence_error(Ball, Context) -->
+    { table_shift_ball(Ball) },
+    [ 'Tabling dependency error' ],
+    swi_extra(Context).
+
+table_shift_ball(dependency(_Head)).
+table_shift_ball(dependency(_Skeleton, _Trie, _Mono)).
+table_shift_ball(call_info(_Skeleton, _Status)).
+table_shift_ball(call_info(_GenSkeleton, _Skeleton, _Status)).
+
+%!  dwim_predicates(+PI, -Dwims)
+%
+%   Find related predicate indicators.
+
 dwim_predicates(Module:Name/_Arity, Dwims) :-
     !,
     findall(Dwim, dwim_predicate(Module:Name, Dwim), Dwims).
@@ -333,6 +365,13 @@ dwim_message([Head|T]) -->
 
 swi_message(io_error(Op, Stream)) -->
     [ 'I/O error in ~w on stream ~p'-[Op, Stream] ].
+swi_message(thread_error(TID, false)) -->
+    [ 'Thread ~p died due to failure:'-[TID] ].
+swi_message(thread_error(TID, exception(Error))) -->
+    [ 'Thread ~p died abnormally:'-[TID], nl ],
+    translate_message(Error).
+swi_message(idg_dependency_error(Shared, Private)) -->
+    [ 'Shared table for ~p may not depend on private ~p'-[Shared, Private] ].
 swi_message(shell(execute, Cmd)) -->
     [ 'Could not execute `~w'''-[Cmd] ].
 swi_message(shell(signal(Sig), Cmd)) -->
@@ -383,6 +422,14 @@ swi_message(conditional_compilation_error(no_if, What)) -->
     [ ':- ~w without :- if'-[What] ].
 swi_message(duplicate_key(Key)) -->
     [ 'Duplicate key: ~p'-[Key] ].
+swi_message(initialization_error(failed, Goal, File:Line)) -->
+    !,
+    [ '~w:~w: ~p: false'-[File, Line, Goal] ].
+swi_message(initialization_error(Error, Goal, File:Line)) -->
+    [ '~w:~w: ~p '-[File, Line, Goal] ],
+    translate_message(Error).
+swi_message(qlf_format_error(File, Message)) -->
+    [ '~w: Invalid QLF file: ~w'-[File, Message] ].
 
 cond_location(File:Line) -->
     { file_base_name(File, Base) },
@@ -413,6 +460,8 @@ swi_location(stream(Stream, Line, LinePos, CharNo)) -->
     ->  swi_location(file(File, Line, LinePos, CharNo))
     ;   [ 'Stream ~w:~d:~d '-[Stream, Line, LinePos] ]
     ).
+swi_location(autoload(File:Line)) -->
+    [ '~w:~w: '-[File, Line] ].
 swi_location(_) -->
     [].
 
@@ -466,9 +515,9 @@ swi_comment(Msg) -->
 
 
 thread_context -->
-    { thread_self(Me), Me \== main },
+    { thread_self(Me), Me \== main, thread_property(Me, id(Id)) },
     !,
-    ['[Thread ~w] '-[Me]].
+    ['[Thread ~w] '-[Id]].
 thread_context -->
     [].
 
@@ -496,6 +545,17 @@ prolog_message(initialization_failure(Goal, _)) -->
 prolog_message(initialization_exception(E)) -->
     [ 'Prolog initialisation failed:', nl ],
     translate_message(E).
+prolog_message(init_goal_syntax(Error, Text)) -->
+    !,
+    [ '-g ~w: '-[Text] ],
+    translate_message(Error).
+prolog_message(init_goal_failed(failed, @(Goal,File:Line))) -->
+    !,
+    [ '~w:~w: ~p: false'-[File, Line, Goal] ].
+prolog_message(init_goal_failed(Error, @(Goal,File:Line))) -->
+    !,
+    [ '~w:~w: ~p '-[File, Line, Goal] ],
+    translate_message(Error).
 prolog_message(init_goal_failed(failed, Text)) -->
     !,
     [ '-g ~w: false'-[Text] ].
@@ -517,9 +577,9 @@ prolog_message(commandline_arg_type(Flag, Arg)) -->
     [ 'Bad argument to commandline option -~w: ~w'-[Flag, Arg] ].
 prolog_message(missing_feature(Name)) -->
     [ 'This version of SWI-Prolog does not support ~w'-[Name] ].
-prolog_message(singletons(List)) -->
+prolog_message(singletons(_Term, List)) -->
     [ 'Singleton variables: ~w'-[List] ].
-prolog_message(multitons(List)) -->
+prolog_message(multitons(_Term, List)) -->
     [ 'Singleton-marked variables appearing more than once: ~w'-[List] ].
 prolog_message(profile_no_cpu_time) -->
     [ 'No CPU-time info.  Check the SWI-Prolog manual for details' ].
@@ -550,8 +610,12 @@ prolog_message(unknown_in_module_user) -->
     [ 'Using a non-error value for unknown in the global module', nl,
       'causes most of the development environment to stop working.', nl,
       'Please use :- dynamic or limit usage of unknown to a module.', nl,
-      'See http://www.swi-prolog.org/howto/database.html'
+      'See https://www.swi-prolog.org/howto/database.html'
     ].
+prolog_message(deprecated(What)) -->
+    deprecated(What).
+prolog_message(untable(PI)) -->
+    [ 'Reconsult: removed tabling for ~p'-[PI] ].
 
 
                  /*******************************
@@ -592,10 +656,12 @@ prolog_message(no_exported_op(Module, Op)) -->
 prolog_message(discontiguous((-)/2,_)) -->
     prolog_message(minus_in_identifier).
 prolog_message(discontiguous(Proc,Current)) -->
-    [ 'Clauses of ~p are not together in the source-file'-[Proc], nl ],
-    current_definition(Proc, '  Earlier definition at '),
-    [ '  Current predicate: ~p'-[Current], nl,
-      '  Use :- discontiguous ~p. to suppress this message'-[Proc]
+    [ 'Clauses of ', ansi(code, '~p', [Proc]),
+      ' are not together in the source-file', nl ],
+    current_definition(Proc, 'Earlier definition at '),
+    [ 'Current predicate: ', ansi(code, '~p', [Current]), nl,
+      'Use ', ansi(code, ':- discontiguous ~p.', [Proc]),
+      ' to suppress this message'
     ].
 prolog_message(decl_no_effect(Goal)) -->
     [ 'Deprecated declaration has no effect: ~p'-[Goal] ].
@@ -619,7 +685,7 @@ prolog_message(load_file(done(Level, File, Action, Module, Time, Clauses))) -->
 prolog_message(dwim_undefined(Goal, Alternatives)) -->
     { goal_to_predicate_indicator(Goal, Pred)
     },
-    [ 'Undefined procedure: ~q'-[Pred], nl,
+    [ 'Unknown procedure: ~q'-[Pred], nl,
       '    However, there are definitions for:', nl
     ],
     dwim_message(Alternatives).
@@ -631,11 +697,22 @@ prolog_message(error(loop_error(Spec), file_search(Used))) -->
     ],
     used_search(Used).
 prolog_message(minus_in_identifier) -->
-    [ 'The "-" character should not be used to seperate words in an', nl,
+    [ 'The "-" character should not be used to separate words in an', nl,
       'identifier.  Check the SWI-Prolog FAQ for details.'
     ].
 prolog_message(qlf(removed_after_error(File))) -->
     [ 'Removed incomplete QLF file ~w'-[File] ].
+prolog_message(qlf(recompile(Spec,_Pl,_Qlf,Reason))) -->
+    [ '~p: recompiling QLF file'-[Spec] ],
+    qlf_recompile_reason(Reason).
+prolog_message(qlf(can_not_recompile(Spec,QlfFile,_Reason))) -->
+    [ '~p: can not recompile "~w" (access denied)'-[Spec, QlfFile], nl,
+      '\tLoading from source'-[]
+    ].
+prolog_message(qlf(system_lib_out_of_date(Spec,QlfFile))) -->
+    [ '~p: can not recompile "~w" (access denied)'-[Spec, QlfFile], nl,
+      '\tLoading QlfFile'-[]
+    ].
 prolog_message(redefine_module(Module, OldFile, File)) -->
     [ 'Module "~q" already loaded from ~w.'-[Module, OldFile], nl,
       'Wipe and reload from ~w? '-[File], flush
@@ -700,21 +777,27 @@ hidden_module(M) :-
     sub_atom(M, 0, _, _, $).
 
 current_definition(Proc, Prefix) -->
-    { pi_head(Proc, Head),
+    { pi_uhead(Proc, Head),
       predicate_property(Head, file(File)),
       predicate_property(Head, line_count(Line))
     },
-    [ '~w'-[Prefix], '~w:~d'-[File,Line], nl ].
+    [ '~w~w:~d'-[Prefix,File,Line], nl ].
 current_definition(_, _) --> [].
 
-pi_head(Module:Name/Arity, Module:Head) :-
+pi_uhead(Module:Name/Arity, Module:Head) :-
     !,
     atom(Module), atom(Name), integer(Arity),
     functor(Head, Name, Arity).
-pi_head(Name/Arity, user:Head) :-
+pi_uhead(Name/Arity, user:Head) :-
     atom(Name), integer(Arity),
     functor(Head, Name, Arity).
 
+qlf_recompile_reason(old) -->
+    !,
+    [ ' (out of date)'-[] ].
+qlf_recompile_reason(_) -->
+    [ ' (incompatible with current Prolog version)'-[] ].
+
 prolog_message(file_search(cache(Spec, _Cond), Path)) -->
     [ 'File search: ~p --> ~p (cache)'-[Spec, Path] ].
 prolog_message(file_search(found(Spec, Cond), Path)) -->
@@ -726,26 +809,6 @@ prolog_message(file_search(tried(Spec, Cond), Path)) -->
                  *              GC              *
                  *******************************/
 
-prolog_message(gc(start)) -->
-    thread_context,
-    [ 'GC: ', flush ].
-prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
-    [ at_same_line,
-      'gained ~D+~D in ~3f sec; used ~D+~D; free ~D+~D'-
-      [G, T, Time, UG, UT, RG, RT]
-    ].
-prolog_message(shift_stacks(start(_L,_G,_T))) -->
-    thread_context,
-    [ 'Stack-shift: ', flush ].
-prolog_message(shift_stacks(done(Time, L, G, T))) -->
-    { LKB is L//1024,
-      GKB is G//1024,
-      TKB is T//1024
-    },
-    [ at_same_line,
-      'local: ~DKB, global: ~DKB, trail: ~DKB bytes (~2f sec)'-
-      [LKB, GKB, TKB, Time]
-    ].
 prolog_message(agc(start)) -->
     thread_context,
     [ 'AGC: ', flush ].
@@ -764,6 +827,118 @@ prolog_message(cgc(done(CollectedClauses, _CollectedBytes,
       [CollectedClauses, Time, RemainingBytes]
     ].
 
+		 /*******************************
+		 *        STACK OVERFLOW	*
+		 *******************************/
+
+out_of_stack(Context) -->
+    { human_stack_size(Context.localused,   Local),
+      human_stack_size(Context.globalused,  Global),
+      human_stack_size(Context.trailused,   Trail),
+      human_stack_size(Context.stack_limit, Limit),
+      LCO is (100*(Context.depth - Context.environments))/Context.depth
+    },
+    [ 'Stack limit (~s) exceeded'-[Limit], nl,
+      '  Stack sizes: local: ~s, global: ~s, trail: ~s'-[Local,Global,Trail], nl,
+      '  Stack depth: ~D, last-call: ~0f%, Choice points: ~D'-
+         [Context.depth, LCO, Context.choicepoints], nl
+    ],
+    overflow_reason(Context, Resolve),
+    resolve_overflow(Resolve).
+
+human_stack_size(Size, String) :-
+    Size < 100,
+    format(string(String), '~dKb', [Size]).
+human_stack_size(Size, String) :-
+    Size < 100 000,
+    Value is Size / 1024,
+    format(string(String), '~1fMb', [Value]).
+human_stack_size(Size, String) :-
+    Value is Size / (1024*1024),
+    format(string(String), '~1fGb', [Value]).
+
+overflow_reason(Context, fix) -->
+    show_non_termination(Context),
+    !.
+overflow_reason(Context, enlarge) -->
+    { Stack = Context.get(stack) },
+    !,
+    [ '  In:'-[], nl ],
+    stack(Stack).
+overflow_reason(_Context, enlarge) -->
+    [ '  Insufficient global stack'-[] ].
+
+show_non_termination(Context) -->
+    (   { Stack = Context.get(cycle) }
+    ->  [ '  Probable infinite recursion (cycle):'-[], nl ]
+    ;   { Stack = Context.get(non_terminating) }
+    ->  [ '  Possible non-terminating recursion:'-[], nl ]
+    ),
+    stack(Stack).
+
+stack([]) --> [].
+stack([frame(Depth, M:Goal, _)|T]) -->
+    [ '    [~D] ~q:'-[Depth, M] ],
+    stack_goal(Goal),
+    [ nl ],
+    stack(T).
+
+stack_goal(Goal) -->
+    { compound(Goal),
+      !,
+      compound_name_arity(Goal, Name, Arity)
+    },
+    [ '~q('-[Name] ],
+    stack_goal_args(1, Arity, Goal),
+    [ ')'-[] ].
+stack_goal(Goal) -->
+    [ '~q'-[Goal] ].
+
+stack_goal_args(I, Arity, Goal) -->
+    { I =< Arity,
+      !,
+      arg(I, Goal, A),
+      I2 is I + 1
+    },
+    stack_goal_arg(A),
+    (   { I2 =< Arity }
+    ->  [ ', '-[] ],
+        stack_goal_args(I2, Arity, Goal)
+    ;   []
+    ).
+stack_goal_args(_, _, _) -->
+    [].
+
+stack_goal_arg(A) -->
+    { nonvar(A),
+      A = [Len|T],
+      !
+    },
+    (   {Len == cyclic_term}
+    ->  [ '[cyclic list]'-[] ]
+    ;   {T == []}
+    ->  [ '[length:~D]'-[Len] ]
+    ;   [ '[length:~D|~p]'-[Len, T] ]
+    ).
+stack_goal_arg(A) -->
+    { nonvar(A),
+      A = _/_,
+      !
+    },
+    [ '<compound ~p>'-[A] ].
+stack_goal_arg(A) -->
+    [ '~p'-[A] ].
+
+resolve_overflow(fix) -->
+    [].
+resolve_overflow(enlarge) -->
+    { current_prolog_flag(stack_limit, LimitBytes),
+      NewLimit is LimitBytes * 2
+    },
+    [ nl,
+      'Use the --stack_limit=size[KMG] command line option or'-[], nl,
+      '?- set_prolog_flag(stack_limit, ~I). to double the limit.'-[NewLimit]
+    ].
 
 
                  /*******************************
@@ -783,6 +958,30 @@ prolog_message(autoload(Pred, File)) -->
     [ 'autoloading ~p from ~w'-[Pred, File] ].
 prolog_message(autoload(read_index(Dir))) -->
     [ 'Loading autoload index for ~w'-[Dir] ].
+prolog_message(autoload(disabled(Loaded))) -->
+    [ 'Disabled autoloading (loaded ~D files)'-[Loaded] ].
+prolog_message(autoload(already_defined(PI, From))) -->
+    [ ansi(code, '~p', [PI]) ],
+    (   { '$pi_head'(PI, Head),
+          predicate_property(Head, built_in)
+        }
+    ->  [' is a built-in predicate']
+    ;   [ ' is already imported from module ',
+          ansi(code, '~p', [From])
+        ]
+    ).
+
+swi_message(autoload(Msg)) -->
+    [ nl, '  ' ],
+    autoload_message(Msg).
+
+autoload_message(not_exported(PI, Spec, _FullFile, _Exports)) -->
+    [ ansi(code, '~w', [Spec]),
+      ' does not export ',
+      ansi(code, '~p', [PI])
+    ].
+autoload_message(no_file(Spec)) -->
+    [ ansi(code, '~p', [Spec]), ': No such file' ].
 
 
                  /*******************************
@@ -843,10 +1042,9 @@ compiler_warning(neq_singleton(A,B), _Clause, Options) -->
     [ 'Test is always true: ~W'-[A\==B, Options] ].
 compiler_warning(unify_singleton(A,B), _Clause, Options) -->
     [ 'Unified variable is not used: ~W'-[A=B, Options] ].
-compiler_warning(var_true(A), _Clause, Options) -->
-    [ 'Test is always true: ~W'-[var(A), Options] ].
-compiler_warning(nonvar_false(A), _Clause, Options) -->
-    [ 'Test is always false: ~W'-[nonvar(A), Options] ].
+compiler_warning(always(Bool, Pred, Arg), _Clause, Options) -->
+    { Goal =.. [Pred,Arg] },
+    [ 'Test is always ~w: ~W'-[Bool, Goal, Options] ].
 compiler_warning(unbalanced_var(V), _Clause, Options) -->
     [ 'Variable not introduced in all branches: ~W'-[V, Options] ].
 compiler_warning(branch_singleton(V), _Clause, Options) -->
@@ -894,14 +1092,17 @@ prolog_message(copyright) -->
       'Please run ?- license. for legal details.'
     ].
 prolog_message(user_versions) -->
-    { findall(Msg, prolog:version_msg(Msg), Msgs) },
-    user_version_messages(Msgs).
+    (   { findall(Msg, prolog:version_msg(Msg), Msgs),
+          Msgs \== []
+        }
+    ->  [nl],
+        user_version_messages(Msgs)
+    ;   []
+    ).
 prolog_message(documentaton) -->
-    [ 'For online help and background, visit http://www.swi-prolog.org', nl,
+    [ 'For online help and background, visit https://www.swi-prolog.org', nl,
       'For built-in help, use ?- help(Topic). or ?- apropos(Word).'
     ].
-prolog_message(author) -->
-    [ 'Jan Wielemaker (jan@swi-prolog.org)' ].
 prolog_message(welcome) -->
     [ 'Welcome to SWI-Prolog (' ],
     prolog_message(threads),
@@ -916,11 +1117,12 @@ prolog_message(welcome) -->
     prolog_message(documentaton),
     [ nl, nl ].
 prolog_message(about) -->
-    [ 'SWI-Prolog version ' ],
+    [ 'SWI-Prolog version (' ],
+    prolog_message(threads),
+    prolog_message(address_bits),
+    ['version ' ],
     prolog_message(version),
-    [ ' by ' ],
-    prolog_message(author),
-    [ nl ],
+    [ ')', nl ],
     prolog_message(copyright).
 prolog_message(halt) -->
     [ 'halt' ].
@@ -941,23 +1143,23 @@ prolog_message(query(QueryResult)) -->
     query_result(QueryResult).
 
 query_result(no) -->            % failure
-    [ ansi([bold,fg(red)], 'false.', []) ],
+    [ ansi(truth(false), 'false.', []) ],
     extra_line.
-query_result(yes([])) -->      % prompt_alternatives_on: groundness
+query_result(yes(true, [])) -->      % prompt_alternatives_on: groundness
     !,
-    [ ansi(bold, 'true.', []) ],
+    [ ansi(truth(true), 'true.', []) ],
     extra_line.
-query_result(yes(Residuals)) -->
-    result([], Residuals),
+query_result(yes(Delays, Residuals)) -->
+    result([], Delays, Residuals),
     extra_line.
 query_result(done) -->          % user typed <CR>
     extra_line.
-query_result(yes(Bindings, Residuals)) -->
-    result(Bindings, Residuals),
-    prompt(yes, Bindings, Residuals).
-query_result(more(Bindings, Residuals)) -->
-    result(Bindings, Residuals),
-    prompt(more, Bindings, Residuals).
+query_result(yes(Bindings, Delays, Residuals)) -->
+    result(Bindings, Delays, Residuals),
+    prompt(yes, Bindings, Delays, Residuals).
+query_result(more(Bindings, Delays, Residuals)) -->
+    result(Bindings, Delays, Residuals),
+    prompt(more, Bindings, Delays, Residuals).
 query_result(help) -->
     [ nl, 'Actions:'-[], nl, nl,
       '; (n, r, space, TAB): redo    t:          trace & redo'-[], nl,
@@ -975,16 +1177,16 @@ query_result(eof) -->
 query_result(toplevel_open_line) -->
     [].
 
-prompt(Answer, [], []-[]) -->
+prompt(Answer, [], true, []-[]) -->
     !,
     prompt(Answer, empty).
-prompt(Answer, _, _) -->
+prompt(Answer, _, _, _) -->
     !,
     prompt(Answer, non_empty).
 
 prompt(yes, empty) -->
     !,
-    [ ansi(bold, 'true.', []) ],
+    [ ansi(truth(true), 'true.', []) ],
     extra_line.
 prompt(yes, _) -->
     !,
@@ -992,18 +1194,29 @@ prompt(yes, _) -->
     extra_line.
 prompt(more, empty) -->
     !,
-    [ ansi(bold, 'true ', []), flush ].
+    [ ansi(truth(true), 'true ', []), flush ].
 prompt(more, _) -->
     !,
     [ ' '-[], flush ].
 
-result(Bindings, Residuals) -->
+result(Bindings, Delays, Residuals) -->
     { current_prolog_flag(answer_write_options, Options0),
-      Options = [partial(true)|Options0]
+      Options = [partial(true)|Options0],
+      GOptions = [priority(999)|Options0]
     },
+    wfs_residual_program(Delays, GOptions),
     bindings(Bindings, [priority(699)|Options]),
-    bind_res_sep(Bindings, Residuals),
-    residuals(Residuals, [priority(999)|Options]).
+    (   {Residuals == []-[]}
+    ->  bind_delays_sep(Bindings, Delays),
+        delays(Delays, GOptions)
+    ;   bind_res_sep(Bindings, Residuals),
+        residuals(Residuals, GOptions),
+        (   {Delays == true}
+        ->  []
+        ;   [','-[], nl],
+            delays(Delays, GOptions)
+        )
+    ).
 
 bindings([], _) -->
     [].
@@ -1035,7 +1248,7 @@ value(Name, Skel, Subst, Options) -->
 
 substitution([], _) --> !.
 substitution([N=V|T], Options) -->
-    [ ', ', ansi(fg(green), '% where', []), nl,
+    [ ', ', ansi(comment, '% where', []), nl,
       '    ~w = ~W'-[N,V,Options] ],
     substitutions(T, Options).
 
@@ -1050,7 +1263,7 @@ residuals(Normal-Hidden, Options) -->
     bind_res_sep(Normal, Hidden),
     (   {Hidden == []}
     ->  []
-    ;   [ansi(fg(green), '% with pending residual goals', []), nl]
+    ;   [ansi(comment, '% with pending residual goals', []), nl]
     ),
     residuals1(Hidden, Options).
 
@@ -1063,11 +1276,56 @@ residuals1([G|Gs], Options) -->
     ;   [ '~W'-[G, Options] ]
     ).
 
+wfs_residual_program(true, _Options) -->
+    !.
+wfs_residual_program(Goal, _Options) -->
+    { current_prolog_flag(toplevel_list_wfs_residual_program, true),
+      '$current_typein_module'(TypeIn),
+      (   current_predicate(delays_residual_program/2)
+      ->  true
+      ;   use_module(library(wfs), [delays_residual_program/2])
+      ),
+      delays_residual_program(TypeIn:Goal, TypeIn:Program),
+      Program \== []
+    },
+    !,
+    [ ansi(comment, '% WFS residual program', []), nl ],
+    [ ansi(wfs(residual_program), '~@', ['$messages':list_clauses(Program)]) ].
+wfs_residual_program(_, _) --> [].
+
+delays(true, _Options) -->
+    !.
+delays(Goal, Options) -->
+    { current_prolog_flag(toplevel_list_wfs_residual_program, true)
+    },
+    !,
+    [ ansi(truth(undefined), '~W', [Goal, Options]) ].
+delays(_, _Options) -->
+    [ ansi(truth(undefined), undefined, []) ].
+
+:- public list_clauses/1.
+
+list_clauses([]).
+list_clauses([H|T]) :-
+    (   system_undefined(H)
+    ->  true
+    ;   portray_clause(user_output, H, [indent(4)])
+    ),
+    list_clauses(T).
+
+system_undefined((undefined :- tnot(undefined))).
+system_undefined((answer_count_restraint :- tnot(answer_count_restraint))).
+system_undefined((radial_restraint :- tnot(radial_restraint))).
+
 bind_res_sep(_, []) --> !.
 bind_res_sep(_, []-[]) --> !.
 bind_res_sep([], _) --> !.
 bind_res_sep(_, _) --> [','-[], nl].
 
+bind_delays_sep([], _) --> !.
+bind_delays_sep(_, true) --> !.
+bind_delays_sep(_, _) --> [','-[], nl].
+
 extra_line -->
     { current_prolog_flag(toplevel_extra_white_line, true) },
     !,
@@ -1190,20 +1448,18 @@ prolog_message(frame(Frame, backtrace, _PC)) -->
     !,
     { prolog_frame_attribute(Frame, level, Level)
     },
-    [ ansi(bold, '~t[~D] ~10|', [Level]) ],
+    [ ansi(frame(level), '~t[~D] ~10|', [Level]) ],
     frame_context(Frame),
     frame_goal(Frame).
 prolog_message(frame(Frame, choice, PC)) -->
     !,
     prolog_message(frame(Frame, backtrace, PC)).
 prolog_message(frame(_, cut_call, _)) --> !, [].
-prolog_message(frame(Frame, trace(Port), _PC)) -->
+prolog_message(frame(Goal, trace(Port))) -->
     !,
     [ ' T ' ],
     port(Port),
-    frame_level(Frame),
-    frame_context(Frame),
-    frame_goal(Frame).
+    goal(Goal).
 prolog_message(frame(Frame, Port, _PC)) -->
     frame_flags(Frame),
     port(Port),
@@ -1214,8 +1470,12 @@ prolog_message(frame(Frame, Port, _PC)) -->
     [ flush ].
 
 frame_goal(Frame) -->
-    { prolog_frame_attribute(Frame, goal, Goal0),
-      clean_goal(Goal0, Goal),
+    { prolog_frame_attribute(Frame, goal, Goal)
+    },
+    goal(Goal).
+
+goal(Goal0) -->
+    { clean_goal(Goal0, Goal),
       current_prolog_flag(debugger_write_options, Options)
     },
     [ '~W'-[Goal, Options] ].
@@ -1255,17 +1515,17 @@ frame_flags(Frame) -->
     [ '~w~w '-[T, S] ].
 
 port(Port) -->
-    { port_name(Port, Colour, Name)
+    { port_name(Port, Name)
     },
     !,
-    [ ansi([bold,fg(Colour)], '~w: ', [Name]) ].
+    [ ansi(port(Port), '~w: ', [Name]) ].
 
-port_name(call,      green,   'Call').
-port_name(exit,      green,   'Exit').
-port_name(fail,      red,     'Fail').
-port_name(redo,      yellow,  'Redo').
-port_name(unify,     blue,    'Unify').
-port_name(exception, magenta, 'Exception').
+port_name(call,      'Call').
+port_name(exit,      'Exit').
+port_name(fail,      'Fail').
+port_name(redo,      'Redo').
+port_name(unify,     'Unify').
+port_name(exception, 'Exception').
 
 clean_goal(M:Goal, Goal) :-
     hidden_module(M),
@@ -1319,24 +1579,92 @@ prolog_message(pack(no_arch(Entry, Arch))) -->
 
 prolog_message(null_byte_in_path(Component)) -->
     [ '0-byte in PATH component: ~p (skipped directory)'-[Component] ].
-prolog_message(invalid_tmp_var(Var, Value, Reason)) -->
-    [ 'Cannot use '-[] ], env(Var),
-    [ ' as temporary file directory: ~p: ~w'-[Value, Reason] ].
+prolog_message(invalid_tmp_dir(Dir, Reason)) -->
+    [ 'Cannot use ~p as temporary file directory: ~w'-[Dir, Reason] ].
+prolog_message(ambiguous_stream_pair(Pair)) -->
+    [ 'Ambiguous operation on stream pair ~p'-[Pair] ].
+prolog_message(backcomp(init_file_moved(FoundFile))) -->
+    { absolute_file_name(app_config('init.pl'), InitFile,
+                         [ file_errors(fail)
+                         ])
+    },
+    [ 'The location of the config file has moved'-[], nl,
+      '  from "~w"'-[FoundFile], nl,
+      '  to   "~w"'-[InitFile], nl,
+      '  See https://www.swi-prolog.org/modified/config-files.html'-[]
+    ].
+
+		 /*******************************
+		 *          DEPRECATED		*
+		 *******************************/
+
+deprecated(Term) -->
+    prolog:deprecated(Term),
+    !.
+deprecated(set_prolog_stack(_Stack,limit)) -->
+    [ 'set_prolog_stack/2: limit(Size) sets the combined limit.'-[], nl,
+      'See https://www.swi-prolog.org/changes/stack-limit.html'
+    ].
+
+		 /*******************************
+		 *           TRIPWIRES		*
+		 *******************************/
+
+tripwire_message(Wire, Context) -->
+    [ 'Trapped tripwire ~w for '-[Wire] ],
+    tripwire_context(Wire, Context).
+
+tripwire_context(_, ATrie) -->
+    { '$is_answer_trie'(ATrie),
+      !,
+      '$tabling':atrie_goal(ATrie, QGoal),
+      user_predicate_indicator(QGoal, Goal)
+    },
+    [ '~p'-[Goal] ].
+tripwire_context(_, Ctx) -->
+    [ '~p'-[Ctx] ].
+
+
+		 /*******************************
+		 *        DEFAULT THEME		*
+		 *******************************/
+
+:- public default_theme/2.
+
+default_theme(var,                    [fg(red)]).
+default_theme(code,                   [fg(blue)]).
+default_theme(comment,                [fg(green)]).
+default_theme(warning,                [fg(red)]).
+default_theme(error,                  [bold, fg(red)]).
+default_theme(truth(false),           [bold, fg(red)]).
+default_theme(truth(true),            [bold]).
+default_theme(truth(undefined),       [bold, fg(cyan)]).
+default_theme(wfs(residual_program),  [fg(cyan)]).
+default_theme(frame(level),           [bold]).
+default_theme(port(call),             [bold, fg(green)]).
+default_theme(port(exit),             [bold, fg(green)]).
+default_theme(port(fail),             [bold, fg(red)]).
+default_theme(port(redo),             [bold, fg(yellow)]).
+default_theme(port(unify),            [bold, fg(blue)]).
+default_theme(port(exception),        [bold, fg(magenta)]).
+default_theme(message(informational), [fg(green)]).
+default_theme(message(information),   [fg(green)]).
+default_theme(message(debug(_)),      [fg(blue)]).
+default_theme(message(Level),         Attrs) :-
+    nonvar(Level),
+    default_theme(Level, Attrs).
 
-env(Name) -->
-    { current_prolog_flag(windows, true) },
-    [ '%~w%'-[Name] ].
-env(Name) -->
-    [ '$~w'-[Name] ].
 
                  /*******************************
                  *      PRINTING MESSAGES       *
                  *******************************/
 
 :- multifile
-    user:message_hook/3.
+    user:message_hook/3,
+    prolog:message_prefix_hook/2.
 :- dynamic
-    user:message_hook/3.
+    user:message_hook/3,
+    prolog:message_prefix_hook/2.
 :- thread_local
     user:thread_message_hook/3.
 
@@ -1346,6 +1674,35 @@ env(Name) -->
 %   system.
 
 print_message(Level, Term) :-
+    setup_call_cleanup(
+        push_msg(Term),
+        print_message_guarded(Level, Term),
+        pop_msg),
+    !.
+print_message(Level, Term) :-
+    (   Level \== silent
+    ->  format(user_error, 'Recursive ~w message: ~q~n', [Level, Term])
+    ;   true
+    ).
+
+push_msg(Term) :-
+    nb_current('$inprint_message', Messages),
+    !,
+    \+ ( '$member'(Msg, Messages),
+         Msg =@= Term
+       ),
+    b_setval('$inprint_message', [Term|Messages]).
+push_msg(Term) :-
+    b_setval('$inprint_message', [Term]).
+
+pop_msg :-
+    (   nb_current('$inprint_message', [_|Messages]),
+        Messages \== []
+    ->  b_setval('$inprint_message', Messages)
+    ;   nb_delete('$inprint_message')
+    ).
+
+print_message_guarded(Level, Term) :-
     (   must_print(Level, Term)
     ->  (   translate_message(Term, Lines, [])
         ->  (   nonvar(Term),
@@ -1381,7 +1738,7 @@ print_system_message(Term, Kind, Lines) :-
     Term \= error(syntax_error(_), _),
     msg_property(Kind, location_prefix(File:Line, LocPrefix, LinePrefix)),
     !,
-    insert_prefix(Lines, LinePrefix, PrefixLines),
+    insert_prefix(Lines, LinePrefix, Ctx, PrefixLines),
     '$append'([ begin(Kind, Ctx),
                 LocPrefix,
                 nl
@@ -1417,34 +1774,70 @@ msg_property(query, stream(user_output)) :- !.
 msg_property(_, stream(user_error)) :- !.
 msg_property(error,
              location_prefix(File:Line,
-                             '~NERROR: ~w:~d:'-[File,Line], '~N\t')) :- !.
+                             '~NERROR: ~w:~d:'-[File,Line],
+                             '~NERROR:    ')) :- !.
 msg_property(warning,
              location_prefix(File:Line,
-                             '~NWarning: ~w:~d:'-[File,Line], '~N\t')) :- !.
+                             '~NWarning: ~w:~d:'-[File,Line],
+                             '~NWarning:    ')) :- !.
 msg_property(error,   wait(0.1)) :- !.
 
-msg_prefix(debug(_),      '~N% ').
-msg_prefix(warning,           Prefix) :-
-    (   thread_message_id(Id)
-    ->  Prefix = '~NWarning: [Thread ~w] '-Id
-    ;   Prefix = '~NWarning: '
-    ).
-msg_prefix(error,             Prefix) :-
-    (   thread_message_id(Id)
-    ->  Prefix = '~NERROR: [Thread ~w] '-Id
-    ;   Prefix = '~NERROR: '
-    ).
+msg_prefix(debug(_), Prefix) :-
+    msg_context('~N% ', Prefix).
+msg_prefix(warning, Prefix) :-
+    msg_context('~NWarning: ', Prefix).
+msg_prefix(error, Prefix) :-
+    msg_context('~NERROR: ', Prefix).
 msg_prefix(informational, '~N% ').
 msg_prefix(information,   '~N% ').
 
-thread_message_id(Id) :-
+%!  msg_context(+Prefix0, -Prefix) is det.
+%
+%   Add contextual information to a message.   This uses the Prolog flag
+%   `message_context`. Recognised context terms are:
+%
+%     - time
+%     - time(Format)
+%     - thread
+%
+%   In addition, the hook prolog:message_prefix_hook/2   is  called that
+%   allows for additional context information.
+
+msg_context(Prefix0, Prefix) :-
+    current_prolog_flag(message_context, Context),
+    is_list(Context),
+    !,
+    add_message_context(Context, Prefix0, Prefix).
+msg_context(Prefix, Prefix).
+
+add_message_context([], Prefix, Prefix).
+add_message_context([H|T], Prefix0, Prefix) :-
+    (   add_message_context1(H, Prefix0, Prefix1)
+    ->  true
+    ;   Prefix1 = Prefix0
+    ),
+    add_message_context(T, Prefix1, Prefix).
+
+add_message_context1(Context, Prefix0, Prefix) :-
+    prolog:message_prefix_hook(Context, Extra),
+    atomics_to_string([Prefix0, Extra, ' '], Prefix).
+add_message_context1(time, Prefix0, Prefix) :-
+    get_time(Now),
+    format_time(string(S), '%T.%3f ', Now),
+    string_concat(Prefix0, S, Prefix).
+add_message_context1(time(Format), Prefix0, Prefix) :-
+    get_time(Now),
+    format_time(string(S), Format, Now),
+    atomics_to_string([Prefix0, S, ' '], Prefix).
+add_message_context1(thread, Prefix0, Prefix) :-
     thread_self(Id0),
     Id0 \== main,
-    \+ current_prolog_flag(thread_message_prefix, false),
+    !,
     (   atom(Id0)
     ->  Id = Id0
     ;   thread_property(Id0, id(Id))
-    ).
+    ),
+    format(string(Prefix), '~w[Thread ~w] ', [Prefix0, Id]).
 
 %!  print_message_lines(+Stream, +PrefixOrKind, +Lines)
 %
@@ -1454,7 +1847,7 @@ thread_message_id(Id) :-
 print_message_lines(Stream, kind(Kind), Lines) :-
     !,
     msg_property(Kind, prefix(Prefix)),
-    insert_prefix(Lines, Prefix, PrefixLines),
+    insert_prefix(Lines, Prefix, Ctx, PrefixLines),
     '$append'([ begin(Kind, Ctx)
               | PrefixLines
               ],
@@ -1463,25 +1856,29 @@ print_message_lines(Stream, kind(Kind), Lines) :-
               AllLines),
     print_message_lines(Stream, AllLines).
 print_message_lines(Stream, Prefix, Lines) :-
-    insert_prefix(Lines, Prefix, PrefixLines),
+    insert_prefix(Lines, Prefix, _, PrefixLines),
     print_message_lines(Stream, PrefixLines).
 
-%!  insert_prefix(+Lines, +Prefix, -PrefixedLines)
+%!  insert_prefix(+Lines, +Prefix, +Ctx, -PrefixedLines)
 
-insert_prefix([at_same_line|Lines0], Prefix, Lines) :-
+insert_prefix([at_same_line|Lines0], Prefix, Ctx, Lines) :-
     !,
-    prefix_nl(Lines0, Prefix, Lines).
-insert_prefix(Lines0, Prefix, [prefix(Prefix)|Lines]) :-
-    prefix_nl(Lines0, Prefix, Lines).
+    prefix_nl(Lines0, Prefix, Ctx, Lines).
+insert_prefix(Lines0, Prefix, Ctx, [prefix(Prefix)|Lines]) :-
+    prefix_nl(Lines0, Prefix, Ctx, Lines).
 
-prefix_nl([], _, [nl]).
-prefix_nl([nl], _, [nl]) :- !.
-prefix_nl([flush], _, [flush]) :- !.
-prefix_nl([nl|T0], Prefix, [nl, prefix(Prefix)|T]) :-
+prefix_nl([], _, _, [nl]).
+prefix_nl([nl], _, _, [nl]) :- !.
+prefix_nl([flush], _, _, [flush]) :- !.
+prefix_nl([nl|T0], Prefix, Ctx, [nl, prefix(Prefix)|T]) :-
     !,
-    prefix_nl(T0, Prefix, T).
-prefix_nl([H|T0], Prefix, [H|T]) :-
-    prefix_nl(T0, Prefix, T).
+    prefix_nl(T0, Prefix, Ctx, T).
+prefix_nl([ansi(Attrs,Fmt,Args)|T0], Prefix, Ctx,
+          [ansi(Attrs,Fmt,Args,Ctx)|T]) :-
+    !,
+    prefix_nl(T0, Prefix, Ctx, T).
+prefix_nl([H|T0], Prefix, Ctx, [H|T]) :-
+    prefix_nl(T0, Prefix, Ctx, T).
 
 %!  print_message_lines(+Stream, +Lines)
 
@@ -1506,24 +1903,46 @@ line_element(S, nl) :-
     nl(S).
 line_element(S, prefix(Fmt-Args)) :-
     !,
-    format(S, Fmt, Args).
+    safe_format(S, Fmt, Args).
 line_element(S, prefix(Fmt)) :-
     !,
-    format(S, Fmt, []).
+    safe_format(S, Fmt, []).
 line_element(S, flush) :-
     !,
     flush_output(S).
 line_element(S, Fmt-Args) :-
     !,
-    format(S, Fmt, Args).
+    safe_format(S, Fmt, Args).
 line_element(S, ansi(_, Fmt, Args)) :-
     !,
-    format(S, Fmt, Args).
+    safe_format(S, Fmt, Args).
+line_element(S, ansi(_, Fmt, Args, _Ctx)) :-
+    !,
+    safe_format(S, Fmt, Args).
 line_element(_, begin(_Level, _Ctx)) :- !.
 line_element(_, end(_Ctx)) :- !.
 line_element(S, Fmt) :-
-    format(S, Fmt, []).
+    safe_format(S, Fmt, []).
 
+%!  safe_format(+Stream, +Format, +Args) is det.
+
+safe_format(S, Fmt, Args) :-
+    E = error(_,_),
+    catch(format(S,Fmt,Args), E,
+          format_failed(S,Fmt,Args,E)).
+
+format_failed(S, _Fmt, _Args, E) :-
+    E = error(io_error(_,S),_),
+    !,
+    throw(E).
+format_failed(S, Fmt, Args, error(E,_)) :-
+    format(S, '~N    [[ EXCEPTION while printing message ~q~n\c
+                        ~7|with arguments ~W:~n\c
+                        ~7|raised: ~W~n~4|]]~n',
+           [ Fmt,
+             Args, [quoted(true), max_depth(10)],
+             E, [quoted(true), max_depth(10)]
+           ]).
 
 %!  message_to_string(+Term, -String)
 %
@@ -1544,15 +1963,20 @@ actions_to_format([nl|T], Fmt, Args) :-
     !,
     actions_to_format(T, Fmt0, Args),
     atom_concat('~n', Fmt0, Fmt).
-actions_to_format([Skip|T], Fmt, Args) :-
-    action_skip(Skip),
+actions_to_format([ansi(_Attrs, Fmt0, Args0)|Tail], Fmt, Args) :-
     !,
-    actions_to_format(T, Fmt, Args).
+    actions_to_format(Tail, Fmt1, Args1),
+    atom_concat(Fmt0, Fmt1, Fmt),
+    append_args(Args0, Args1, Args).
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :-
     !,
     actions_to_format(Tail, Fmt1, Args1),
     atom_concat(Fmt0, Fmt1, Fmt),
     append_args(Args0, Args1, Args).
+actions_to_format([Skip|T], Fmt, Args) :-
+    action_skip(Skip),
+    !,
+    actions_to_format(T, Fmt, Args).
 actions_to_format([Term|Tail], Fmt, Args) :-
     atomic(Term),
     !,
@@ -1565,7 +1989,6 @@ actions_to_format([Term|Tail], Fmt, Args) :-
 
 action_skip(at_same_line).
 action_skip(flush).
-action_skip(ansi(_Attrs, _Fmt, _Args)).
 action_skip(begin(_Level, _Ctx)).
 action_skip(end(_Ctx)).
 
@@ -1591,6 +2014,7 @@ append_args(Args0, Args1, Args) :-
 
 print_once(compatibility(_), _).
 print_once(null_byte_in_path(_), _).
+print_once(deprecated(_), _).
 
 %!  must_print(+Level, +Message)
 %
