@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker and Richard O'Keefe
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2014-2017, VU University Amsterdam
+    Copyright (c)  2014-2019, VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -34,8 +35,20 @@
 
 :- module(check_installation,
           [ check_installation/0,
-            check_installation/1                % -Issues
+            check_installation/1,               % -Issues
+            check_config_files/0,
+            update_config_files/0,
+            test_installation/0,
+            test_installation/1                 % +Options
           ]).
+:- autoload(library(apply),[maplist/2,maplist/3]).
+:- autoload(library(archive),[archive_open/3,archive_close/1]).
+:- autoload(library(lists),[append/3,member/2]).
+:- autoload(library(option),[option/2,merge_options/3]).
+:- autoload(library(pcre),[re_config/1]).
+:- autoload(library(prolog_source),[path_segments_atom/2]).
+:- use_module(library(settings),[setting/2]).
+
 
 /** <module> Check installation issues and features
 
@@ -68,6 +81,10 @@ shared objects/DLLs can be loaded.
 %     features.
 
 % Feature tests
+component(tcmalloc,
+          _{ test:test_tcmalloc,
+             url:'tcmalloc.html'
+           }).
 component(gmp,
           _{ test:current_prolog_flag(bounded, false),
              url:'gmp.html'
@@ -93,6 +110,7 @@ component(library(pcre), _{features:pcre_features}).
 component(library(pdt_console), _{}).
 component(library(porter_stem), _{}).
 component(library(process), _{}).
+component(library(protobufs), _{}).
 component(library(editline), _{os:unix}).
 component(library(readline), _{os:unix}).
 component(library(readutil), _{}).
@@ -105,6 +123,7 @@ component(library(sha), _{}).
 component(library(snowball), _{}).
 component(library(socket), _{}).
 component(library(ssl), _{}).
+component(library(crypto), _{}).
 component(library(syslog), _{os:unix}).
 component(library(table), _{}).
 component(library(time), _{}).
@@ -113,6 +132,7 @@ component(library(unicode), _{}).
 component(library(uri), _{}).
 component(library(uuid), _{}).
 component(library(zlib), _{}).
+component(library(yaml), _{}).
 
 issue_base('http://www.swi-prolog.org/build/issues/').
 
@@ -131,21 +151,26 @@ issue_base('http://www.swi-prolog.org/build/issues/').
 %        are present (e.g., unbounded arithmetic support)
 %     2. Test that all standard libraries that depend on foreign
 %        code are present.
+%     3. provides a test_installation predicate to run the tests
+%        at runtime if the system was built with -DINSTALL_TESTS
 %
 %   If issues are found it prints a   diagnostic message with a link
 %   to a wiki page with additional information about the issue.
 
 check_installation :-
     print_message(informational, installation(checking)),
-    check_installation_(Issues),
+    check_installation_(InstallIssues),
     check_on_path,
+    check_config_files(ConfigIssues),
+    maplist(print_message(warning), ConfigIssues),
+    append(InstallIssues, ConfigIssues, Issues),
     (   Issues == []
     ->  print_message(informational, installation(perfect))
     ;   length(Issues, Count),
         print_message(warning, installation(imperfect(Count)))
     ).
 
-%!  check_insmtallation(-Issues:list(pair)) is det.
+%!  check_installation(-Issues:list(pair)) is det.
 %
 %   As check_installation/0, but additionally  returns   a  list  of
 %   Component-Problem pairs. Problem is  one of `optional_not_found`
@@ -291,6 +316,17 @@ error_kind(error).
                  *         SPECIAL TESTS        *
                  *******************************/
 
+%!  test_tcmalloc
+
+:- if(current_predicate(malloc_property/1)).
+test_tcmalloc :-
+    malloc_property('generic.current_allocated_bytes'(Bytes)),
+    Bytes > 1 000 000.
+:- else.
+test_tcmalloc :-
+    fail.
+:- endif.
+
 %!  archive_features
 %
 %   Report features supported by library(archive).
@@ -384,8 +420,13 @@ jquery_file :-
 %   Validate that Prolog is installed in $PATH
 
 check_on_path :-
-    current_prolog_flag(executable, EXE),
+    current_prolog_flag(executable, EXEFlag),
+    prolog_to_os_filename(EXE, EXEFlag),
     file_base_name(EXE, Prog),
+    absolute_file_name(EXE, AbsExe,
+                       [ access(execute)
+                       ]),
+    prolog_to_os_filename(AbsExe, OsExe),
     (   absolute_file_name(path(Prog), OnPath,
                            [ access(execute),
                              file_errors(fail)
@@ -398,13 +439,65 @@ check_on_path :-
                                  solutions(all)
                                ]),
             same_file(EXE, OnPathAny)
-        ->  print_message(warning, installation(not_first_on_path(EXE, OnPath)))
-        ;   print_message(warning, installation(not_same_on_path(EXE, OnPath)))
+        ->  print_message(warning, installation(not_first_on_path(OsExe, OnPath)))
+        ;   print_message(warning, installation(not_same_on_path(OsExe, OnPath)))
         )
-    ;   print_message(warning, installation(not_on_path(EXE, Prog)))
+    ;   print_message(warning, installation(not_on_path(OsExe, Prog)))
     ).
 
 
+		 /*******************************
+		 *           RUN TESTS		*
+		 *******************************/
+
+%!  test_installation is semidet.
+%!  test_installation(+Options) is semidet.
+%
+%   Run regression tests in the installed system. Requires the system to
+%   be built using
+%
+%	cmake -DINSTALL_TESTS=ON
+%
+%   Options processed:
+%
+%     - packages(+Boolean)
+%       When `false`, do not test the packages
+%     - package(+Package)
+%       Only test package package.
+
+test_installation :-
+    test_installation([]).
+
+test_installation(Options) :-
+    absolute_file_name(swi(test/test),
+                       TestFile,
+                       [ access(read),
+                         file_errors(fail),
+                         file_type(prolog)
+                       ]),
+    !,
+    test_installation_run(TestFile, Options).
+test_installation(_Options) :-
+    print_message(warning, installation(testing(no_installed_tests))).
+
+test_installation_run(TestFile, Options) :-
+    (   option(package(_), Options)
+    ->  merge_options(Options,
+                      [ core(false),
+                        subdirs(false)
+                      ], TestOptions)
+    ;   merge_options(Options,
+                      [ packages(true)
+                      ], TestOptions)
+    ),
+    load_files(user:TestFile),
+    current_prolog_flag(verbose, Old),
+    setup_call_cleanup(
+        set_prolog_flag(verbose, silent),
+        user:test([], TestOptions),
+        set_prolog_flag(verbose, Old)).
+
+
                  /*******************************
                  *            MESSAGES          *
                  *******************************/
@@ -504,6 +597,9 @@ message(jquery(found(Path))) -->
     [ '  jQuery from ~w'-[Path] ].
 message(jquery(not_found(File))) -->
     [ '  Cannot find jQuery (~w)'-[File] ].
+message(testing(no_installed_tests)) -->
+    [ '  Runtime testing is not enabled.', nl],
+    [ '  Please recompile the system with INSTALL_TESTS enabled.' ].
 
 
 public_executable(EXE, PublicProg) :-
@@ -569,3 +665,143 @@ list_names([H|T]) -->
         list_names(T)
     ).
 
+
+		 /*******************************
+		 *          CONFIG FILES	*
+		 *******************************/
+
+%!  check_config_files
+%
+%   Examines the locations of config files.  The config files have moved
+%   in version 8.1.15
+
+check_config_files :-
+    check_config_files(Issues),
+    maplist(print_message(warning), Issues).
+
+check_config_files(Issues) :-
+    findall(Issue, check_config_file(Issue), Issues).
+
+check_config_file(config(Id, move(Type, OldFile, NewFile))) :-
+    old_config(Type, Id, OldFile),
+    access_file(OldFile, exist),
+    \+ ( new_config(Type, Id, NewFile),
+         access_file(NewFile, exist)
+       ),
+    once(new_config(Type, Id, NewFile)).
+check_config_file(config(Id, different(Type, OldFile, NewFile))) :-
+    old_config(Type, Id, OldFile),
+    access_file(OldFile, exist),
+    new_config(Type, Id, NewFile),
+    access_file(NewFile, exist),
+    \+ same_file(OldFile, NewFile).
+
+%!  update_config_files
+%
+%   Move config files from their old location to  the new if the file or
+%   directory exists in the old location but not in the new.
+
+update_config_files :-
+    old_config(Type, Id, OldFile),
+    access_file(OldFile, exist),
+    \+ ( new_config(Type, Id, NewFile),
+         access_file(NewFile, exist)
+       ),
+    (   new_config(Type, Id, NewFile),
+        \+ same_file(OldFile, NewFile),
+        create_parent_dir(NewFile)
+    ->  catch(rename_file(OldFile, NewFile), E,
+              print_message(warning, E)),
+        print_message(informational, config(Id, moved(Type, OldFile, NewFile)))
+    ),
+    fail.
+update_config_files.
+
+old_config(file, init, File) :-
+    current_prolog_flag(windows, true),
+    win_folder(appdata, Base),
+    atom_concat(Base, '/SWI-Prolog/swipl.ini', File).
+old_config(file, init, File) :-
+    expand_file_name('~/.swiplrc', [File]).
+old_config(directory, lib, Dir) :-
+    expand_file_name('~/lib/prolog', [Dir]).
+old_config(directory, xpce, Dir) :-
+    expand_file_name('~/.xpce', [Dir]).
+old_config(directory, history, Dir) :-
+    expand_file_name('~/.swipl-dir-history', [Dir]).
+old_config(directory, pack, Dir) :-
+    (   catch(expand_file_name('~/lib/swipl/pack', [Dir]), _, fail)
+    ;   absolute_file_name(swi(pack), Dir,
+                           [ file_type(directory), solutions(all) ])
+    ).
+
+new_config(file, init, File) :-
+    absolute_file_name(user_app_config('init.pl'), File,
+                       [ solutions(all) ]).
+new_config(directory, lib, Dir) :-
+    config_dir(user_app_config(lib), Dir).
+new_config(directory, xpce, Dir) :-
+    config_dir(user_app_config(xpce), Dir).
+new_config(directory, history, Dir) :-
+    config_dir(user_app_config('dir-history'), Dir).
+new_config(directory, pack, Dir) :-
+    config_dir([app_data(pack), swi(pack)], Dir).
+
+config_dir(Aliases, Dir) :-
+    is_list(Aliases),
+    !,
+    (   member(Alias, Aliases),
+        absolute_file_name(Alias, Dir,
+                           [ file_type(directory), solutions(all) ])
+    *-> true
+    ;   member(Alias, Aliases),
+        absolute_file_name(Alias, Dir,
+                           [ solutions(all) ])
+    ).
+config_dir(Alias, Dir) :-
+    (   absolute_file_name(Alias, Dir,
+                           [ file_type(directory), solutions(all) ])
+    *-> true
+    ;   absolute_file_name(Alias, Dir,
+                           [ solutions(all) ])
+    ).
+
+create_parent_dir(NewFile) :-
+    file_directory_name(NewFile, Dir),
+    create_parent_dir_(Dir).
+
+create_parent_dir_(Dir) :-
+    exists_directory(Dir),
+    '$my_file'(Dir),
+    !.
+create_parent_dir_(Dir) :-
+    file_directory_name(Dir, Parent),
+    Parent \== Dir,
+    create_parent_dir_(Parent),
+    make_directory(Dir).
+
+prolog:message(config(Id, Issue)) -->
+    [ 'Config: '-[] ],
+    config_description(Id),
+    config_issue(Issue).
+
+config_description(init) -->
+    [ '(user initialization file) '-[], nl ].
+config_description(lib) -->
+    [ '(user library) '-[], nl ].
+config_description(pack) -->
+    [ '(add-ons) '-[], nl ].
+config_description(history) -->
+    [ '(command line history) '-[], nl ].
+config_description(xpce) -->
+    [ '(gui) '-[], nl ].
+
+config_issue(move(Type, Old, New)) -->
+    [ '  found ~w "~w"'-[Type, Old], nl ],
+    [ '  new location is "~w"'-[New] ].
+config_issue(moved(Type, Old, New)) -->
+    [ '  found ~w "~w"'-[Type, Old], nl ],
+    [ '  moved to new location "~w"'-[New] ].
+config_issue(different(Type, Old, New)) -->
+    [ '  found different ~w "~w"'-[Type, Old], nl ],
+    [ '  new location is "~w"'-[New] ].
