@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2012-2018, VU University Amsterdam
+    Copyright (c)  2012-2019, VU University Amsterdam
                               CWI, Amsterdam
     All rights reserved.
 
@@ -61,7 +61,7 @@
 :- use_module(library(http/http_open)).
 :- use_module(library(http/json)).
 :- use_module(library(http/http_client), []).   % plugin for POST support
-
+:- use_module(library(prolog_config)).
 
 /** <module> A package manager for Prolog
 
@@ -90,7 +90,7 @@ makes installed packages available as libaries.
                  *          CONSTANTS           *
                  *******************************/
 
-:- setting(server, atom, 'http://www.swi-prolog.org/pack/',
+:- setting(server, atom, 'https://www.swi-prolog.org/pack/',
            'Server to exchange pack information').
 
 
@@ -1085,6 +1085,7 @@ foreign_file('configure.ac').
 foreign_file('configure').
 foreign_file('Makefile').
 foreign_file('makefile').
+foreign_file('CMakeLists.txt').
 
 
 %!  configure_foreign(+PackDir, +Options) is det.
@@ -1092,6 +1093,11 @@ foreign_file('makefile').
 %   Run configure if it exists.  If =|configure.ac|= or =|configure.in|=
 %   exists, first run =autoheader= and =autoconf=
 
+configure_foreign(PackDir, Options) :-
+    directory_file_path(PackDir, 'CMakeLists.txt', CMakeFile),
+    exists_file(CMakeFile),
+    !,
+    cmake_configure_foreign(PackDir, Options).
 configure_foreign(PackDir, Options) :-
     make_configure(PackDir, Options),
     directory_file_path(PackDir, configure, Configure),
@@ -1120,6 +1126,18 @@ make_configure(_, _).
 autoconf_master('configure.ac').
 autoconf_master('configure.in').
 
+%!  cmake_configure_foreign(+PackDir, +Options) is det.
+%
+%   Create a `build` directory in PackDir and run `cmake ..`
+
+cmake_configure_foreign(PackDir, _Options) :-
+    directory_file_path(PackDir, build, BuildDir),
+    make_directory_path(BuildDir),
+    current_prolog_flag(executable, Exe),
+    format(atom(CDEF), '-DSWIPL=~w', [Exe]),
+    run_process(path(cmake), [CDEF, '..'],
+                [directory(BuildDir)]).
+
 
 %!  make_foreign(+PackDir, +Options) is det.
 %
@@ -1136,8 +1154,31 @@ pack_make(PackDir, Targets, _Options) :-
     ProcessOptions = [ directory(PackDir), env(BuildEnv) ],
     forall(member(Target, Targets),
            run_process(path(make), [Target], ProcessOptions)).
+pack_make(PackDir, Targets, _Options) :-
+    directory_file_path(PackDir, 'CMakeLists.txt', CMakefile),
+    exists_file(CMakefile),
+    directory_file_path(PackDir, 'build', BuildDir),
+    exists_directory(BuildDir),
+    !,
+    (   Targets == [distclean]
+    ->  delete_directory_contents(BuildDir)
+    ;   build_environment(BuildEnv),
+        ProcessOptions = [ directory(BuildDir), env(BuildEnv) ],
+        forall(member(Target, Targets),
+               run_cmake_target(Target, BuildDir, ProcessOptions))
+    ).
 pack_make(_, _, _).
 
+run_cmake_target(check, BuildDir, ProcessOptions) :-
+    !,
+    (   directory_file_path(BuildDir, 'CTestTestfile.cmake', TestFile),
+        exists_file(TestFile)
+    ->  run_process(path(ctest), [], ProcessOptions)
+    ;   true
+    ).
+run_cmake_target(Target, _, ProcessOptions) :-
+    run_process(path(make), [Target], ProcessOptions).
+
 %!  save_build_environment(+PackDir)
 %
 %   Create  a  shell-script  build.env  that    contains  the  build
@@ -1253,6 +1294,9 @@ def_environment('LDSOFLAGS', Value) :-
     ->  current_prolog_flag(home, Home),
         atomic_list_concat(['-L"', Home, '/bin"'], SystemLib),
         System = [SystemLib]
+    ;   apple_bundle_libdir(LibDir)
+    ->  atomic_list_concat(['-L"', LibDir, '"'], SystemLib),
+        System = [SystemLib]
     ;   current_prolog_flag(c_libplso, '')
     ->  System = []                 % ELF systems do not need this
     ;   prolog_library_dir(SystemLibDir),
