@@ -2,7 +2,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2004-2018, University of Amsterdam
+    Copyright (c)  2004-2020, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
     All rights reserved.
@@ -179,6 +179,7 @@ typedef struct _plsocket
 { int		    magic;		/* PLSOCK_MAGIC */
   SOCKET	    socket;		/* The OS socket */
   int		    flags;		/* Misc flags */
+  int		    domain;		/* AF_* */
   atom_t	    symbol;		/* <socket>(%p) */
   IOSTREAM *	    input;		/* input stream */
   IOSTREAM *	    output;		/* output stream */
@@ -519,6 +520,11 @@ is_nbio_socket(nbio_sock_t socket)
 { return socket && socket->magic == PLSOCK_MAGIC;
 }
 
+int
+nbio_domain(nbio_sock_t socket)
+{ return socket->domain;
+}
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Allocate a wrapper for an OS  socket.   The  wrapper  is allocated in an
 array of pointers, to keep small  integer   identifiers  we can use with
@@ -857,7 +863,21 @@ nbio_socket(int domain, int type, int protocol)
   { closesocket(sock);
     return NULL;
   }
-
+  s->domain = domain;
+#ifdef __WINDOWS__
+  /* On older versions of Windows (win7 and before) the default send
+     buffer size is only 8k. On a high latency link this can seriously
+     hamper performance. Default it to max(64k, current-value)
+  */
+  { int val;
+    int val_len = sizeof(val);
+    if (getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&val, &val_len) == 0 &&
+	val < 65535)
+    { val = 65535;
+      setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&val, sizeof(int));
+    }
+  }
+#endif
   return s;
 }
 
@@ -1013,6 +1033,18 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 
       break;
     }
+    case TCP_SNDBUF:
+    { int val = va_arg(args, int);
+      if ( setsockopt(socket->socket, SOL_SOCKET, SO_SNDBUF,
+		      (const char *)&val, sizeof(int)) == -1 )
+      { nbio_error(GET_ERRNO, TCP_ERRNO);
+	rc = -1;
+      } else
+      { rc = 0;
+      }
+
+      break;
+    }
     default:
       rc = -1;
       assert(0);
@@ -1201,9 +1233,12 @@ nbio_connect(nbio_sock_t socket,
 #endif
 	continue;
       }
-#ifdef __WINDOWS__
+#if defined(WSAEISCONN)				 /* Windows */
       if ( GET_ERRNO == WSAEISCONN )
         break;
+#elif defined(EISCONN)
+      if ( GET_ERRNO == EISCONN )
+        break;
 #endif
       nbio_error(GET_ERRNO, TCP_ERRNO);
       return -1;
