@@ -79,6 +79,14 @@ test(null_error, [condition(has_exe(sh))]) :-
     process_create(path(sh),
                    ['-c', 'echo "THIS IS AN ERROR" 1>&2'],
                    [stderr(null)]).
+test(stream_input, [condition(has_exe(wc)), X == "0"]) :-
+    open('/dev/null', read, In, [type(binary)]),
+    process_create(path(wc),
+                   ['-c'],
+                   [stdin(stream(In)), stdout(pipe(Out))]),
+    close(In),
+    read_process(Out, X0),
+    split_string(X0, "", " \r\n", [X]).
 test(read_error, [condition(has_exe(sh)),X == 'error\n']) :-
     process_create(path(sh),
                    ['-c', 'echo "error" 1>&2'],
@@ -107,7 +115,8 @@ test(cwd, [true, condition(\+current_prolog_flag(windows, true))]) :-
     read_process(Out, CWD0),
     normalize_space(atom(CWD), CWD0),
     same_file(CWD, Tmp).
-test(cwd, [true, condition(current_prolog_flag(windows, true))]) :-
+test(cwd, [true, condition(( current_prolog_flag(windows, true),
+                             \+current_prolog_flag(wine_version, _)))]) :-
     tmp_dir(Tmp),
     getenv('COMSPEC', Shell),
     process_create(Shell, ['/c', cd],
