@@ -54,7 +54,8 @@
 :- use_module(library(main)).
 :- use_module(library(readutil)).
 
-:- if(exists_source(library(http/http_ssl_plugin))).
+:- if(( exists_source(library(http/http_ssl_plugin)),
+        \+ current_prolog_flag(pldoc_to_tex,true))).
 :- use_module(library(ssl)).
 :- use_module(library(http/http_ssl_plugin)).
 :- endif.
@@ -98,7 +99,7 @@ code.  This is often a sequence of use_module/1 directives.
   ==
 
 The   program   entry   point   is     http_daemon/0,   declared   using
-initialization/2. This main be overruled using   a new declaration after
+initialization/2. This may be overruled using   a  new declaration after
 loading  this  library.  The  new  entry    point  will  typically  call
 http_daemon/1 to start the server in a preconfigured way.
 
@@ -827,7 +828,8 @@ wait(Options) :-
     repeat,
         State = deadline(Count),
         Deadline is FirstDeadline+Count*Interval,
-        (   thread_get_message(Me, Msg, [deadline(Deadline)])
+        (   thread_idle(thread_get_message(Me, Msg, [deadline(Deadline)]),
+                        long)
         ->  catch(ignore(handle_message(Msg)), E,
                   print_message(error, E)),
             Msg == quit,
@@ -841,7 +843,7 @@ wait(Options) :-
 wait(_) :-
     thread_self(Me),
     repeat,
-        thread_get_message(Me, Msg),
+        thread_idle(thread_get_message(Me, Msg), long),
         catch(ignore(handle_message(Msg)), E,
               print_message(error, E)),
         Msg == quit,
