@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2007-2018, University of Amsterdam
+    Copyright (c)  2007-2020, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
     All rights reserved.
@@ -34,8 +34,6 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-:- if(current_predicate(is_dict/1)).
-
 :- module(json,
           [ json_read/2,                % +Stream, -JSONTerm
             json_read/3,                % +Stream, -JSONTerm, +Options
@@ -51,29 +49,17 @@
             json_write_dict/3,          % +Stream, +Dict, +Options
             atom_json_dict/3            % ?Atom, ?JSONDict, +Options
           ]).
-
-:- else.
-
-:- module(json,
-          [ json_read/2,                % +Stream, -JSONTerm
-            json_read/3,                % +Stream, -JSONTerm, +Options
-            atom_json_term/3,           % ?Atom, ?JSONTerm, +Options
-            json_write/2,               % +Stream, +Term
-            json_write/3,               % +Stream, +Term, +Options
-            is_json_term/1,             % @Term
-            is_json_term/2
-          ]).
-
-:- endif.
-
 :- use_module(library(record)).
 :- use_module(library(memfile)).
 :- use_module(library(error)).
 :- use_module(library(option)).
-:- use_module(library(debug)).
+:- use_module(library(lists)).
 
 :- use_foreign_library(foreign(json)).
 
+:- multifile
+    json_write_hook/4.                          % +Term, +Stream, +State, +Options
+
 :- predicate_options(json_read/3, 3,
                      [ null(ground),
                        true(ground),
@@ -136,12 +122,13 @@ terms.
               null:ground = @(null),
               true:ground = @(true),
               false:ground = @(false),
+              end_of_file:ground = error,
               value_string_as:oneof([atom,string]) = atom,
               tag:atom = '',
               default_tag:atom).
 
 default_json_dict_options(
-    json_options(null, true, false, string, '', _)).
+    json_options(null, true, false, error, string, '', _)).
 
 
                  /*******************************
@@ -210,7 +197,7 @@ type_term(chars,  Result, chars(Result)).
 %   Here is a complete example in  JSON and its corresponding Prolog
 %   term.
 %
-%     ==
+%     ```
 %     { "name":"Demo term",
 %       "created": {
 %         "day":null,
@@ -220,47 +207,79 @@ type_term(chars,  Result, chars(Result)).
 %       "confirmed":true,
 %       "members":[1,2,3]
 %     }
-%     ==
+%     ```
 %
-%     ==
+%     ```
 %     json([ name='Demo term',
 %            created=json([day= @null, month='December', year=2007]),
 %            confirmed= @true,
 %            members=[1, 2, 3]
 %          ])
-%     ==
+%     ```
 %
 %   The following options are processed:
 %
-%           * null(+NullTerm)
-%           Term used to represent JSON =null=.  Default @(null)
-%           * true(+TrueTerm)
-%           Term used to represent JSON =true=.  Default @(true)
-%           * false(+FalseTerm)
-%           Term used to represent JSON =false=.  Default @(false)
-%           * value_string_as(+Type)
-%           Prolog type used for strings used as value.  Default
-%           is =atom=.  The alternative is =string=, producing a
-%           packed string object.  Please note that =codes= or
-%           =chars= would produce ambiguous output and is therefore
-%           not supported.
+%     - null(+NullTerm)
+%       Term used to represent JSON =null=.  Default @(null)
+%     - true(+TrueTerm)
+%       Term used to represent JSON =true=.  Default @(true)
+%     - false(+FalseTerm)
+%       Term used to represent JSON =false=.  Default @(false)
+%     - end_of_file(+ErrorOrTerm)
+%       If end of file is reached after skipping white space
+%       but before any input is processed take the following
+%       action (default `error`):
+%         - If ErrorOrTerm == `error`, throw an unexpected
+%           end of file syntax error
+%         - Otherwise return ErrorOrTerm.
+%       Returning an status term is required to process
+%       [Concatenated
+%       JSON](https://en.wikipedia.org/wiki/JSON_streaming#Concatenated_JSON).
+%       Suggested values are `@(eof)` or `end_of_file`.
+%     - value_string_as(+Type)
+%       Prolog type used for strings used as value. Default is `atom`.
+%       The alternative is `string`, producing a packed string object.
+%       Please note that `codes` or `chars` would produce ambiguous
+%       output and are therefore not supported.
 %
 %   @see    json_read_dict/3 to read a JSON term using the version 7
 %           extended data types.
 
 json_read(Stream, Term) :-
     default_json_options(Options),
-    (   json_value(Stream, Term, _, Options)
+    (   json_value_top(Stream, Term, Options)
     ->  true
     ;   syntax_error(illegal_json, Stream)
     ).
 json_read(Stream, Term, Options) :-
     make_json_options(Options, OptionTerm, _RestOptions),
-    (   json_value(Stream, Term, _, OptionTerm)
+    (   json_value_top(Stream, Term, OptionTerm)
     ->  true
     ;   syntax_error(illegal_json, Stream)
     ).
 
+json_value_top(Stream, Term, Options) :-
+    stream_property(Stream, type(binary)),
+    !,
+    setup_call_cleanup(
+        set_stream(Stream, encoding(utf8)),
+        json_value_top_(Stream, Term, Options),
+        set_stream(Stream, type(binary))).
+json_value_top(Stream, Term, Options) :-
+    json_value_top_(Stream, Term, Options).
+
+json_value_top_(Stream, Term, Options) :-
+    get_code(Stream, C0),
+    ws(C0, Stream, C1),
+    (   C1 == -1
+    ->  json_options_end_of_file(Options, Action),
+        (   Action == error
+        ->  syntax_error(unexpected_end_of_file, Stream)
+        ;   Term = Action
+        )
+    ;   json_term_top(C1, Stream, Term, Options)
+    ).
+
 json_value(Stream, Term, Next, Options) :-
     get_code(Stream, C0),
     ws(C0, Stream, C1),
@@ -269,34 +288,33 @@ json_value(Stream, Term, Next, Options) :-
     ;   json_term(C1, Stream, Term, Next, Options)
     ).
 
-json_term(0'{, Stream, json(Pairs), Next, Options) :-
+json_term(C0, Stream, JSON, Next, Options) :-
+    json_term_top(C0, Stream, JSON, Options),
+    get_code(Stream, Next).
+
+json_term_top(0'{, Stream, json(Pairs), Options) :-
     !,
     ws(Stream, C),
-    json_pairs(C, Stream, Pairs, Options),
-    get_code(Stream, Next).
-json_term(0'[, Stream, Array, Next, Options) :-
+    json_pairs(C, Stream, Pairs, Options).
+json_term_top(0'[, Stream, Array, Options) :-
     !,
     ws(Stream, C),
-    json_array(C, Stream, Array, Options),
-    get_code(Stream, Next).
-json_term(0'", Stream, String, Next, Options) :-
+    json_array(C, Stream, Array, Options).
+json_term_top(0'", Stream, String, Options) :-
     !,
     get_code(Stream, C1),
     json_string_codes(C1, Stream, Codes),
     json_options_value_string_as(Options, Type),
-    codes_to_type(Type, Codes, String),
-    get_code(Stream, Next).
-json_term(0'-, Stream, Number, Next, _Options) :-
+    codes_to_type(Type, Codes, String).
+json_term_top(0'-, Stream, Number, _Options) :-
     !,
-    json_read_number(Stream, 0'-, Number, Next).
-json_term(D, Stream, Number, Next, _Options) :-
+    json_read_number(Stream, 0'-, Number).
+json_term_top(D, Stream, Number, _Options) :-
     between(0'0, 0'9, D),
     !,
-    json_read_number(Stream, D, Number, Next).
-json_term(C, Stream, Constant, Next, Options) :-
-    get_code(Stream, C1),
-    json_identifier_codes(C1, Stream, Codes, Next),
-    atom_codes(ID, [C|Codes]),
+    json_read_number(Stream, D, Number).
+json_term_top(C, Stream, Constant, Options) :-
+    json_read_constant(C, Stream, ID),
     json_constant(ID, Constant, Options).
 
 json_pairs(0'}, _, [], _) :- !.
@@ -378,13 +396,24 @@ escape(0'u, Stream, C) :-
     code_type(C4, xdigit(D4)),
     C is D1<<12+D2<<8+D3<<4+D4.
 
-json_identifier_codes(C1, Stream, [C1|T], Next) :-
-    between(0'a, 0'z, C1),
+json_read_constant(0't, Stream, true) :-
     !,
-    get_code(Stream, C2),
-    json_identifier_codes(C2, Stream, T, Next).
-json_identifier_codes(C, _, [], C).
+    must_see(`rue`, Stream, true).
+json_read_constant(0'f, Stream, false) :-
+    !,
+    must_see(`alse`, Stream, false).
+json_read_constant(0'n, Stream, null) :-
+    !,
+    must_see(`ull`, Stream, null).
 
+must_see([], _Stream, _).
+must_see([H|T], Stream, Name) :-
+    get_code(Stream, C),
+    (   C == H
+    ->  true
+    ;   syntax_error(json_expected(Name), Stream)
+    ),
+    must_see(T, Stream, Name).
 
 json_constant(true, Constant, Options) :-
     !,
@@ -442,19 +471,23 @@ stream_error_context(Stream, stream(Stream, Line, LinePos, CharNo)) :-
 %!  json_write(+Stream, +Term) is det.
 %!  json_write(+Stream, +Term, +Options) is det.
 %
-%   Write a JSON term to Stream.  The   JSON  object  is of the same
-%   format as produced by json_read/2, though we allow for some more
-%   flexibility with regard to pairs in  objects. All of Name=Value,
-%   Name-Value and Name(Value) produce the  same output.
+%   Write a JSON term to Stream. The JSON   object is of the same format
+%   as  produced  by  json_read/2,  though  we    allow  for  some  more
+%   flexibility with regard to pairs  in   objects.  All  of Name=Value,
+%   Name-Value and Name(Value) produce the same output.
 %
-%   Values can be of the form  #(Term),   which  causes `Term` to be
-%   _stringified_ if it is not an atom or string. Stringification is
+%   Values can be of  the  form  #(Term),   which  causes  `Term`  to be
+%   _stringified_ if it is not  an   atom  or string. Stringification is
 %   based on term_string/2.
 %
-%   The version 7 _dict_ type is supported as well.  Optionally,  if
-%   the dict has a _tag_,  a property  "type":"tag"  can be added to
-%   the object.  This behaviour can be controlled using the  =tag=
-%   option (see below). For example:
+%   Rational numbers are emitted as  floating   point  numbers. The hook
+%   json_write_hook/4  can  be  used   to    realize   domain   specific
+%   alternatives.
+%
+%   The version 7 _dict_ type is supported   as well. Optionally, if the
+%   dict has a _tag_, a  property  "type":"tag"   can  be  added  to the
+%   object. This behaviour can be controlled using the =tag= option (see
+%   below). For example:
 %
 %     ==
 %     ?- json_write(current_output, point{x:1,y:2}).
@@ -502,6 +535,25 @@ stream_error_context(Stream, stream(Stream, Line, LinePos, CharNo)) :-
 %   objects  can  be  safely  embedded  into  an  HTML  =|<script>|=
 %   element.
 
+%!  json_write_hook(+Term, +Stream, +State, +Options) is semidet.
+%
+%   Hook that can be used to  emit   a  JSON  representation for Term to
+%   Stream. If the  predicate  succeeds  it   __must__  have  written  a
+%   __valid__ JSON data element and if it fails it may not have produced
+%   any output. This facility may be used  to map arbitrary Prolog terms
+%   to JSON. It was added to manage   the  precision with which floating
+%   point numbers are emitted.
+%
+%   Note that this hook is shared by all   users  of this library. It is
+%   generally  adviced  to  map  a  unique    compound   term  to  avoid
+%   interference with normal output.
+%
+%   @arg State and Options are opaque handles to the current output
+%   state and settings.  Future versions may provide documented access
+%   to these terms.  Currently it is adviced to ignore these arguments.
+
+
+
 :- record json_write_state(indent:nonneg = 0,
                        step:positive_integer = 2,
                        tab:positive_integer = 8,
@@ -523,7 +575,6 @@ json_write_term(Var, _, _, _) :-
 json_write_term(json(Pairs), Stream, State, Options) :-
     !,
     json_write_object(Pairs, Stream, State, Options).
-:- if(current_predicate(is_dict/1)).
 json_write_term(Dict, Stream, State, Options) :-
     is_dict(Dict),
     !,
@@ -535,7 +586,6 @@ json_write_term(Dict, Stream, State, Options) :-
     ;   Pairs = Pairs0
     ),
     json_write_object(Pairs, Stream, State, Options).
-:- endif.
 json_write_term(List, Stream, State, Options) :-
     is_list(List),
     !,
@@ -555,10 +605,20 @@ json_write_term(List, Stream, State, Options) :-
         indent(Stream, State),
         write(Stream, ']')
     ).
+
+json_write_term(Term, Stream, State, Options) :-
+    json_write_hook(Term, Stream, State, Options),
+    !.
 json_write_term(Number, Stream, _State, _Options) :-
     number(Number),
     !,
-    write(Stream, Number).
+    (   float(Number)
+    ->  write(Stream, Number)
+    ;   integer(Number)
+    ->  write(Stream, Number)
+    ;   Float is float(Number)              % rational number
+    ->  write(Stream, Float)
+    ).
 json_write_term(True, Stream, _State, Options) :-
     json_options_true(Options, True),
     !,
@@ -713,7 +773,6 @@ json_print_length(json(Pairs), Options, Max, Len0, Len) :-
     Len1 =< Max,
     must_be(list, Pairs),
     pairs_print_length(Pairs, Options, Max, Len1, Len).
-:- if(current_predicate(is_dict/1)).
 json_print_length(Dict, Options, Max, Len0, Len) :-
     is_dict(Dict),
     !,
@@ -721,7 +780,6 @@ json_print_length(Dict, Options, Max, Len0, Len) :-
     Len1 is Len0 + 2,
     Len1 =< Max,
     pairs_print_length(Pairs, Options, Max, Len1, Len).
-:- endif.
 json_print_length(Array, Options, Max, Len0, Len) :-
     is_list(Array),
     !,
@@ -857,8 +915,6 @@ is_json_pair(Options, Name=Value) :-
     atom(Name),
     is_json_term2(Options, Value).
 
-:- if(current_predicate(is_dict/1)).
-
                  /*******************************
                  *         DICT SUPPORT         *
                  *******************************/
@@ -895,6 +951,8 @@ is_json_pair(Options, Name=Value) :-
 %       Default the atom `true`.
 %     * false(+FalseTerm)
 %       Default the atom `false`
+%     * end_of_file(+ErrorOrTerm)
+%       Action on reading end-of-file. See json_read/3 for details.
 %     * value_string_as(+Type)
 %       Prolog type used for strings used as value.  Default
 %       is `string`.  The alternative is `atom`, producing a
@@ -905,7 +963,7 @@ json_read_dict(Stream, Dict) :-
 
 json_read_dict(Stream, Dict, Options) :-
     make_json_dict_options(Options, OptionTerm, _RestOptions),
-    (   json_value(Stream, Term, _, OptionTerm)
+    (   json_value_top(Stream, Term, OptionTerm)
     ->  true
     ;   syntax_error(illegal_json, Stream)
     ),
@@ -931,8 +989,16 @@ term_to_dict(Value0, Value, _Options) :-
     !,
     Value = Value0.
 term_to_dict(List0, List, Options) :-
-    assertion(is_list(List0)),
+    is_list(List0),
+    !,
     terms_to_dicts(List0, List, Options).
+term_to_dict(Special, Special, Options) :-
+    (   json_options_true(Options, Special)
+    ;   json_options_false(Options, Special)
+    ;   json_options_null(Options, Special)
+    ;   json_options_end_of_file(Options, Special)
+    ),
+    !.
 
 json_dict_pairs([], [], _).
 json_dict_pairs([Name=Value0|T0], [Name=Value|T], Options) :-
@@ -986,9 +1052,7 @@ atom_json_dict(Atom, Term, Options) :-
     ground(Atom),
     !,
     setup_call_cleanup(
-        ( text_memfile(Atom, MF),
-          open_memory_file(MF, read, In, [free_on_close(true)])
-        ),
+        open_string(Atom, In),
         json_read_dict(In, Term, Options),
         close(In)).
 atom_json_dict(Result, Term, Options) :-
@@ -1000,17 +1064,6 @@ atom_json_dict(Result, Term, Options) :-
     with_output_to(Out,
                    json_write_dict(current_output, Term, Options1)).
 
-text_memfile(Atom, MF) :-
-    atom(Atom),
-    !,
-    atom_to_memory_file(Atom, MF).
-text_memfile(String, MF) :-
-    string(String),
-    !,
-    new_memory_file(MF),
-    insert_memory_file(MF, 0, String).
-
-:- endif.
 
                  /*******************************
                  *           MESSAGES           *
