@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2002-2018, University of Amsterdam
+    Copyright (c)  2002-2020, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
     All rights reserved.
@@ -83,6 +83,10 @@
 
 /** <module> Threaded HTTP server
 
+Most   code   doesn't   need  to   use  this   directly;  instead   use
+library(http/http_server),  which  combines   this  library  with   the
+typical HTTP libraries that most servers need.
+
 This library defines the HTTP server  frontend of choice for SWI-Prolog.
 It is based on the multi-threading   capabilities of SWI-Prolog and thus
 exploits multiple cores  to  serve   requests  concurrently.  The server
@@ -121,6 +125,9 @@ self-signed SSL certificate.
     http:create_pool/1,
     http:schedule_workers/1.
 
+:- meta_predicate
+    thread_repeat_wait(0).
+
 %!  http_server(:Goal, :Options) is det.
 %
 %   Create a server at Port that calls Goal for each parsed request.
@@ -217,7 +224,7 @@ make_socket(Port, _:Options0, Options) :-
     tcp_socket(Socket),
     tcp_setopt(Socket, reuseaddr),
     tcp_bind(Socket, Port),
-    tcp_listen(Socket, 5),
+    tcp_listen(Socket, 64),
     make_addr_atom('httpd', Port, Queue),
     Options = [ queue(Queue),
                 tcp_socket(Socket)
@@ -253,6 +260,7 @@ create_server(Goal, Address, Options) :-
     get_time(StartTime),
     memberchk(queue(Queue), Options),
     scheme(Scheme, Options),
+    autoload_https(Scheme),
     address_port(Address, Port),
     make_addr_atom(Scheme, Port, Alias),
     thread_self(Initiator),
@@ -276,6 +284,12 @@ scheme(http, _).
 address_port(_Host:Port, Port) :- !.
 address_port(Port, Port).
 
+autoload_https(https) :-
+    \+ clause(accept_hook(_Goal, _Options), _),
+    exists_source(library(http/http_ssl_plugin)),
+    !,
+    use_module(library(http/http_ssl_plugin)).
+autoload_https(_).
 
 %!  http_current_server(:Goal, ?Port) is nondet.
 %
@@ -402,6 +416,7 @@ accept_server3(Goal, Options) :-
 accept_server3(Goal, Options) :-
     memberchk(tcp_socket(Socket), Options),
     memberchk(queue(Queue), Options),
+    debug(http(connection), 'Waiting for connection', []),
     tcp_accept(Socket, Client, Peer),
     debug(http(connection), 'New HTTP connection from ~p', [Peer]),
     thread_send_message(Queue, tcp_client(Client, Goal, Peer)),
@@ -456,17 +471,23 @@ connect(Address) :-
 %   hook http:schedule_workers/1 to extend  the   worker  pool. This
 %   predicate can be used by accept_hook/2.
 
+http_enough_workers(Queue, _Why, _Peer) :-
+    message_queue_property(Queue, waiting(_0)),
+    !,
+    debug(http(scheduler), '~D waiting for work; ok', [_0]).
 http_enough_workers(Queue, Why, Peer) :-
     message_queue_property(Queue, size(Size)),
     (   enough(Size, Why)
-    ->  true
+    ->  debug(http(scheduler), '~D in queue; ok', [Size])
     ;   current_server(Port, _, _, Queue, _, _),
-        catch(http:schedule_workers(_{port:Port,
-                                      reason:Why,
-                                      peer:Peer,
-                                      waiting:Size,
-                                      queue:Queue
-                                     }),
+        Data = _{ port:Port,
+                  reason:Why,
+                  peer:Peer,
+                  waiting:Size,
+                  queue:Queue
+                },
+        debug(http(scheduler), 'Asking to reschedule: ~p', [Data]),
+        catch(http:schedule_workers(Data),
               Error,
               print_message(error, Error))
     ->  true
@@ -567,19 +588,12 @@ resize_pool(Queue, Size) :-
 %   stops.
 
 http_worker(Options) :-
-    thread_at_exit(done_worker),
+    debug(http(scheduler), 'New worker', []),
+    prolog_listen(this_thread_exit, done_worker),
     option(queue(Queue), Options),
     option(max_idle_time(MaxIdle), Options, infinite),
-    repeat,
-      garbage_collect,
-      trim_stacks,
+    thread_repeat_wait(get_work(Queue, Message, MaxIdle)),
       debug(http(worker), 'Waiting for a job ...', []),
-      (   MaxIdle == infinite
-      ->  thread_get_message(Queue, Message)
-      ;   thread_get_message(Queue, Message, [timeout(MaxIdle)])
-      ->  true
-      ;   Message = quit(idle)
-      ),
       debug(http(worker), 'Got job ~p', [Message]),
       (   Message = quit(Sender)
       ->  !,
@@ -607,6 +621,15 @@ http_worker(Options) :-
           )
       ).
 
+get_work(Queue, Message, infinite) :-
+    !,
+    thread_get_message(Queue, Message).
+get_work(Queue, Message, MaxIdle) :-
+    (   thread_get_message(Queue, Message, [timeout(MaxIdle)])
+    ->  true
+    ;   Message = quit(idle)
+    ).
+
 
 %!  open_client(+Message, +Queue, -Goal, -In, -Out,
 %!              +Options, -ClientOptions) is semidet.
@@ -730,19 +753,21 @@ recreate_worker(exception(Error), Queue) :-
 recreate_on_error('$aborted').
 recreate_on_error(time_limit_exceeded).
 
-%       thread_httpd:message_level(+Exception, -Level)
+%!  thread_httpd:message_level(+Exception, -Level)
 %
-%       Determine the message stream used for  exceptions that may occur
-%       during server_loop/5. Being multifile, clauses   can be added by
-%       the   application   to   refine   error   handling.   See   also
-%       message_hook/3 for further programming error handling.
+%   Determine the message stream used  for   exceptions  that  may occur
+%   during server_loop/5. Being multifile, clauses can   be added by the
+%   application to refine error handling.   See  also message_hook/3 for
+%   further programming error handling.
 
 :- multifile
     message_level/2.
 
-message_level(error(io_error(read, _), _),      silent).
-message_level(error(timeout_error(read, _), _), informational).
-message_level(keep_alive_timeout,               silent).
+message_level(error(io_error(read, _), _),               silent).
+message_level(error(socket_error(epipe,_), _),           silent).
+message_level(error(http_write_short(_Obj,_Written), _), silent).
+message_level(error(timeout_error(read, _), _),          informational).
+message_level(keep_alive_timeout,                        silent).
 
 current_message_level(Term, Level) :-
     (   message_level(Term, Level)
@@ -796,7 +821,12 @@ http_process(Goal, In, Out, Options) :-
                  ]),
     next(Connection, Request).
 
-next(switch_protocol(SwitchGoal, _SwitchOptions), Request) :-
+next(Connection, Request) :-
+    next_(Connection, Request), !.
+next(Connection, Request) :-
+    print_message(warning, goal_failed(next(Connection,Request))).
+
+next_(switch_protocol(SwitchGoal, _SwitchOptions), Request) :-
     !,
     memberchk(pool(client(_Queue, _Goal, In, Out)), Request),
     (   catch(call(SwitchGoal, In, Out), E,
@@ -805,14 +835,14 @@ next(switch_protocol(SwitchGoal, _SwitchOptions), Request) :-
     ->  true
     ;   http_close_connection(Request)
     ).
-next(spawned(ThreadId), _) :-
+next_(spawned(ThreadId), _) :-
     !,
     debug(http(spawn), 'Handler spawned to thread ~w', [ThreadId]).
-next(Connection, Request) :-
+next_(Connection, Request) :-
     downcase_atom(Connection, 'keep-alive'),
     http_requeue(Request),
     !.
-next(_, Request) :-
+next_(_, Request) :-
     http_close_connection(Request).
 
 
@@ -900,6 +930,67 @@ create_pool(Pool) :-
     thread_pool_create(Pool, 10, []).
 
 
+		 /*******************************
+		 *         WAIT POLICIES	*
+		 *******************************/
+
+:- meta_predicate
+    thread_repeat_wait(0).
+
+%!  thread_repeat_wait(:Goal) is multi.
+%
+%   Acts as `repeat,  thread_idle(Goal)`,  choosing   whether  to  use a
+%   `long` or `short` idle time based on the average firing rate.
+
+thread_repeat_wait(Goal) :-
+    new_rate_mma(5, 1000, State),
+    repeat,
+      update_rate_mma(State, MMA),
+      long(MMA, IsLong),
+      (   IsLong == brief
+      ->  call(Goal)
+      ;   thread_idle(Goal, IsLong)
+      ).
+
+long(MMA, brief) :-
+    MMA < 0.05,
+    !.
+long(MMA, short) :-
+    MMA < 1,
+    !.
+long(_, long).
+
+%!  new_rate_mma(+N, +Resolution, -State) is det.
+%!  update_rate_mma(!State, -MMA) is det.
+%
+%   Implement _Modified Moving  Average_  computing   the  average  time
+%   between requests as an exponential moving averate with alpha=1/N.
+%
+%   @arg Resolution is the time resolution  in 1/Resolution seconds. All
+%   storage is done in integers to avoid  the need for stack freezing in
+%   nb_setarg/3.
+%
+%   @see https://en.wikipedia.org/wiki/Moving_average
+
+new_rate_mma(N, Resolution, rstate(Base, 0, MaxI, Resolution, N, 0)) :-
+    current_prolog_flag(max_tagged_integer, MaxI),
+    get_time(Base).
+
+update_rate_mma(State, MMAr) :-
+    State = rstate(Base, Last, MaxI, Resolution, N, MMA0),
+    get_time(Now),
+    Stamp is round((Now-Base)*Resolution),
+    (   Stamp > MaxI
+    ->  nb_setarg(1, State, Now),
+        nb_setarg(2, State, 0)
+    ;   true
+    ),
+    Diff is Stamp-Last,
+    nb_setarg(2, State, Stamp),
+    MMA is round(((N-1)*MMA0+Diff)/N),
+    nb_setarg(6, State, MMA),
+    MMAr is MMA/float(Resolution).
+
 
                  /*******************************
                  *            MESSAGES          *
