@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2006-2018, University of Amsterdam
+    Copyright (c)  2006-2019, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
     All rights reserved.
@@ -145,14 +145,13 @@ take_block(Lines, BaseIndent, List, Rest) :-
     ).
 take_block([N-['|'|RL1]|LT], _, Table, Rest) :-
     phrase(row(R0), RL1),
-    rest_table(LT, N, RL, Rest),
-    !,
-    Table = table(class=wiki, [tr(R0)|RL]).
-take_block([0-[-,-|More]|LT], _, Block, LT) :-  % seperation line
+    take_table(LT, N, R0, Table, Rest),
+    !.
+take_block([0-[-,-|More]|LT], _, Block, LT) :-  % separation line
     maplist(=(-), More),
     !,
     Block = hr([]).
-take_block([_-Line|LT], _, Block, LT) :-        % seperation line
+take_block([_-Line|LT], _, Block, LT) :-        % separation line
     ruler(Line),
     !,
     Block = hr([]).
@@ -171,12 +170,26 @@ take_block([I-L1|LT], BaseIndent, Elem, Rest) :-
     rest_par(LT, PT, I, BaseIndent, MaxI, Rest),
     (   MaxI >= BaseIndent+16
     ->  Elem = center(Par)
-    ;   MaxI >= BaseIndent+4
-    ->  Elem = blockquote(Par)
+    ;   phrase(blockquote(BQ), Par)
+    ->  Elem = blockquote(BQ)
     ;   Elem = p(Par)
     ).
 take_block([Verb|Lines], _, Verb, Lines).
 
+blockquote(Clean) -->
+    [>, ' '],
+    bq_lines(Clean).
+
+bq_lines([' '|Par]) -->
+    ['\n'], !, [>,' '],
+    bq_lines(Par).
+bq_lines([H|T]) -->
+    [H],
+    bq_lines(T).
+bq_lines([]) -->
+    [].
+
+
 %!  ruler(+Line) is semidet.
 %
 %   True if Line contains 3 ruler chars and otherwise spaces.
@@ -393,29 +406,81 @@ face_token(=) --> [=].
 face_token(*) --> [*].
 face_token('_') --> ['_'].
 
-rest_table([N-Line|LT], N, RL, Rest) :-
-    md_table_structure_line(Line),
+take_table(Lines, Indent, Row0, Table, Rest) :-
+    rest_table(Lines, Indent, Rows, Rest),
+    (   Rows = [align(Align)|Rows1]
+    ->  maplist(align_row(Align), Rows1, Rows2),
+        (   maplist(=(td([])), Row0)        % empty header
+        ->  Table = table(class=wiki, Rows2)
+        ;   maplist(td_to_th, Row0, Header),
+            Table = table(class=wiki, [tr(Header)|Rows2])
+        )
+    ;   Table = table(class=wiki, [tr(Row0)|Rows])
+    ).
+
+td_to_th(td(X), th(X)) :- !.
+td_to_th(X, X).
+
+align_row(Align, tr(Row0), tr(Row)) :-
+    align_cells(Align, Row0, Row).
+
+align_cells([Align|AT], [Cell0|T0], [Cell|T]) :-
+    align_cell(Align, Cell0, Cell),
+    align_cells(AT, T0, T).
+align_cells(_, Cells, Cells).
+
+align_cell(Align, td(Content), td(class=Align, Content)).
+
+%!  rest_table(+Lines, +Indent, -Rows, -RestLines).
+
+rest_table([N-Line|LT], N, [align(Align)|RL], Rest) :-
+    phrase(column_alignment(Align), Line),
     !,
-    rest_table(LT, N, RL, Rest).
-rest_table([N-['|'|RL1]|LT], N, [tr(R0)|RL], Rest) :-
+    rest_table2(LT, N, RL, Rest).
+rest_table(Lines, N, RL, Rest) :-
+    rest_table2(Lines, N, RL, Rest).
+
+rest_table2([N-['|'|RL1]|LT], N, [tr(R0)|RL], Rest) :-
     !,
     phrase(row(R0), RL1),
-    rest_table(LT, N, RL, Rest).
-rest_table(Rest, _, [], Rest).
+    rest_table2(LT, N, RL, Rest).
+rest_table2(Rest, _, [], Rest).
 
-%!  md_table_structure_line(+Chars)
+%!  column_alignment(-Alignment) is semidet.
 %
-%   True if Chars represents Markdown  table structure. We currently
-%   ignore the structure information.
+%   Process an alignment line.
 
-md_table_structure_line(Line) :-
-    forall(member(Char, Line),
-           md_table_structure_char(Char)).
+column_alignment([H|T]) -->
+    ['|'],
+    (   colspec(H)
+    ->  column_alignment(T)
+    ;   {T=[]}
+    ).
 
-md_table_structure_char(' ').
-md_table_structure_char('-').
-md_table_structure_char('|').
-md_table_structure_char(':').
+colspec(Align) -->
+    ws_tokens, [':'], dashes3,
+    (   [':']
+    ->  {Align = center}
+    ;   {Align = left}
+    ),
+    ws_tokens.
+colspec(Align) -->
+    ws_tokens, dashes3,
+    (   [':']
+    ->  {Align = right}
+    ;   {Align = left}
+    ),
+    ws_tokens.
+
+dashes3 -->
+    [-,-,-],
+    dashes.
+
+dashes --> [-], !, dashes.
+dashes --> [].
+
+ws_tokens --> [' '], !, ws_tokens.
+ws_tokens --> [].
 
 %!  rest_par(+Lines, -Par,
 %!           +BaseIndent, +MaxI0, -MaxI, -RestLines) is det.
@@ -519,6 +584,7 @@ label_cont(Name) --> [w(Name)].
 md_section_line(Line1, Line2, Header) :-
     Line1 \== [],
     section_underline(Line2, Type),
+    is_list(Line1),
     phrase(wiki_words(_), Line1),  % Should not have structure elements
     !,
     (   phrase(labeled_section_line(Title, Attrs), Line1)
@@ -746,6 +812,7 @@ structure_term(\include(What,Type,Opts), include(What,Type,Opts), []) :- !.
 structure_term(dl(Att, Args), dl(Att), [Args]) :- !.
 structure_term(dt(Att, Args), dt(Att), [Args]) :- !.
 structure_term(table(Att, Args), table(Att), [Args]) :- !.
+structure_term(td(Att, Args), td(Att), [Args]) :- !.
 structure_term(h1(Att, Args), h1(Att), [Args]) :- !.
 structure_term(h2(Att, Args), h2(Att), [Args]) :- !.
 structure_term(h3(Att, Args), h3(Att), [Args]) :- !.
@@ -773,6 +840,7 @@ structure_tag(dd).
 structure_tag(table).
 structure_tag(tr).
 structure_tag(td).
+structure_tag(th).
 structure_tag(blockquote).
 structure_tag(center).
 
@@ -839,9 +907,13 @@ wiki_faces_int(WithFaces, ArgNames) -->
 
 wiki_faces_int([], _, _) -->
     [].
-wiki_faces_int([H|T], ArgNames, Options) -->
+wiki_faces_int(List, ArgNames, Options) -->
     wiki_face(H, ArgNames, Options),
     !,
+    {   is_list(H)
+    ->  append(H, T, List)
+    ;   List = [H|T]
+    },
     wiki_faces(T, ArgNames, Options).
 wiki_faces_int([Before,EmphTerm|T], ArgNames, Options) -->
     emphasis_before(Before),
@@ -899,6 +971,16 @@ wiki_face(code(Code), _, _) -->
 wiki_face(code(Code), _, _) -->
     [=,'|'], wiki_words(Code), ['|',=],
     !.
+wiki_face(PredRef, _, _) -->
+    ['`'], take_predref(PredRef), ['`'],
+    !.
+wiki_face(\nopredref(Pred), _, _) -->
+    ['`', '`'], take_predref(\predref(Pred)), ['`', '`'],
+    !.
+wiki_face([flag, ' ', \flagref(Flag)], _, _) -->
+    [ w('flag'), ' ', '`', w(Flag), '`' ],
+    { current_prolog_flag(Flag, _) },
+    !.
 wiki_face(code(Code), _, _) -->
     ['`','`'], wiki_words(Code), ['`','`'],
     !.
@@ -911,12 +993,12 @@ wiki_face(Code, _, _) -->
           code_face(Text, Term, Vars, Code)
         }
     ).
-wiki_face(Face, _, _) -->
+wiki_face(Face, _, Options) -->
     [ w(Name) ], arg_list(List),
     { atomic_list_concat([Name|List], Text),
       E = error(_,_),
       catch(atom_to_term(Text, Term, Vars), E, fail),
-      term_face(Text, Term, Vars, Face)
+      term_face(Text, Term, Vars, Face, Options)
     },
     !.
 wiki_face(br([]), _, _) -->
@@ -929,32 +1011,11 @@ wiki_face(_, _, Options) -->
       !,
       fail
     }.
-wiki_face(\predref(Name/Arity), _, _) -->
-    [ w(Name), '/' ], arity(Arity),
-    { functor_name(Name)
-    },
-    !.
-wiki_face(\predref(Module:(Name/Arity)), _, _) -->
-    [ w(Module), ':', w(Name), '/' ], arity(Arity),
-    { functor_name(Name)
-    },
-    !.
-wiki_face(\predref(Name/Arity), _, _) -->
-    prolog_symbol_char(S0),
-    symbol_string(SRest), [ '/' ], arity(Arity),
-    !,
-    { atom_chars(Name, [S0|SRest])
-    }.
-wiki_face(\predref(Name//Arity), _, _) -->
-    [ w(Name), '/', '/' ], arity(Arity),
-    { functor_name(Name)
-    },
-    !.
-wiki_face(\predref(Module:(Name//Arity)), _, _) -->
-    [ w(Module), ':', w(Name), '/', '/' ], arity(Arity),
-    { functor_name(Name)
-    },
+wiki_face(PredRef, _, _) -->
+    take_predref(PredRef),
     !.
+wiki_face(\cite(Citations), _, _) -->
+    ['['], citations(Citations), [']'].
 wiki_face(\include(Name, Type, Options), _, _) -->
     ['[','['], file_name(Base, Ext), [']',']'],
     { autolink_extension(Ext, Type),
@@ -1194,21 +1255,21 @@ arg_list_close([H|T], Depth) -->
     arg_list_close(T, Depth).
 
 
-%!  term_face(+Text, +Term, +Vars, -Face) is semidet.
+%!  term_face(+Text, +Term, +Vars, -Face, +Options) is semidet.
 %
 %   Process embedded Prolog-terms. Currently   processes  Alias(Arg)
 %   terms that refer to files.  Future   versions  will also provide
 %   pretty-printing of Prolog terms.
 
-term_face(_Text, Term, _Vars, \file(Name, FileOptions)) :-
+term_face(_Text, Term, _Vars, \file(Name, FileOptions), Options) :-
     ground(Term),
     compound(Term),
-    compound_name_arguments(Term, Alias, [_]),
+    compound_name_arity(Term, Alias, 1),
     user:file_search_path(Alias, _),
-    existing_file(Term, FileOptions, []),
+    existing_file(Term, FileOptions, [], Options),
     !,
     format(atom(Name), '~q', [Term]).
-term_face(Text, Term, Vars, Face) :-
+term_face(Text, Term, Vars, Face, _Options) :-
     code_face(Text, Term, Vars, Face).
 
 untag([], []).
@@ -1231,6 +1292,29 @@ image_label(\include(Name, image, Options)) -->
     file_options(RestOptions).
 
 
+take_predref(\predref(Name/Arity)) -->
+    [ w(Name), '/' ], arity(Arity),
+    { functor_name(Name)
+    }.
+take_predref(\predref(Module:(Name/Arity))) -->
+    [ w(Module), ':', w(Name), '/' ], arity(Arity),
+    { functor_name(Name)
+    }.
+take_predref(\predref(Name/Arity)) -->
+    prolog_symbol_char(S0),
+    symbol_string(SRest), [ '/' ], arity(Arity),
+    !,
+    { atom_chars(Name, [S0|SRest])
+    }.
+take_predref(\predref(Name//Arity)) -->
+    [ w(Name), '/', '/' ], arity(Arity),
+    { functor_name(Name)
+    }.
+take_predref(\predref(Module:(Name//Arity))) -->
+    [ w(Module), ':', w(Name), '/', '/' ], arity(Arity),
+    { functor_name(Name)
+    }.
+
 %!  file_options(-Options) is det.
 %
 %   Extracts additional processing options for  files. The format is
@@ -1323,6 +1407,15 @@ wiki_link(a(href(Ref), Label), Options) -->
       catch(expand_url_path(Term, Ref), E, fail),
       option(label(Label), Options, Ref)
     }.
+wiki_link(a(href(Ref), Label), Options) -->
+    [#, w(First)],
+    { option(end(End), Options) },
+    tokens_no_whitespace(Rest),
+    peek_end_url(End),
+    !,
+    { atomic_list_concat([#,First|Rest], Ref),
+      option(label(Label), Options, Ref)
+    }.
 wiki_link(a(href(Ref), Label), Options) -->
     [<],
     (   { option(relative(true), Options),
@@ -1369,7 +1462,8 @@ file_name(FileBase, Extension) -->
       absolute_file_name(Spec, Path,
                          [ access(read),
                            extensions([pl]),
-                           file_type(prolog)
+                           file_type(prolog),
+                           file_errors(fail)
                          ]),
       file_name_extension(FileBase, Extension, Path)
     }.
@@ -1398,37 +1492,40 @@ file_extension(Ext) -->
     }.
 
 
-%!  resolve_file(+Name, -Options, ?RestOptions) is det.
+%!  resolve_file(+Name, -FileOptions, ?RestOptions, +Options) is det.
 %
 %   Find the actual file based on the pldoc_file global variable. If
 %   present  and  the   file   is    resolvable,   add   an   option
 %   absolute_path(Path) that reflects the current   location  of the
 %   file.
 
-resolve_file(Name, Options, Rest) :-
-    existing_file(Name, Options, Rest),
+resolve_file(Name, FileOptions, Rest) :-
+    existing_file(Name, FileOptions, Rest, []),
     !.
 resolve_file(_, Options, Options).
 
 
-existing_file(Name, Options, Rest) :-
+existing_file(Name, FileOptions, Rest, Options) :-
+    \+ Options.get(link) == false,
     E = error(_,_),
-    catch(existing_file_p(Name, Options, Rest), E, fail).
+    catch(existing_file_p(Name, FileOptions, Rest), E, fail).
 
-existing_file_p(Name, Options, Rest) :-
-    nb_current(pldoc_file, RelativeTo),
-    RelativeTo \== [],
+existing_file_p(Name, FileOptions, Rest) :-
+    (   nb_current(pldoc_file, RelativeTo),
+        RelativeTo \== []
+    ->  Extra = [relative_to(RelativeTo)|Extra1]
+    ;   Extra = Extra1
+    ),
     (   compound(Name)
-    ->  Extra = [file_type(prolog)]
-    ;   Extra = []
+    ->  Extra1 = [file_type(prolog)]
+    ;   Extra1 = []
     ),
     absolute_file_name(Name, Path,
-                       [ relative_to(RelativeTo),
-                         access(read),
+                       [ access(read),
                          file_errors(fail)
                        | Extra
                        ]),
-    Options = [ absolute_path(Path) | Rest ].
+    FileOptions = [ absolute_path(Path) | Rest ].
 
 %!  arity(-Arity:int)// is semidet.
 %
@@ -1536,6 +1633,34 @@ autolink_file('README', wiki).
 autolink_file('TODO', wiki).
 autolink_file('ChangeLog', wiki).
 
+%!  citations(-List)//
+%
+%   Parse @cite1[;@cite2]* into a list of citations.
+
+citations([H|T]) -->
+    citation(H),
+    (   [';']
+    ->  citations(T)
+    ;   {T=[]}
+    ).
+
+citation(Atom) -->
+    [@], wiki_words(Atoms),
+    { length(Atoms, Len),
+      Len > 10, !,
+      fail
+    ; true
+    },
+    end_citation,
+    !,
+    { atomic_list_concat(Atoms, Atom)
+    }.
+
+end_citation, [';'] --> [';'].
+end_citation, ['@'] --> ['@'].
+end_citation, [']'] --> [']'].
+
+
                  /*******************************
                  *           SECTIONS           *
                  *******************************/
