@@ -3,22 +3,34 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2002-2011, University of Amsterdam
-			      VU University Amsterdam
+    Copyright (c)  2003-2016, University of Amsterdam
+                              VU University Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #ifndef RDFDB_H_INCLUDED
@@ -64,18 +76,17 @@ supposed to be local to the SWI-Prolog kernel are declared using
 #include "debug.h"
 #include "memory.h"
 #include "hash.h"
-#include "error.h"
 #include "skiplist.h"
 #ifdef WITH_MD5
 #include "md5.h"
 #endif
 #include "mutex.h"
 #include "resource.h"
+#include "xsd.h"
 
 #define RDF_VERSION 30000		/* 3.0.0 */
 
-#define URL_subPropertyOf \
-	"http://www.w3.org/2000/01/rdf-schema#subPropertyOf"
+#define URL_subPropertyOf "http://www.w3.org/2000/01/rdf-schema#subPropertyOf"
 
 
 		 /*******************************
@@ -431,6 +442,20 @@ typedef struct query_admin
 } query_admin;
 
 
+#define PREFIX_INITIAL_ENTRIES 16
+
+typedef struct prefix
+{ atom_t	 alias;
+  atom_info      uri;
+  struct prefix *next;
+} prefix;
+
+typedef struct prefix_table
+{ prefix       **entries;
+  size_t	 size;
+  size_t	 count;
+} prefix_table;
+
 #define JOINED_DEFER 1
 
 #ifdef JOINED_DEFER
@@ -455,6 +480,7 @@ typedef struct rdf_db
   size_t	agenda_created;		/* #visited nodes in agenda */
   graph_hash_table graphs;		/* Graph table */
   graph	       *last_graph;		/* last accessed graph */
+  prefix_table *prefixes;		/* alias->uri table */
   query_admin	queries;		/* Active query administration */
 
   size_t	duplicates;		/* #duplicate triples */
@@ -490,6 +516,7 @@ typedef struct rdf_db
     simpleMutex gc;			/* DB garbage collection lock */
     simpleMutex duplicates;		/* Duplicate init lock */
     simpleMutex erase;			/* Erase triples */
+    simpleMutex prefixes;		/* Prefix expansion */
   } locks;
 
   struct
@@ -627,5 +654,8 @@ COMMON(int)	rdf_broadcast(broadcast_id id, void *a1, void *a2);
 COMMON(int)	rdf_is_broadcasting(broadcast_id id);
 COMMON(void)	consider_triple_rehash(rdf_db *db, size_t extra);
 COMMON(int)	rdf_create_gc_thread(rdf_db *db);
+COMMON(atom_t)  expand_prefix(rdf_db *db, atom_t alias, atom_t local);
+
+#include "error.h"
 
 #endif /*RDFDB_H_INCLUDED*/
