@@ -165,7 +165,7 @@ typedef enum format
 { D_AUTO = 0,
   D_TURTLE,
   D_TRIG,
-  D_TRIG_NO_GRAPH			/* read TRiG as Turtle */
+  D_TRIG_NO_GRAPH			/* read TriG as Turtle */
 } format;
 
 typedef enum recover_mode
@@ -1139,7 +1139,7 @@ make_absolute_resource(turtle_state *ts, const wchar_t *uri)
 		 *	   PARSER OBJECT	*
 		 *******************************/
 
-static void	clear_turtle_parser(turtle_state *ts);
+static int	clear_turtle_parser(turtle_state *ts);
 static int	init_base_uri(turtle_state *ts);
 static int	read_predicate_object_list(turtle_state *ts, const char *end);
 static int	read_object(turtle_state *ts);
@@ -1170,13 +1170,18 @@ new_turtle_parser(IOSTREAM *s)
 }
 
 
-static void
+static int
 clear_turtle_parser(turtle_state *ts)
-{ if ( ts->base_uri )	     free(ts->base_uri);
+{ int rc;
+
+  if ( ts->base_uri )	     free(ts->base_uri);
   if ( ts->empty_prefix )    free(ts->empty_prefix);
   if ( ts->bnode.buffer )    free(ts->bnode.buffer);
 
-  if ( ts->input ) PL_release_stream(ts->input);
+  if ( ts->input )
+    rc = PL_release_stream(ts->input);
+  else
+    rc = TRUE;
 
   set_subject(ts, NULL, NULL);
   set_predicate(ts, NULL, NULL);
@@ -1189,6 +1194,8 @@ clear_turtle_parser(turtle_state *ts)
   clear_hash_table(&ts->blank_node_map);
 
   memset(ts, 0, sizeof(*ts));			/* second call does nothing */
+
+  return rc;
 }
 
 
@@ -2393,6 +2400,7 @@ read_object(turtle_state *ts)
 	free_resource(ts, r);
 	return rc;
       }
+      return FALSE;
     }
     case '[':
     { resource *r;
@@ -2753,12 +2761,22 @@ starts_graph(turtle_state *ts)
 }
 
 
+#define	NO_GRAPH \
+	if ( graph_keyword_seen ) \
+	  return syntax_error(ts, "graph IRI expected after GRAPH keyword") \
+
 static int
-graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
+graph_or_final_predicate_object_list(turtle_state *ts, resource *r,
+				     int *must_be_graph)
 { int rc;
+  int graph_keyword_seen = *must_be_graph;
+
+  *must_be_graph = FALSE;
 
   if ( !skip_ws(ts) )
+  { NO_GRAPH;
     return FALSE;
+  }
 
   if ( (rc = starts_graph(ts)) == TRUE )
   { switch ( ts->format )
@@ -2776,7 +2794,7 @@ graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
 	}
       case D_TURTLE:
 	syntax_warning(ts, "Unexpected \"<graph> {\" in Turtle format "
-		       "(assuming TRiG, ignoring graphs)");
+		       "(assuming TriG, ignoring graphs)");
         set_format(ts, D_TRIG_NO_GRAPH);
 	/*FALLTHROUGH*/
       case D_TRIG_NO_GRAPH:
@@ -2789,6 +2807,7 @@ graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
   } else if ( rc == FALSE )
   { set_subject(ts, r, NULL);
 
+    NO_GRAPH;
     return final_predicate_object_list(ts);
   } else
     return FALSE;			/* error from starts_graph() */
@@ -2797,17 +2816,20 @@ graph_or_final_predicate_object_list(turtle_state *ts, resource *r)
 
 static int
 statement(turtle_state *ts)
-{ retry:
+{ int graph_keyword_seen = FALSE;
 
+retry:
   if ( !skip_ws(ts) )
     return FALSE;
 
   switch ( ts->current_char )
   { case '@':				/* directive */
-    { return next(ts) && directive(ts);
+    { NO_GRAPH;
+      return next(ts) && directive(ts);
     }
     case '}':
-    { if ( ts->format == D_TRIG )
+    { NO_GRAPH;
+      if ( ts->format == D_TRIG )
       { set_graph(ts, NULL, NULL);
 	return next(ts);			/* return } as empty triple set */
       } else if ( ts->format == D_TRIG_NO_GRAPH )
@@ -2816,7 +2838,8 @@ statement(turtle_state *ts)
 	return syntax_error(ts, "Unexpected \"}\" in Turtle format");
     }
     case '{':
-    { switch( ts->format )
+    { NO_GRAPH;
+      switch( ts->format )
       { case D_AUTO:
 	  set_format(ts, D_TRIG);
 	  /*FALLTHROUGH*/
@@ -2833,7 +2856,7 @@ statement(turtle_state *ts)
 	  }
 	case D_TURTLE:
 	  syntax_warning(ts, "Unexpected \"{\" in Turtle format "
-			 "(assuming TRiG, ignoring graphs)");
+			 "(assuming TriG, ignoring graphs)");
 	  set_format(ts, D_TRIG_NO_GRAPH);
 	  /*FALLTHROUGH*/
         case D_TRIG_NO_GRAPH:
@@ -2846,12 +2869,13 @@ statement(turtle_state *ts)
     { resource *r;
 
       if ( (r=read_iri_ref(ts)) )
-	return graph_or_final_predicate_object_list(ts, r);
+	return graph_or_final_predicate_object_list(ts, r, &graph_keyword_seen);
       return FALSE;
     }
     case '[':
     { resource *r;
 
+      NO_GRAPH;
       if ( !next(ts) || !skip_ws(ts) )
 	return FALSE;
       if ( ts->current_char == ']' )
@@ -2871,6 +2895,7 @@ statement(turtle_state *ts)
     case '(':
     { resource *r;
 
+      NO_GRAPH;
       if ( (r=read_collection(ts)) )
       { set_subject(ts, r, NULL);
 	return final_predicate_object_list(ts);
@@ -2880,6 +2905,7 @@ statement(turtle_state *ts)
     case '_':
     { resource *r;
 
+      NO_GRAPH;
       if ( (r=read_blank_node_label(ts)) )
       { set_subject(ts, r, NULL);
 	return final_predicate_object_list(ts);
@@ -2899,19 +2925,21 @@ statement(turtle_state *ts)
 	  r = resolve_iri(ts, NULL, baseBuf(&pn_local));
 	  discardBuf(&pn_local);
 	  if ( r )
-	    return graph_or_final_predicate_object_list(ts, r);
+	    return graph_or_final_predicate_object_list(ts, r,
+							&graph_keyword_seen);
 	}
       } else
       { resource *r;
 
 	if ( (r=resolve_iri(ts, NULL, NULL)) )
-	  return graph_or_final_predicate_object_list(ts, r);
+	  return graph_or_final_predicate_object_list(ts, r, &graph_keyword_seen);
       }
 
       return FALSE;
     }
     case -1:
-    { return TRUE;				/* End of file */
+    { NO_GRAPH;
+      return TRUE;				/* End of file */
     }
     default:
     { string_buffer pn_prefix;
@@ -2919,12 +2947,31 @@ statement(turtle_state *ts)
       if ( read_pn_prefix(ts, &pn_prefix) )
       { if ( wcscasecmp(baseBuf(&pn_prefix), L"BASE") == 0 )
 	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
 	  return sparql_base_directive(ts);
 	}
 	if ( wcscasecmp(baseBuf(&pn_prefix), L"PREFIX") == 0 )
 	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
 	  return sparql_prefix_directive(ts);
 	}
+	if ( wcscasecmp(baseBuf(&pn_prefix), L"GRAPH") == 0 &&
+	     ts->current_char != ':' )
+	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
+	  if ( ts->format == D_TURTLE )
+	  { syntax_warning(ts, "Unexpected \"GRAPH\" in Turtle format "
+			 "(assuming TriG, ignoring graphs)");
+	    set_format(ts, D_TRIG_NO_GRAPH);
+	  } else if ( ts->format == D_AUTO )
+	  { set_format(ts, D_TRIG);
+	  }
+	  if ( !next(ts) )
+	    return FALSE;
+	  graph_keyword_seen = TRUE;
+
+	  goto retry;
+	}
 
 	if ( ts->current_char == ':' )
 	{ if ( !next(ts) )
@@ -2942,7 +2989,8 @@ statement(turtle_state *ts)
 	      discardBuf(&pn_local);
 	      if ( r )
 	      { discardBuf(&pn_prefix);
-		return graph_or_final_predicate_object_list(ts, r);
+		return graph_or_final_predicate_object_list(ts, r,
+							    &graph_keyword_seen);
 	      }
 	    }
 	  } else
@@ -2952,10 +3000,12 @@ statement(turtle_state *ts)
 	    discardBuf(&pn_prefix);
 
 	    if ( r )
-	      return graph_or_final_predicate_object_list(ts, r);
+	      return graph_or_final_predicate_object_list(ts, r,
+							  &graph_keyword_seen);
 	  }
 	} else
 	{ discardBuf(&pn_prefix);
+	  NO_GRAPH;
 	  return syntax_error(ts, "Expected \":\"");
 	}
       }
@@ -3189,10 +3239,7 @@ destroy_turtle_parser(term_t parser)
 { turtle_state *ts;
 
   if ( get_turtle_parser(parser, &ts) )
-  { clear_turtle_parser(ts);
-
-    return TRUE;
-  }
+    return clear_turtle_parser(ts);
 
   return FALSE;
 }
