@@ -104,11 +104,11 @@ typedef enum
   S_DECL,				/* inside a declaration */
   S_MDECL0,				/* Seen <! */
   S_STRING,				/* inside a "string" or 'string' */
+  S_VAL0,				/* Seen = in decl */
   S_DECLCMT0,				/* Seen <...- */
   S_DECLCMT,				/* Seen <...-- */
   S_DECLCMTE0,				/* Seen <...--..- */
   S_CMTO,				/* Seen <!- */
-  S_CMT1,				/* Seen <!-- */
   S_CMT,				/* Seen <!--X... */
   S_CMTE0,				/* Seem <!--...- */
   S_CMTE1,				/* Seem <!--...-- */
