@@ -90,13 +90,12 @@ prolog_trace_interception_gui(Port, Frame, CHP, Action) :-
         Action = continue
     ).
 
-:- multifile
-    user:prolog_event_hook/1.
-
-user:prolog_event_hook(frame_finished(Frame)) :-
-    retractall(finished_frame(Frame)),
-    fail.
+:- initialization
+    prolog_unlisten(frame_finished, retract_frame),
+    prolog_listen(frame_finished, retract_frame).
 
+retract_frame(Frame) :-
+    retractall(finished_frame(Frame)).
 
 %!  map_action(+GuiAction, +Frame, -Action) is det.
 %
@@ -287,7 +286,7 @@ hide_children_frame(Frame) :-
 show(StartFrame, CHP, Up, exception(Except)) :-
     !,
     show(StartFrame, CHP, Up, exception, exception),
-    message_to_string(Except, Message),
+    message_to_string(Except, Message, 200),
     send_tracer(report(warning, 'Exception: %s', Message)).
 show(StartFrame, CHP, Up, Port) :-
     show(StartFrame, CHP, Up, Port, Port),
@@ -314,6 +313,16 @@ show(StartFrame, CHP, Up, Port, Style) :-
                         bindings
                       ]).
 
+message_to_string(Except, Message, MaxLength) :-
+    catch(message_to_string(Except, Message0), _, fail),
+    string_length(Message0, Len),
+    (   Len > MaxLength
+    ->  sub_string(Message0, 0, MaxLength, _, Base),
+        string_concat(Base, ' ...', Message)
+    ;   Message = Message0
+    ).
+message_to_string(_, Message, _) :-
+    Message = "<exception too large>".
 
 %!  find_frame(+Up, +StartFrame, +Port, -PC, -Frame) is det.
 %
