@@ -48,8 +48,8 @@
 #define OS_VERSION "unknown"
 #endif
 
-#ifndef MACHINE
-#define MACHINE "unknown"
+#ifndef PCE_MACHINE
+#define PCE_MACHINE "unknown"
 #endif
 
 
@@ -396,7 +396,7 @@ __pce_export int pceSetProfileHooks(pce_profile_hooks *hooks);
 		*         INITIALISATION	*
 		********************************/
 
-__pce_export int pceInitialise(int handles, const char *home,
+__pce_export int pceInitialise(int handles, const char *home, const char *appdata,
 			       int argc, char **argv);
 
 
@@ -463,7 +463,7 @@ __pce_export void	Cflush(void);
 __pce_export char *	Cgetline(char *line, int size);
 
 #ifdef PCE_INCLUDED
-	     int	Cputstr(String s);
+	     int	Cputstr(PceString s);
 #endif
 
 
