@@ -1,31 +1,41 @@
-/*  $Id$
-
-    Part of XPCE --- The SWI-Prolog GUI toolkit
+/*  Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
     E-mail:        jan@swi.psy.uva.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (c)  1985-2002, University of Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <h/kernel.h>
 #include <h/unix.h>			/* storeStringFile() prototype */
 
-static CharArray	stringToCharArray(String s);
+static CharArray	stringToCharArray(PceString s);
 
 		/********************************
 		*         CREATE/CONVERT	*
@@ -177,7 +187,7 @@ getModifyCharArray(CharArray n, CharArray n2)
 
 
 static CharArray
-ModifiedCharArray(CharArray n, String buf)
+ModifiedCharArray(CharArray n, PceString buf)
 {
   Class class = classOfObject(n);
 
@@ -210,7 +220,7 @@ getCapitaliseCharArray(CharArray n)
 { if ( n->data.s_size == 0 )
     answer(n);
   else
-  { String d = &n->data;
+  { PceString d = &n->data;
     int size = d->s_size;
     LocalString(buf, d->s_iswide, size);
     int i=1, o=1;
@@ -242,7 +252,7 @@ by spaces.
 
 CharArray
 getLabelNameCharArray(CharArray n)
-{ String s = &n->data;
+{ PceString s = &n->data;
   int size = s->s_size;
   int i;
 
@@ -274,7 +284,7 @@ getLabelNameCharArray(CharArray n)
 
 CharArray
 getDowncaseCharArray(CharArray n)
-{ String s = &n->data;
+{ PceString s = &n->data;
   int size = s->s_size;
   LocalString(buf, s->s_iswide, size);
   int i;
@@ -289,7 +299,7 @@ getDowncaseCharArray(CharArray n)
 
 static CharArray
 getUpcaseCharArray(CharArray n)
-{ String s = &n->data;
+{ PceString s = &n->data;
   int size = s->s_size;
   LocalString(buf, s->s_iswide, size);
   int i;
@@ -304,7 +314,7 @@ getUpcaseCharArray(CharArray n)
 
 static CharArray
 getStripCharArray(CharArray n, Name how)
-{ String s = &n->data;
+{ PceString s = &n->data;
   int size = s->s_size;
   LocalString(buf, s->s_iswide, size);
   int i=0, o=0, lnb=0;
@@ -338,7 +348,7 @@ getStripCharArray(CharArray n, Name how)
 
 static Chain
 getSplitCharArray(CharArray in, CharArray br)
-{ String s1 = &in->data;
+{ PceString s1 = &in->data;
   int size = s1->s_size;
   int i=0, last=0;
   Chain ch = answerObject(ClassChain, EAV);
@@ -347,7 +357,7 @@ getSplitCharArray(CharArray in, CharArray br)
   str_cphdr(&buf, s1);
 
   if ( notDefault(br) )			/* given pattern */
-  { String b = &br->data;
+  { PceString b = &br->data;
 
     while( i<=size-b->s_size )
     { if ( str_prefix_offset(s1, i, b) )
@@ -402,8 +412,8 @@ getSplitCharArray(CharArray in, CharArray br)
 
 CharArray
 getAppendCharArray(CharArray n1, CharArray n2)
-{ String s1 = &n1->data;
-  String s2 = &n2->data;
+{ PceString s1 = &n1->data;
+  PceString s2 = &n2->data;
   int iswide = (s1->s_iswide || s2->s_iswide);
   LocalString(buf, iswide, s1->s_size + s2->s_size);
 
@@ -541,7 +551,7 @@ base64_code(unsigned int in)
 
 static CharArray
 getBase64EncodeCharArray(CharArray in)
-{ String s = &in->data;
+{ PceString s = &in->data;
   int size = s->s_size;
   int triples = (size+2)/3;
   LocalString(buf, FALSE, triples*4);
@@ -578,7 +588,7 @@ getBase64EncodeCharArray(CharArray in)
 
 static CharArray
 getBase64DecodeCharArray(CharArray in)
-{ String s = &in->data;
+{ PceString s = &in->data;
   int size = s->s_size;
   LocalString(buf, FALSE, (size/4)*3);
   int i, o = 0;
@@ -793,7 +803,7 @@ CtoScratchCharArray(const char *s)
 
 
 CharArray
-StringToScratchCharArray(const String s)
+StringToScratchCharArray(const PceString s)
 { CharArray name = scratch_char_arrays;
   int n;
 
@@ -828,7 +838,7 @@ CtoCharArray(char *s)
 
 
 static CharArray
-stringToCharArray(String s)
+stringToCharArray(PceString s)
 { CharArray name = StringToScratchCharArray(s);
   CharArray rval = answerObject(ClassCharArray, name, EAV);
 
