@@ -47,6 +47,8 @@ static void	expose_frame(Widget w, FrameObj fr, Region xregion);
 static void	destroyFrame(Widget, FrameObj, XtPointer);
 static status   updateAreaFrame(FrameObj fr, Int border);
 static int	ws_group_frame(FrameObj fr);
+static int	ws_set_pid_frame(FrameObj fr);
+static void	ws_set_net_icon_frame(FrameObj fr);
 
 #define MainWindow(fr)	     ( isNil(fr->members->head) ? (Any) fr : \
 			       fr->members->head->value )
@@ -147,6 +149,15 @@ ws_uncreate_frame(FrameObj fr)
 }
 
 
+static Image
+getIconFrame(FrameObj fr)
+{ if ( notNil(fr->application) && notNil(fr->application->icon_image) )
+    return fr->application->icon_image;
+
+  return fr->icon_image;
+}
+
+
 status
 ws_create_frame(FrameObj fr)
 { Arg args[25];
@@ -154,6 +165,7 @@ ws_create_frame(FrameObj fr)
   Widget w;
   DisplayObj d = fr->display;
   DisplayWsXref r = d->ws_ref;
+  Image icon;
 
   XtSetArg(args[n], XtNtitle,		  nameToMB(fr->label));   n++;
   XtSetArg(args[n], XtNmappedWhenManaged, False);                 n++;
@@ -177,13 +189,13 @@ ws_create_frame(FrameObj fr)
   { XtSetArg(args[n], XtNsaveUnder, True);
     n++;
   }
-  if ( notNil(fr->icon_image) )
+  if ( notNil(icon=getIconFrame(fr)) )
   { XtSetArg(args[n], XtNiconPixmap,
-	     getXrefObject(fr->icon_image, fr->display));
+	     getXrefObject(icon, fr->display));
     n++;
-    if ( notNil(fr->icon_image->mask) )
+    if ( notNil(icon->mask) )
     { XtSetArg(args[n], XtNiconMask,
-	       getXrefObject(fr->icon_image->mask, fr->display));
+	       getXrefObject(icon->mask, fr->display));
       n++;
     }
   }
@@ -263,6 +275,10 @@ ws_realise_frame(FrameObj fr)
 
   ws_frame_background(fr, fr->background); /* Why is this necessary? */
   ws_group_frame(fr);
+  if ( notNil(fr->application) && notNil(fr->application->icon_image) )
+    ws_set_pid_frame(fr);		   /* group in Ubuntu dock */
+
+  ws_set_net_icon_frame(fr);
 }
 
 
@@ -411,6 +427,28 @@ ws_group_frame(FrameObj fr)
   fail;
 }
 
+static int
+ws_set_pid_frame(FrameObj fr)
+{ Widget w = widgetFrame(fr);
+  DisplayWsXref r = fr->display->ws_ref;
+  unsigned int xpid = getpid();
+  static Atom _net_wm_pid = 0;
+
+  if ( !_net_wm_pid )
+    _net_wm_pid = XInternAtom(r->display_xref,
+			      "_NET_WM_PID",
+			      False);
+
+  XChangeProperty(r->display_xref,
+		  XtWindow(w),
+		  _net_wm_pid,
+		  XA_CARDINAL, 32, PropModeReplace,
+		  (unsigned char *) &xpid, 1);
+
+  return TRUE;
+}
+
+
 #ifdef O_XDND
 		 /*******************************
 		 *	   DRAG-AND-DROP	*
@@ -1003,6 +1041,10 @@ getWMFrameFrame(FrameObj fr, int *dxp, int *dyp)
 		Cprintf("w = %ld; root = %ld; parent = %ld; "
 			"dx=%d; dy=%d; bw = %d\n",
 			w, root, parent, dx, dy, bw));
+	} else
+	{ DEBUG(NAME_frame,
+		Cprintf("w = %ld; root = %ld; parent = %ld\n",
+			w, root, parent));
 	}
 
 	if ( parent == root )
@@ -1037,6 +1079,10 @@ ws_frame_bb(FrameObj fr, int *x, int *y, int *w, int *h)
     *w = atts.width + 2*bw;
     *h = atts.height + 2*bw;
 
+    DEBUG(NAME_geometry,
+	  Cprintf("ws_frame_bb(%s): %d %d %d %d\n",
+		  pp(fr), *x, *y, *w, *h));
+
     succeed;
   }
 
@@ -1148,15 +1194,13 @@ ws_x_geometry_frame(FrameObj fr, Name spec, Monitor mon)
     }
 
     if ( ok )
-    { int mw = (w < MIN_VISIBLE ? MIN_VISIBLE : w);
-
-      if ( y < 0 )			/* above the screen */
-	y = 0;
-      if ( y > dh-MIN_VISIBLE )		/* below the screen */
+    { if ( y < 1 )			/* above the screen */
+	y = 1;
+      else if ( y > dh-MIN_VISIBLE )	/* below the screen */
 	y = dh - MIN_VISIBLE;
-      if ( x+mw < MIN_VISIBLE )		/* left of the screen */
-	x = MIN_VISIBLE-mw;
-      if ( x > dw-MIN_VISIBLE )		/* right of the screen */
+      if ( x < 1 )			/* left of the screen */
+	x = 1;
+      else if ( x > dw-MIN_VISIBLE )	/* right of the screen */
 	x = dw - MIN_VISIBLE;
     }
 
@@ -1358,6 +1402,42 @@ ws_set_icon_frame(FrameObj fr)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+See https://stackoverflow.com/questions/10699927/xlib-argb-window-icon
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void
+ws_set_net_icon_frame(FrameObj fr)
+{ unsigned long *buffer;
+  size_t length;
+  Image icon = getIconFrame(fr);
+
+  if ( notNil(icon) &&
+       getXrefObject(icon, fr->display) &&
+       (buffer=ws_image_to_rgba(icon, DEFAULT, &length)) )
+  { Widget w = widgetFrame(fr);
+    DisplayWsXref r = fr->display->ws_ref;
+    static Atom _net_wm_icon = 0;
+    static Atom cardinal     = 0;
+
+    if ( !_net_wm_icon ) {
+      _net_wm_icon = XInternAtom(r->display_xref,
+				 "_NET_WM_ICON",
+				 False);
+      cardinal = XInternAtom(r->display_xref, "CARDINAL", False);
+    }
+
+    XChangeProperty(r->display_xref,
+		    XtWindow(w),
+		    _net_wm_icon,
+		    cardinal, 32,
+		    PropModeReplace, (const unsigned char*) buffer, length);
+
+    free(buffer);
+  }
+}
+
+
 void
 ws_set_icon_label_frame(FrameObj fr)
 { Widget w = widgetFrame(fr);
