@@ -237,7 +237,7 @@ pl_format_predicate(term_t chr, term_t descr)
   if ( !format_predicates )
     format_predicates = newHTable(8);
 
-  addNewHTable(format_predicates, (void *)(intptr_t)c, proc);
+  updateHTable(format_predicates, (void *)(intptr_t)c, proc);
 
   succeed;
 }
@@ -601,7 +601,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  } else			/* r,R */
 		  { if ( arg == DEFAULT )
 		      FMT_ERROR("r,R requires radix specifier");
-		    if ( arg < 1 || arg > 36 )
+		    if ( arg < 2 || arg > 36 )
 		    { term_t r = PL_new_term_ref();
 
 		      PL_put_integer(r, arg);
