@@ -35,7 +35,6 @@
 */
 
 #ifdef __WINDOWS__
-#define _WIN32_WINNT 0x0600		/* Vista */
 #include "windows/uxnt.h"
 #ifdef WIN64
 #include "config/win64.h"
@@ -72,7 +71,6 @@ locking is required.
 #define NEEDS_SWINSOCK
 #define PL_STREAM_IMPL
 #include "pl-stream.h"
-#define PL_ARITY_AS_SIZE
 #include "SWI-Prolog.h"
 #include "pl-utf8.h"
 #include <sys/types.h>
@@ -2975,11 +2973,54 @@ IOFUNCTIONS Sfilefunctions =
 };
 
 
+		 /*******************************
+		 *	    TTY STREAMS		*
+		 *******************************/
+
+static ssize_t
+Sread_tty(void *handle, char *buf, size_t size)
+{ ssize_t bytes = Sread_file(handle, buf, size);
+
+  if ( bytes < 0 && errno == EBADF )
+  { errno = 0;
+    return 0;
+  }
+
+  return bytes;
+}
+
+
+static ssize_t
+Swrite_tty(void *handle, char *buf, size_t size)
+{ ssize_t bytes = Swrite_file(handle, buf, size);
+
+  if ( bytes == -1 && errno == EBADF )
+  { errno = 0;
+    return size;
+  }
+
+  return bytes;
+}
+
+
+static int
+Sclose_tty(void *handle)
+{ int rc = Sclose_file(handle);
+
+  if ( rc == -1 && errno == EBADF )
+  { errno = 0;
+    return 0;
+  }
+
+  return rc;
+}
+
+
 IOFUNCTIONS Sttyfunctions =
-{ Sread_file,
-  Swrite_file,
-  NULL,
-  Sclose_file,
+{ Sread_tty,
+  Swrite_tty,
+  Sseek_file,
+  Sclose_tty,
   Scontrol_file,
 #ifdef O_LARGEFILES
   Sseek_file64
@@ -3593,12 +3634,35 @@ Sclose_memfile(void *handle)
   return -1;
 }
 
+static int
+Scontrol_memfile(void *handle, int action, void *arg)
+{ memfile *mf = handle;
+
+  switch(action)
+  { case SIO_SETENCODING:
+      errno = EPERM;
+      return -1;
+    case SIO_FLUSHOUTPUT:
+      return 0;
+    case SIO_GETSIZE:
+    { int64_t *szp = arg;
+
+      *szp = mf->here;
+      return 0;
+    }
+    default:
+      errno = EINVAL;
+      return -1;
+  }
+}
+
 
 IOFUNCTIONS Smemfunctions =
 { Sread_memfile,
   Swrite_memfile,
   Sseek_memfile,
-  Sclose_memfile
+  Sclose_memfile,
+  Scontrol_memfile
 };
 
 
