@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2012-2018, VU University Amsterdam
+    Copyright (c)  2012-2020, VU University Amsterdam
 			      CWI, Amsterdam
     All rights reserved.
 
@@ -77,12 +77,13 @@ COMMON(void) GC_linger(void *ptr);
 
 typedef struct linger_list
 { struct linger_list *next;		/* Next lingering object */
+  gen_t		generation;		/* Linger generation */
   void		*object;		/* The lingering data */
   void	       (*unalloc)(void* obj);   /* actually free the object */
 } linger_list;
 
 COMMON(void)	linger(linger_list** list, void (*unalloc)(void *), void *object);
-COMMON(void)	free_lingering(linger_list *list);
+COMMON(void)	free_lingering(linger_list **list, gen_t generation);
 
 
 		 /*******************************
@@ -90,6 +91,8 @@ COMMON(void)	free_lingering(linger_list *list);
 		 *******************************/
 
 COMMON(void)		initAlloc(void);
+COMMON(int)		initTCMalloc(void);
+COMMON(size_t)		heapUsed(void);
 #ifndef DMALLOC
 COMMON(void *)		allocHeap(size_t n);
 COMMON(void *)		allocHeapOrHalt(size_t n);
@@ -118,6 +121,16 @@ COMMON(int64_t)		valBignum__LD(word w ARG_LD);
 COMMON(int)		equalIndirect(word r1, word r2);
 COMMON(size_t)		gsizeIndirectFromCode(Code PC);
 COMMON(word)		globalIndirectFromCode(Code *PC);
+COMMON(void *)		tmp_malloc(size_t req);
+COMMON(void *)		tmp_realloc(void *mem, size_t req);
+COMMON(void)		tmp_free(void *mem);
+COMMON(size_t)		tmp_nalloc(size_t req);
+COMMON(size_t)		tmp_nrealloc(void *mem, size_t req);
+COMMON(void *)		stack_malloc(size_t req);
+COMMON(void *)		stack_realloc(void *mem, size_t req);
+COMMON(void)		stack_free(void *mem);
+COMMON(size_t)		stack_nalloc(size_t req);
+COMMON(size_t)		stack_nrealloc(void *mem, size_t req);
 #ifndef xmalloc
 COMMON(void *)		xmalloc(size_t size);
 COMMON(void *)		xrealloc(void *mem, size_t size);
