@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2017, University of Amsterdam
+    Copyright (c)  1985-2019, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -213,7 +213,7 @@ bump_atom_references(Atom a, unsigned int ref)
     if ( ATOM_REF_COUNT(nref) == 0 )
       return TRUE;			/* reached max references */
 
-    if ( COMPARE_AND_SWAP(&a->references, ref, nref) )
+    if ( COMPARE_AND_SWAP_UINT(&a->references, ref, nref) )
     { if ( ATOM_REF_COUNT(ref) == 0 )
 	ATOMIC_DEC(&GD->atoms.unregistered);
       return TRUE;
@@ -258,13 +258,15 @@ PL_register_blob_type(PL_blob_t *type)
 	  t = t->next;
 
 	t->next = type;
-	type->rank = t->rank+1;
       }
       if ( true(type, PL_BLOB_TEXT) )
-      { if ( true(type, PL_BLOB_WCHAR) )
+      { type->rank = ++GD->atoms.text_rank;
+	if ( true(type, PL_BLOB_WCHAR) )
 	  type->padding = sizeof(pl_wchar_t);
 	else
 	  type->padding = sizeof(char);
+      } else
+      { type->rank = --GD->atoms.nontext_rank;
       }
 
       if ( !GD->atoms.initialised )
@@ -389,7 +391,7 @@ static const ccharp atoms[] = {
 
 #define acquire_atom_bucket(b) (void)0
 
-#define release_atom_bucket(b) (void)0
+#define release_atom_bucket() (void)0
 
 #endif
 
@@ -420,7 +422,7 @@ allocateAtomBlock(int idx)
     { newblock[i].type = ATOM_TYPE_INVALID;
       newblock[i].name = "<virgin>";
     }
-    if ( !COMPARE_AND_SWAP(&GD->atoms.array.blocks[idx],
+    if ( !COMPARE_AND_SWAP_PTR(&GD->atoms.array.blocks[idx],
 			   NULL, newblock-bs) )
       PL_free(newblock);		/* done by someone else */
   }
@@ -451,7 +453,7 @@ reserveAtom(void)
       ref = a->references;
 
       if ( ATOM_IS_FREE(ref) &&
-	   COMPARE_AND_SWAP(&a->references, ref, ATOM_RESERVED_REFERENCE) )
+	   COMPARE_AND_SWAP_UINT(&a->references, ref, ATOM_RESERVED_REFERENCE) )
       { assert(a->type == ATOM_TYPE_INVALID);
 	GD->atoms.no_hole_before = index+1;
         a->atom = (index<<LMASK_BITS)|TAG_ATOM;
@@ -475,8 +477,8 @@ reserveAtom(void)
     ref = a->references;
 
     if ( ATOM_IS_FREE(ref) &&
-	 COMPARE_AND_SWAP(&a->references, ref, ATOM_RESERVED_REFERENCE) )
-    { COMPARE_AND_SWAP(&GD->atoms.highest, index, index+1);
+	 COMPARE_AND_SWAP_UINT(&a->references, ref, ATOM_RESERVED_REFERENCE) )
+    { COMPARE_AND_SWAP_SIZE(&GD->atoms.highest, index, index+1);
       a->atom = (index<<LMASK_BITS)|TAG_ATOM;
 
       return a;
@@ -513,7 +515,7 @@ is activated.
 static int
 same_name(const Atom a, const char *s, size_t length, const PL_blob_t *type)
 { if ( false(type, PL_BLOB_NOCOPY) )
-    return memcmp(s, a->name, length) == 0;
+    return length == 0 || memcmp(s, a->name, length) == 0;
   else
     return s == a->name;
 }
@@ -609,13 +611,13 @@ redo:
   if ( true(type, PL_BLOB_UNIQUE) )
   { a->next = table[v];
     if ( !( !GD->atoms.rehashing &&	/* See (**) above */
-            COMPARE_AND_SWAP(&table[v], head, a) &&
+            COMPARE_AND_SWAP_PTR(&table[v], head, a) &&
 	    table == GD->atoms.table->table ) )
     { if ( false(type, PL_BLOB_NOCOPY) )
         PL_free(a->name);
       a->type = ATOM_TYPE_INVALID;
       a->name = "<race>";
-      MemoryBarrier();
+      MEMORY_BARRIER();
       a->references = 0;
       goto redo;
     }
@@ -663,8 +665,8 @@ static char *t_tracking;
 
 static int
 tracking(const Atom a)
-{ return /*a->type == &text_atom && */
-         strprefix(a->name, t_tracking);
+{ return ( a->type == &text_atom &&
+	   strcmp(a->name, t_tracking) == 0 );
 }
 
 
@@ -679,7 +681,7 @@ _PL_debug_register_atom(atom_t a,
   { Atom atom = fetchAtomArray(i);
     unsigned int refs;
 
-    refs = register_atom(atom);
+    refs = ATOM_REF_COUNT(register_atom(atom));
     if ( atomLogFd && tracking(atom) )
       Sfprintf(atomLogFd, "%s:%d: %s(): ++ (%d) for `%s' (#%d)\n",
 	       file, line, func, refs, atom->name, i);
@@ -826,18 +828,18 @@ invalidateAtom(Atom a, unsigned int ref)
 #define ATOM_PRE_DESTROY_REFERENCE \
 	(ATOM_DESTROY_REFERENCE|ATOM_RESERVED_REFERENCE)
 
-  if ( !COMPARE_AND_SWAP(&a->references, ref, ATOM_PRE_DESTROY_REFERENCE) )
+  if ( !COMPARE_AND_SWAP_UINT(&a->references, ref, ATOM_PRE_DESTROY_REFERENCE) )
   { return FALSE;
   }
 
   if ( a->type->release )
   { if ( !(*a->type->release)(a->atom) )
-    { COMPARE_AND_SWAP(&a->references, ATOM_PRE_DESTROY_REFERENCE, ref);
+    { COMPARE_AND_SWAP_UINT(&a->references, ATOM_PRE_DESTROY_REFERENCE, ref);
       return FALSE;
     }
   } else if ( GD->atoms.gc_hook )
   { if ( !(*GD->atoms.gc_hook)(a->atom) )
-    { COMPARE_AND_SWAP(&a->references, ATOM_PRE_DESTROY_REFERENCE, ref);
+    { COMPARE_AND_SWAP_UINT(&a->references, ATOM_PRE_DESTROY_REFERENCE, ref);
       return FALSE;				/* foreign hooks says `no' */
     }
   }
@@ -859,7 +861,7 @@ invalidateAtom(Atom a, unsigned int ref)
     ap = &table->table[a->hash_value & mask];
 
     if ( *ap == a )
-    { if ( !COMPARE_AND_SWAP(&table->table[a->hash_value & mask], a, a->next) )
+    { if ( !COMPARE_AND_SWAP_PTR(&table->table[a->hash_value & mask], a, a->next) )
       { goto redo;
       }
     }
@@ -919,7 +921,7 @@ destroyAtom(Atom a, Atom **buckets)
 
   a->name = "<reclaimed>";
   a->type = ATOM_TYPE_INVALID;
-  MemoryBarrier();
+  MEMORY_BARRIER();
   a->references = 0;
 
   index = indexAtom(a->atom);
@@ -1019,7 +1021,7 @@ pl_garbage_collect_atoms(void)
   if ( GD->cleaning != CLN_NORMAL )	/* Cleaning up */
     return TRUE;
 
-  if ( !COMPARE_AND_SWAP(&GD->atoms.gc_active, FALSE, TRUE) )
+  if ( !COMPARE_AND_SWAP_INT(&GD->atoms.gc_active, FALSE, TRUE) )
     return TRUE;
 
   if ( verbose )
@@ -1108,7 +1110,7 @@ register_atom(volatile Atom p)
     unsigned int nref = ref+1;
 
     if ( ATOM_REF_COUNT(nref) != 0 )
-    { if ( COMPARE_AND_SWAP(&p->references, ref, nref) )
+    { if ( COMPARE_AND_SWAP_UINT(&p->references, ref, nref) )
       { if ( ATOM_REF_COUNT(nref) == 1 )
 	  ATOMIC_DEC(&GD->atoms.unregistered);
 	return nref;
@@ -1207,7 +1209,7 @@ unregister_atom(volatile Atom p)
 
       if ( ATOM_REF_COUNT(newref) == 0 )
 	newref |= ATOM_MARKED_REFERENCE;
-    } while( !COMPARE_AND_SWAP(&p->references, oldref, newref) );
+    } while( !COMPARE_AND_SWAP_UINT(&p->references, oldref, newref) );
     refs = ATOM_REF_COUNT(newref);
 #ifdef O_DEBUG_ATOMGC
     if ( refs == 0 && atomLogFd && tracking(p) )
@@ -1263,6 +1265,23 @@ PL_unregister_atom(atom_t a)
 }
 
 
+/* TRUE if `a` is a possible candidate for AGC
+*/
+int
+is_volatile_atom(atom_t a)
+{
+#ifdef O_ATOMGC
+  size_t index = indexAtom(a);
+
+  if ( index >= GD->atoms.builtin )
+  { Atom p = fetchAtomArray(index);
+    return !p->references;
+  }
+#endif
+  return FALSE;
+}
+
+
 		 /*******************************
 		 *	      CHECK		*
 		 *******************************/
@@ -1487,7 +1506,6 @@ registerBuiltinAtoms(void)
   size_t index;
   int idx;
 
-  GD->atoms.builtin_array = PL_malloc(size * sizeof(struct atom));
   GD->statistics.atoms = size;
 
   for( sp = atoms, index = GD->atoms.highest; *sp; sp++, index++ )
@@ -1698,7 +1716,8 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t state ARG_LD)
 	   (!type_name || type_name == btype->atom_name) &&
 	   atom->atom != ATOM_garbage_collected &&
 	   bump_atom_references(atom, refs) )
-      { DEBUG(CHK_SECURE,	/* avoid trap through linkVal__LD() check */
+      { DEBUG(CHK_ATOM_GARBAGE_COLLECTED,
+	      /* avoid trap through linkVal__LD() check */
 	      if ( atom->atom == ATOM_garbage_collected )
 	      { PL_unregister_atom(atom->atom);
 	        continue;
@@ -2155,6 +2174,37 @@ PL_atom_generator_w(const pl_wchar_t *prefix,
 }
 
 
+size_t
+atom_space(void)
+{ size_t array = ((size_t)2<<MSB(GD->atoms.highest))*sizeof(struct atom);
+  size_t index;
+  int i, last=FALSE;
+  size_t table = GD->atoms.table->buckets * sizeof(Atom);
+  size_t data = 0;
+
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    size_t high = GD->atoms.highest;
+    Atom b = GD->atoms.array.blocks[i];
+
+    if ( upto >= high )
+    { upto = high;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b + index;
+
+      if ( ATOM_IS_VALID(a->references) )
+      { data += a->length;		/* TBD: malloc rounding? */
+      }
+    }
+  }
+
+  return array+table+data;
+}
+
+
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
