@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2016-2017, VU University Amsterdam
+    Copyright (c)  2016-2020, VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -33,6 +34,7 @@
 */
 
 #include "pl-incl.h"
+#include "pl-comp.h"
 #include "pl-dbref.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -63,13 +65,23 @@ Where the compiler translates '$reset' into
     active non-Prolog slots.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static term_t
-findReset(LocalFrame fr, term_t ball ARG_LD)
+#define FRESET_NO_FRAME -1
+#define FRESET_FINDALL  -2
+
+static int
+findReset(LocalFrame fr, term_t ball, term_t *rframe ARG_LD)
 { Definition reset3  = PROCEDURE_reset3->definition;
 
   for(; fr; fr = fr->parent)
   { int rc;
     term_t tref;
+    static Procedure proc_fl = NULL;
+
+    if ( !proc_fl )
+      proc_fl = PL_predicate("findall_loop", 4, "$bags");
+
+    if ( fr->predicate == proc_fl->definition )
+      return FRESET_FINDALL;
 
     if ( fr->predicate != reset3 )
       continue;
@@ -79,11 +91,12 @@ findReset(LocalFrame fr, term_t ball ARG_LD)
     fr = (LocalFrame)valTermRef(tref);
 
     if ( rc )
-    { return consTermRef(fr);
+    { *rframe = consTermRef(fr);
+      return TRUE;
     }
   }
 
-  return 0;
+  return FRESET_NO_FRAME;
 }
 
 
@@ -325,7 +338,7 @@ Performs the following steps:
 	'$cont'(Clause, PC, EnvArg1, ...)
 
      Here, Clause is the clause, PC is the program counter inside
-     the clause, EnvArg1 is an array holding the frame argumnets
+     the clause, EnvArg1 is an array holding the frame arguments
      in the same order as the frame layout.  The atom '<inactive>'
      is used for frame slots that are not accessed by the remainder
      of the continuation.
@@ -339,8 +352,9 @@ Performs the following steps:
 Code
 shift(term_t ball ARG_LD)
 { term_t reset;
+  int rc;
 
-  if ( (reset=findReset(environment_frame, ball PASS_LD)) )
+  if ( (rc=findReset(environment_frame, ball, &reset PASS_LD)) == TRUE )
   { term_t cont = PL_new_term_ref();
     LocalFrame resetfr;
     LocalFrame fr;
@@ -384,11 +398,40 @@ shift(term_t ball ARG_LD)
     return resetfr->programPointer;
   }
 
-  PL_existence_error("reset/3", ball);
-  return NULL;
+  { static const char *msg[] =
+    { "No matching reset/3 call",
+      "Cannot catch continuation through findall/3"
+    };
+
+    return PL_error("shift", 1, msg[-1-rc],
+		    ERR_EXISTENCE, ATOM_reset, ball),NULL;
+  }
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Create a new stack frame to restart a continuation.
+
+(*) The continuation holds a clause reference   (blob)  to the clause to
+restart. This protects this clause from   being garbage collected. After
+we restarted the continuation though, it may become subject to the stack
+garbage collector and subsequently the blob  becomes subject to the atom
+garbage collector, both destroying the clause   reference and the clause
+if it was retracted.
+
+We prevent this by pushing the blob onto the environment stack, where it
+will be found by the atom  garbage   collector.  This  protects both the
+reference and the clause itself as long   as the continuation depends on
+it. Note that the atom is simply pushed  onto the stack where it resides
+between stack frames. This is fine as  atom-gc performs a linear scan of
+the environment stack. Also note that as we push this, it is above lTop.
+This is fine as in  this  stage  the   blob  is  still  protected by the
+continuation.
+
+A pushVolatileAtom() is not needed as this   protection only needs to be
+effective after the next GC call and GC won't run concurrently with AGC.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 Code
 push_continuation(term_t continuation, LocalFrame pfr, Code pcret ARG_LD)
 { LocalFrame top, fr;
@@ -404,21 +447,23 @@ retry:
   { Functor f = valueTerm(*cont);
     Word ep = f->arguments;
     Word ap;
-    Clause cl;
     ClauseRef cref;
+    Clause cl;
     intptr_t pcoffset;
     size_t lneeded, lroom;
     int i;
+    word blob = *ep++;
 
-    if ( !(cl = clause_clref(*ep++)) ||
-	 arityFunctor(f->definition) != cl->variables + 2 )
+    if ( !(cref = clause_clref(blob)) ||
+	 arityFunctor(f->definition) != cref->value.clause->variables + 2 )
     { PL_type_error("continuation", continuation);
       return NULL;
     }
 
+    cl = cref->value.clause;
     pcoffset = valInt(*ep++);
 
-    lneeded = SIZEOF_CREF_CLAUSE +
+    lneeded = sizeof(word) +
 	      (size_t)argFrameP((LocalFrame)NULL, cl->variables);
     lroom   = roomStack(local);
     if ( unlikely(lroom < lneeded) )	/* resize the stack */
@@ -432,8 +477,8 @@ retry:
       goto retry;
     }
 
-    cref = (ClauseRef)top;
-    fr   = addPointer(cref, SIZEOF_CREF_CLAUSE);
+    *(Word)top = blob;			/* see (*) */
+    fr   = addPointer(top, sizeof(word));
     top  = addPointer(top, lneeded);
 
     ap = argFrameP(fr, 0);
@@ -459,12 +504,9 @@ retry:
       }
     }
 
+    assert(ap == (Word)top);
     lTop = top;
 
-    cref->next         = NULL;
-    cref->d.key        = 0;
-    cref->value.clause = cl;
-
     fr->programPointer = pcret;
     fr->parent         = pfr;
     fr->clause         = cref;
