@@ -89,8 +89,7 @@ typedef struct
 		 *******************************/
 
 struct PL_global_data
-{ uintptr_t	heap_base;		/* heap-base rounded downwards */
-  pl_defaults_t	defaults;		/* system default settings */
+{ pl_defaults_t	defaults;		/* system default settings */
   pl_options_t	options;		/* command-line options */
   State		stateList;		/* list of loaded states */
   int		initialised;		/* Heap is initialised */
@@ -453,6 +452,9 @@ struct PL_local_data
   { int		pending[2];		/* PL_raise() pending signals */
     int		current;		/* currently processing signal */
     int		is_sync;		/* current signal is synchronous */
+#ifndef __unix__
+    int		forced;			/* Forced signal */
+#endif
   } signal;
 
   struct
@@ -680,8 +682,8 @@ struct PL_local_data
 
 #ifdef O_LIMIT_DEPTH
   struct
-  { uintptr_t limit;
-    uintptr_t reached;
+  { size_t limit;
+    size_t reached;
   } depth_info;
 #endif
 
@@ -790,8 +792,6 @@ GLOBAL PL_local_data_t *PL_current_engine_ptr;
 #define exception_printed	(LD->exception.printed)
 #define gc_status		(LD->gc.status)
 #define debugstatus		(LD->_debugstatus)
-#define depth_limit		(LD->depth_info.limit)
-#define depth_reached		(LD->depth_info.reached)
 #define base_addresses		(LD->bases)
 #define Suser_input		(LD->IO.streams[0])
 #define Suser_output		(LD->IO.streams[1])
