@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2005-2016, University of Amsterdam
+    Copyright (c)  2005-2020, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -34,7 +35,11 @@
 */
 
 /*#define O_DEBUG 1*/
+#include <math.h>
+#include <fenv.h>
+#include <float.h>
 #include "pl-incl.h"
+#include "pl-arith.h"
 #include "pl-inline.h"
 #undef LD
 #define LD LOCAL_LD
@@ -53,7 +58,7 @@ static mpz_t MPZ_MAX_LONG;
 
 
 		 /*******************************
-		 *	MEMMORY MANAGEMENT	*
+		 *	 MEMORY MANAGEMENT	*
 		 *******************************/
 
 #if O_MY_GMP_ALLOC
@@ -236,10 +241,13 @@ globalMPZ(Word at, mpz_t mpz, int flags ARG_LD)
 { DEBUG(CHK_SECURE, assert(!onStackArea(global, at) && !onStackArea(local, at)));
 
   if ( mpz->_mp_alloc )
-  { Word p;
-    size_t size;
-    size_t wsz = mpz_wsize(mpz, &size);
-    word m     = mkIndHdr(wsz+1, TAG_INTEGER);
+  { size_t size, wsz;
+    Word p;
+    word m;
+
+  copy:
+    wsz = mpz_wsize(mpz, &size);
+    m   = mkIndHdr(wsz+1, TAG_INTEGER);
 
     if ( wsizeofInd(m) != wsz+1 )
     { PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_integer);
@@ -260,10 +268,12 @@ globalMPZ(Word at, mpz_t mpz, int flags ARG_LD)
     *p++     = m;
     p[wsz]   = 0L;			/* pad out */
     p[wsz+1] = m;
-    *p++     = (word)mpz->_mp_size;
+    *p++     = mpz_size_stack(mpz->_mp_size);
     memcpy(p, mpz->_mp_d, size);
   } else				/* already on the stack */
   { Word p = (Word)mpz->_mp_d - 2;
+    if ( !onStack(global, p) )
+      goto copy;
 #ifndef NDEBUG
     size_t size;
     size_t wsz = mpz_wsize(mpz, &size);
@@ -276,6 +286,67 @@ globalMPZ(Word at, mpz_t mpz, int flags ARG_LD)
 }
 
 
+static int
+globalMPQ(Word at, mpq_t mpq, int flags ARG_LD)
+{ mpz_t num, den;			/* num/den */
+
+  num[0] = *mpq_numref(mpq);
+  den[0] = *mpq_denref(mpq);
+
+  if ( num->_mp_alloc || den->_mp_alloc )
+  { size_t num_size, den_size, num_wsz, den_wsz;
+    Word p;
+    word m;
+
+  copy:
+    num_wsz = mpz_wsize(num, &num_size);
+    den_wsz = mpz_wsize(den, &den_size);
+    m       = mkIndHdr(num_wsz+den_wsz+2, TAG_INTEGER);
+
+    if ( wsizeofInd(m) != num_wsz+den_wsz+2 )
+    { PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_rational);
+      return 0;
+    }
+
+    if ( !hasGlobalSpace(num_wsz+den_wsz+4) )
+    { int rc = ensureGlobalSpace(num_wsz+den_wsz+4, flags);
+
+      if ( rc != TRUE )
+	return rc;
+    }
+    p = gTop;
+    gTop += num_wsz+den_wsz+4;
+
+    *at = consPtr(p, TAG_INTEGER|STG_GLOBAL);
+    *p++ = m;
+    *p++ = mpq_size_stack(num->_mp_size);
+    *p++ = mpq_size_stack(den->_mp_size);
+    p[num_wsz-1] = 0L;				/* pad out */
+    memcpy(p, num->_mp_d, num_size);
+    p += num_wsz;
+    p[den_wsz-1] = 0L;				/* pad out */
+    memcpy(p, den->_mp_d, den_size);
+    p += den_wsz;
+    *p = m;
+  } else					/* already on the stack */
+  { Word p = (Word)num->_mp_d - 3;
+    if ( !onStack(global, p) )
+      goto copy;
+    DEBUG(CHK_SECURE,
+	  { size_t num_size;
+	    size_t den_size;
+	    size_t num_wsz = mpz_wsize(num, &num_size);
+	    size_t den_wsz = mpz_wsize(den, &den_size);
+	    assert(p[0] == mkIndHdr(num_wsz+den_wsz+2, TAG_INTEGER));
+	    assert((Word)den->_mp_d == (Word)num->_mp_d + num_wsz);
+	  });
+    *at = consPtr(p, TAG_INTEGER|STG_GLOBAL);
+  }
+
+  return TRUE;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 get_integer() fetches the value of a Prolog  term known to be an integer
 into a number structure. If the  value  is   a  MPZ  number,  it must be
@@ -302,7 +373,7 @@ get_integer(word w, Number n)
     } else
     { n->type = V_MPZ;
 
-      n->value.mpz->_mp_size  = (int)*p++;
+      n->value.mpz->_mp_size  = mpz_stack_size(*p++);
       n->value.mpz->_mp_alloc = 0;
       n->value.mpz->_mp_d     = (mp_limb_t*) p;
     }
@@ -310,6 +381,80 @@ get_integer(word w, Number n)
 }
 
 
+void
+get_rational(word w, Number n)
+{ if ( storage(w) == STG_INLINE )
+  { n->type = V_INTEGER,
+    n->value.i = valInt(w);
+  } else
+  { GET_LD
+    Word p = addressIndirect(w);
+    size_t wsize = wsizeofInd(*p);
+
+    p++;
+    if ( wsize == WORDS_PER_INT64 )
+    { n->type = V_INTEGER;
+      memcpy(&n->value.i, p, sizeof(int64_t));
+    } else if ( (*p&MP_RAT_MASK) )
+    { mpz_t num, den;
+      size_t num_size;
+
+      n->type = V_MPQ;
+      num->_mp_size  = mpz_stack_size(*p++);
+      num->_mp_alloc = 0;
+      num->_mp_d     = (mp_limb_t*) (p+1);
+      num_size = mpz_wsize(num, NULL);
+      den->_mp_size  = mpz_stack_size(*p++);
+      den->_mp_alloc = 0;
+      den->_mp_d     = (mp_limb_t*) (p+num_size);
+
+      *mpq_numref(n->value.mpq) = num[0];
+      *mpq_denref(n->value.mpq) = den[0];
+    } else
+    { n->type = V_MPZ;
+
+      n->value.mpz->_mp_size  = mpz_stack_size(*p++);
+      n->value.mpz->_mp_alloc = 0;
+      n->value.mpz->_mp_d     = (mp_limb_t*) p;
+    }
+  }
+}
+
+
+Code
+get_mpz_from_code(Code pc, mpz_t mpz)
+{ size_t wsize = wsizeofInd(*pc);
+
+  pc++;
+  mpz->_mp_size  = mpz_stack_size(*pc);
+  mpz->_mp_alloc = 0;
+  mpz->_mp_d     = (mp_limb_t*)(pc+1);
+
+  return pc+wsize;
+}
+
+Code
+get_mpq_from_code(Code pc, mpq_t mpq)
+{ Word p = pc;
+  size_t wsize = wsizeofInd(*p);
+  p++;
+  int num_size = mpz_stack_size(*p++);
+  int den_size = mpz_stack_size(*p++);
+  size_t limpsize;
+
+  mpq_numref(mpq)->_mp_size  = num_size;
+  mpq_denref(mpq)->_mp_size  = den_size;
+  mpq_numref(mpq)->_mp_alloc = 0;
+  mpq_denref(mpq)->_mp_alloc = 0;
+  mpq_numref(mpq)->_mp_d     = (mp_limb_t*)p;
+  limpsize = sizeof(mp_limb_t) * abs(num_size);
+  p += (limpsize+sizeof(word)-1)/sizeof(word);
+  mpq_denref(mpq)->_mp_d     = (mp_limb_t*)p;
+
+  return pc+wsize+1;
+}
+
+
 		 /*******************************
 		 *	  IMPORT/EXPORT		*
 		 *******************************/
@@ -319,32 +464,64 @@ addMPZToBuffer(Buffer b, mpz_t mpz)
 	Add mpz in a machine independent representation to the given buffer.
 	The data is stored in limps of 1 byte, preceeded by the byte-count
 	as 4-byte big-endian number;
+
+addMPQToBuffer(Buffer b, mpq_t mpq)
+	Similar to mpz, but first writes the two headers and then the
+	two bit patterns.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
-addMPZToBuffer(Buffer b, mpz_t mpz)
-{ size_t size = (mpz_sizeinbase(mpz, 2)+7)/8;
-  ssize_t hdrsize;
-  size_t count;
+static void
+add_mpz_size_buffer(Buffer b, mpz_t mpz, size_t size)
+{ ssize_t hdrsize;
 
-  if ( !growBuffer(b, size+4) )
-    outOfCore();
   if ( mpz_sgn(mpz) < 0 )
     hdrsize = -(ssize_t)size;
   else
     hdrsize = (ssize_t)size;
-  DEBUG(1, Sdprintf("addMPZToBuffer(): Added %d bytes\n", size));
 
   *b->top++ = (char)((hdrsize>>24)&0xff);
   *b->top++ = (char)((hdrsize>>16)&0xff);
   *b->top++ = (char)((hdrsize>> 8)&0xff);
   *b->top++ = (char)((hdrsize    )&0xff);
+}
+
+static void
+add_mpz_bits_buffer(Buffer b, mpz_t mpz, size_t size)
+{ size_t count;
 
   mpz_export(b->top, &count, 1, 1, 1, 0, mpz);
   assert(count == size);
   b->top = b->top + size;
 }
 
+void
+addMPZToBuffer(Buffer b, mpz_t mpz)
+{ size_t size = (mpz_sizeinbase(mpz, 2)+7)/8;
+
+  if ( !growBuffer(b, size+4) )
+    outOfCore();
+  add_mpz_size_buffer(b, mpz, size);
+  add_mpz_bits_buffer(b, mpz, size);
+}
+
+void
+addMPQToBuffer(Buffer b, mpq_t mpq)
+{ mpz_t num, den;			/* num/den */
+  size_t num_size, den_size;
+
+  num[0] = *mpq_numref(mpq);
+  den[0] = *mpq_denref(mpq);
+  num_size = (mpz_sizeinbase(num, 2)+7)/8;
+  den_size = (mpz_sizeinbase(den, 2)+7)/8;
+
+  if ( !growBuffer(b, num_size+den_size+8) )
+    outOfCore();
+
+  add_mpz_size_buffer(b, num, num_size);
+  add_mpz_size_buffer(b, den, den_size);
+  add_mpz_bits_buffer(b, num, num_size);
+  add_mpz_bits_buffer(b, den, den_size);
+}
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 loadMPZFromCharp() loads an MPZ number directly back to the global stack
@@ -354,6 +531,55 @@ avoided by creating a dummy mpz that looks big enough to mpz_import().
 
 #define SHIFTSIGN32 ((sizeof(int)-4)*8)
 
+static char *
+load_mpz_size(const char *data, int *szp)
+{ int size = 0;
+
+  size |= (data[0]&0xff)<<24;
+  size |= (data[1]&0xff)<<16;
+  size |= (data[2]&0xff)<<8;
+  size |= (data[3]&0xff);
+  size = (size << SHIFTSIGN32)>>SHIFTSIGN32;	/* sign extend */
+
+  *szp = size;
+  data += 4;
+
+  return (char *)data;
+}
+
+static char *
+load_abs_mpz_size(const char *data, int *szp, int *neg)
+{ int size;
+
+  data = load_mpz_size(data, &size);
+  if ( neg )
+  { if ( size < 0 )
+    { size = -size;
+      *neg = TRUE;
+    } else
+    { *neg = FALSE;
+    }
+  } else if ( size < 0 )
+    size = -size;
+  *szp = size;
+
+  return (char *)data;
+}
+
+static char *
+load_mpz_bits(const char *data, size_t size, size_t limpsize, int neg, Word p)
+{ mpz_t mpz;
+
+  mpz->_mp_size  = limpsize;
+  mpz->_mp_alloc = limpsize;
+  mpz->_mp_d     = (mp_limb_t*)p;
+
+  mpz_import(mpz, size, 1, 1, 1, 0, data);
+  assert((Word)mpz->_mp_d == p);	/* check no (re-)allocation is done */
+
+  return (char*)data+size;
+}
+
 char *
 loadMPZFromCharp(const char *data, Word r, Word *store)
 { GET_LD
@@ -361,24 +587,10 @@ loadMPZFromCharp(const char *data, Word r, Word *store)
   size_t limpsize;
   size_t wsize;
   int neg;
-  mpz_t mpz;
   Word p;
   word m;
 
-  size |= (data[0]&0xff)<<24;
-  size |= (data[1]&0xff)<<16;
-  size |= (data[2]&0xff)<<8;
-  size |= (data[3]&0xff);
-  size = (size << SHIFTSIGN32)>>SHIFTSIGN32;	/* sign extend */
-  data += 4;
-
-  DEBUG(1, Sdprintf("loadMPZFromCharp(): size = %d bytes\n", size));
-
-  if ( size < 0 )
-  { neg = TRUE;
-    size = -size;
-  } else
-    neg = FALSE;
+  data = load_abs_mpz_size(data, &size, &neg);
 
   limpsize = (size+sizeof(mp_limb_t)-1)/sizeof(mp_limb_t);
   wsize = (limpsize*sizeof(mp_limb_t)+sizeof(word)-1)/sizeof(word);
@@ -389,35 +601,73 @@ loadMPZFromCharp(const char *data, Word r, Word *store)
   *p++ = m;
   p[wsize] = 0L;			/* pad out */
   p[wsize+1] = m;
-  *p++ = (neg ? -limpsize : limpsize);
-  mpz->_mp_size  = limpsize;
-  mpz->_mp_alloc = limpsize;
-  mpz->_mp_d     = (mp_limb_t*)p;
+  *p++ = mpz_size_stack(neg ? -limpsize : limpsize);
 
-  mpz_import(mpz, size, 1, 1, 1, 0, data);
-  assert((Word)mpz->_mp_d == p);		/* check no (re-)allocation is done */
-
-  return (char *)data+size;
+  return load_mpz_bits(data, size, limpsize, neg, p);
 }
 
+char *
+loadMPQFromCharp(const char *data, Word r, Word *store)
+{ GET_LD
+  int num_size;
+  int den_size;
+  size_t num_limpsize, num_wsize;
+  size_t den_limpsize, den_wsize;
+  int num_neg, den_neg;
+  size_t wsize;
+  Word p;
+  word m;
+
+  data = load_abs_mpz_size(data, &num_size, &num_neg);
+  data = load_abs_mpz_size(data, &den_size, &den_neg);
+
+  num_limpsize = (num_size+sizeof(mp_limb_t)-1)/sizeof(mp_limb_t);
+  num_wsize = (num_limpsize*sizeof(mp_limb_t)+sizeof(word)-1)/sizeof(word);
+  den_limpsize = (den_size+sizeof(mp_limb_t)-1)/sizeof(mp_limb_t);
+  den_wsize = (den_limpsize*sizeof(mp_limb_t)+sizeof(word)-1)/sizeof(word);
+  wsize = num_wsize+den_wsize;
+
+  p = *store;
+  *store += (wsize+4);
+  *r = consPtr(p, TAG_INTEGER|STG_GLOBAL);
+  m = mkIndHdr(wsize+2, TAG_INTEGER);
+  *p++ = m;
+  *p++ = mpq_size_stack(num_neg ? -num_limpsize : num_limpsize);
+  *p++ = mpq_size_stack(den_neg ? -den_limpsize : den_limpsize);
+  p[num_wsize-1] = 0;
+  data = load_mpz_bits(data, num_size, num_limpsize, num_neg, p);
+  p += num_wsize;
+  p[den_wsize-1] = 0;
+  data = load_mpz_bits(data, den_size, den_limpsize, den_neg, p);
+  p += den_wsize;
+  *p++ = m;
+  assert(p == *store);
+
+  return (char *)data;
+}
 
 char *
 skipMPZOnCharp(const char *data)
-{ int size = 0;
+{ int size;
 
-  size |= (data[0]&0xff)<<24;
-  size |= (data[1]&0xff)<<16;
-  size |= (data[2]&0xff)<<8;
-  size |= (data[3]&0xff);
-  size = (size << SHIFTSIGN32)>>SHIFTSIGN32;	/* sign extend */
-  data += 4;
-
-  if ( size < 0 )
-    size = -size;
+  data = load_abs_mpz_size(data, &size, NULL);
 
   return (char *)data + size;
 }
 
+char *
+skipMPQOnCharp(const char *data)
+{ int num_size;
+  int den_size;
+
+  data = load_abs_mpz_size(data, &num_size, NULL);
+  data = load_abs_mpz_size(data, &den_size, NULL);
+  data += num_size;
+  data += den_size;
+
+  return (char *)data;
+}
+
 #undef SHIFTSIGN32
 
 
@@ -524,6 +774,13 @@ promoteToMPQNumber(number *n)
     case V_FLOAT:
     { double v = n->value.f;
 
+      switch(fpclassify(v))
+      { case FP_NAN:
+	  return PL_error(NULL, 0, NULL, ERR_AR_UNDEF);
+        case FP_INFINITE:
+	  return PL_error(NULL, 0, NULL, ERR_AR_RAT_OVERFLOW);
+      }
+
       n->type = V_MPQ;
       mpq_init(n->value.mpq);
       mpq_set_d(n->value.mpq, v);
@@ -817,36 +1074,9 @@ put_number(Word at, Number n, int flags ARG_LD)
       return put_mpz(at, n->value.mpz, flags PASS_LD);
     case V_MPQ:
     { if ( mpz_cmp_ui(mpq_denref(n->value.mpq), 1L) == 0 )
-      { return put_mpz(at, mpq_numref(n->value.mpq), flags PASS_LD);
-      } else
-      { word num, den;
-	Word p;
-	size_t req = ( mpz_wsize(mpq_numref(n->value.mpq), NULL)+3 +
-		       mpz_wsize(mpq_denref(n->value.mpq), NULL)+3 + 3 );
-
-	if ( !hasGlobalSpace(req) )
-	{ int rc = ensureGlobalSpace(req, flags);
-
-	  if ( rc != TRUE )
-	    return rc;
-	}
-
-	if ( !(put_mpz(&num, mpq_numref(n->value.mpq), 0 PASS_LD)) ||
-	     !(put_mpz(&den, mpq_denref(n->value.mpq), 0 PASS_LD)) )
-	  fail;
-
-	p = gTop;
-	gTop += 3;
-	assert(gTop <= gMax);
-
-
-	p[0] = FUNCTOR_rdiv2;
-	p[1] = num;
-	p[2] = den;
-
-	*at = consPtr(p, TAG_COMPOUND|STG_GLOBAL);
-	return TRUE;
-      }
+	return put_mpz(at, mpq_numref(n->value.mpq), flags PASS_LD);
+      else
+	return globalMPQ(at, n->value.mpq, flags PASS_LD);
     }
 #endif
     case V_FLOAT:
@@ -899,22 +1129,25 @@ PL_unify_number__LD(term_t t, Number n ARG_LD)
       break;
 #ifdef O_GMP
     case V_MPQ:
-    { word w;
-      term_t tmp;
-      int rc;
+    { if ( isRational(*p) )
+      { number n2;
+	int rc;
 
-      if ( !(tmp=PL_new_term_ref()) )
-	return FALSE;
-      if ( (rc=put_number(&w, n, ALLOW_GC PASS_LD)) != TRUE )
-	return raiseStackOverflow(rc);
+	get_rational(*p, &n2);
+	rc = (cmpNumbers(n, &n2) == CMP_EQUAL);
+	clearNumber(&n2);
 
-      *valTermRef(tmp) = w;
-      return PL_unify(t, tmp);
+	return rc;
+      }
+      break;
     }
 #endif
     case V_FLOAT:
       if ( isFloat(*p) )
-	return n->value.f == valFloat(*p);
+      {	Word ap = valIndirectP(*p);
+
+	return memcmp((char*)&n->value.f, ap, sizeof(n->value.f)) == 0;
+      }
       break;
   }
 
@@ -938,8 +1171,8 @@ PL_put_number__LD(term_t t, Number n ARG_LD)
 
 void
 get_number(word w, Number n ARG_LD)
-{ if ( isInteger(w) )
-  { get_integer(w, n);
+{ if ( isRational(w) )
+  { get_rational(w, n);
   } else
   { n->type = V_FLOAT;
     n->value.f = valFloat(w);
@@ -952,8 +1185,8 @@ PL_get_number__LD(term_t t, Number n ARG_LD)
 { Word p = valTermRef(t);
 
   deRef(p);
-  if ( isInteger(*p) )
-  { get_integer(*p, n);
+  if ( isRational(*p) )
+  { get_rational(*p, n);
     succeed;
   }
   if ( isFloat(*p) )
@@ -979,6 +1212,23 @@ PL_get_number(term_t t, Number n)
 		 *	     PROMOTION		*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+The GMP functions which convert  mpz's   and  mpq's to double's truncate
+(round to zero) if necessary ignoring the current IEEE rounding mode. To
+correct this incorrect  rounding  when  the   mode  is  `to_postive`  or
+`to_negative` all calls  to  mpX_get_d()  should   be  wrapped  in  this
+function. Note that this function has no GMP dependencies.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+double
+mpX_round(double d) {
+  switch (fegetround()) {
+    case FE_UPWARD  : return (d>=0) ? nexttoward(d, INFINITY) : d;
+    case FE_DOWNWARD: return (d<=0) ? nexttoward(d,-INFINITY) : d;
+    default: return d;
+  }
+}
+
 int
 promoteToFloatNumber(Number n)
 { switch(n->type)
@@ -988,10 +1238,7 @@ promoteToFloatNumber(Number n)
       break;
 #ifdef O_GMP
     case V_MPZ:
-    { double val = mpz_get_d(n->value.mpz);
-
-      if ( !check_float(val) )
-	return FALSE;
+    { double val = mpX_round(mpz_get_d(n->value.mpz));
 
       clearNumber(n);
       n->value.f = val;
@@ -999,10 +1246,7 @@ promoteToFloatNumber(Number n)
       break;
     }
     case V_MPQ:
-    { double val = mpq_get_d(n->value.mpq);
-
-      if ( !check_float(val) )
-	return FALSE;
+    { double val = mpq_to_double(n->value.mpq);
 
       clearNumber(n);
       n->value.f = val;
@@ -1014,7 +1258,7 @@ promoteToFloatNumber(Number n)
       break;
   }
 
-  return TRUE;
+  return check_float(n);
 }
 
 
@@ -1063,8 +1307,8 @@ cmpNumbers() compares two numbers. First, both   numbers are promoted to
 the lowest (in V_* ordering) common type, after which they are compared.
 Note that if the common type is V_FLOAT, but not both are V_FLOAT we can
 run into troubles because big  integers  may   be  out  of range for the
-double representation. We trust mpz_get_d()   and mpq_get_d() return +/-
-float infinity and this compares  correctly   with  the  other argument,
+double representation. We trust mpz_get_d()   and mpq_to_double() return
++/- float infinity and this compares  correctly with the other argument,
 which is guaranteed to be a valid float.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
@@ -1073,16 +1317,19 @@ cmpFloatNumbers(Number n1, Number n2)
 { if ( n1->type == V_FLOAT )
   { double d2;
 
+    if ( isnan(n1->value.f) )
+      return CMP_NOTEQ;
+
     switch(n2->type)
     { case V_INTEGER:
 	d2 = (double)n2->value.i;
 	break;
 #ifdef O_GMP
       case V_MPZ:
-	d2 = mpz_get_d(n2->value.mpz);
+	d2 = mpX_round(mpz_get_d(n2->value.mpz));
 	break;
       case V_MPQ:
-	d2 = mpq_get_d(n2->value.mpq);
+	d2 = mpq_to_double(n2->value.mpq);
 	break;
 #endif
       default:
@@ -1097,16 +1344,19 @@ cmpFloatNumbers(Number n1, Number n2)
 
     assert(n2->type == V_FLOAT);
 
+    if ( isnan(n2->value.f) )
+      return CMP_NOTEQ;
+
     switch(n1->type)
     { case V_INTEGER:
 	d1 = (double)n1->value.i;
 	break;
 #ifdef O_GMP
       case V_MPZ:
-	d1 = mpz_get_d(n1->value.mpz);
+	d1 = mpX_round(mpz_get_d(n1->value.mpz));
 	break;
       case V_MPQ:
-	d1 = mpq_get_d(n1->value.mpq);
+	d1 = mpq_to_double(n1->value.mpq);
 	break;
 #endif
       default:
@@ -1157,7 +1407,6 @@ cmpNumbers(Number n1, Number n2)
   return CMP_EQUAL;
 }
 
-
 void
 cpNumber(Number to, Number from)
 { to->type = from->type;
@@ -1181,13 +1430,211 @@ cpNumber(Number to, Number from)
   }
 }
 
+#ifdef O_GMP
+
+		 /*******************************
+		 *	 FLOAT <-> RATIONAL	*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This code is copied from ECLiPSe  7.0_53.   This  code is covered by the
+CMPL 1.1 (Cisco-style Mozilla Public License  Version 1.1), available at
+www.eclipse-clp.org/license.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+/*
+ * Divide two bignums giving a double float result. The naive solution
+ *	return mpz_to_double(num) / mpz_to_double(den);
+ * suffers from floating point overflows when the numbers are huge
+ * and is inefficient because it looks at unnecessarily many digits.
+ *
+ * IEEE double precision is 53 bits mantissa and 12 bits signed exponent.
+ * So the largest integer representable with doubles is 1024 bits wide,
+ * of which the first 53 are ones, i.e. it lies between 2^1023 and 2^1024.
+ * If the dividend's MSB is more than 1024 bits higher than the divisor's,
+ * the result will always be floating point infinity (no need to divide).
+ * If we do divide, we first drop excess integer precision by keeping only
+ * DBL_PRECISION_LIMBS and ignoring the lower limbs for both operands
+ * (i.e. we effectively scale the integers down, or right-shift them).
+ */
+
+#define MIN_LIMB_DIFF (2+1024/GMP_NUMB_BITS)
+#define DBL_PRECISION_LIMBS (2+53/GMP_NUMB_BITS)
+#define MAX_ULONG_DBL ((double)~0L+1.0)
+
+static double
+mpz_fdiv(mpz_t num, mpz_t den)
+{ mp_ptr longer_d, shorter_d;
+  mp_size_t shorter_size, longer_size, ignored_limbs = 0;
+  int negative, swapped;
+  /* By declaring res volatile we make sure that the result is rounded
+   * to double precision instead of being returned with extended precision
+   * in a floating point register, which can have confusing consequences */
+  volatile double res;
+
+  shorter_size = num->_mp_size;
+  longer_size = den->_mp_size;
+  negative = 0;
+
+  if ( shorter_size < 0 )
+  { shorter_size = -shorter_size;
+    negative = !negative;
+  }
+  if ( longer_size < 0 )
+  { longer_size = -longer_size;
+    negative = !negative;
+  }
+  if ( shorter_size > longer_size )
+  { longer_size = shorter_size;
+    longer_d = num->_mp_d;
+    shorter_size = (den->_mp_size >= 0 ? den->_mp_size : -den->_mp_size);
+    shorter_d = den->_mp_d;
+    swapped = 1;			/* abs(res) > 1 */
+  } else
+  { longer_d = den->_mp_d;
+    shorter_d = num->_mp_d;
+    swapped = 0;			/* abs(res) < 1 */
+  }
+
+  if ( longer_size - shorter_size > MIN_LIMB_DIFF )
+  { res = swapped ? HUGE_VAL : 0.0;
+  } else
+  { double l,s;
+    long int le, se;
+    mpz_t li, si;
+    int r_mode;
+
+    /* we ignore limbs that are not significant for the result */
+    if ( longer_size > MIN_LIMB_DIFF )	/* more can't be represented */
+    { ignored_limbs = longer_size - MIN_LIMB_DIFF;
+      longer_size -= ignored_limbs;
+      shorter_size -= ignored_limbs;
+    }
+    if ( shorter_size > DBL_PRECISION_LIMBS )	/* more exceeds the precision */
+    { ignored_limbs += shorter_size - DBL_PRECISION_LIMBS;
+      longer_size -= shorter_size - DBL_PRECISION_LIMBS;
+      shorter_size = DBL_PRECISION_LIMBS;
+    }
+    longer_d += ignored_limbs;
+    shorter_d += ignored_limbs;
+    li->_mp_alloc = li->_mp_size = longer_size; li->_mp_d = longer_d;
+    si->_mp_alloc = si->_mp_size = shorter_size; si->_mp_d = shorter_d;
+
+    l = mpz_get_d_2exp(&le, li);
+    s = mpz_get_d_2exp(&se, si);
+
+    /* if result negative, some rounding modes must be swapped;
+       avoid if unnecessary */
+    if ( negative )
+    { r_mode = fegetround();
+
+      switch (r_mode)
+      { case FE_UPWARD:   fesetround(FE_DOWNWARD); break;
+	case FE_DOWNWARD: fesetround(FE_UPWARD);
+      }
+    }
+
+    if ( swapped )
+      res = (l/s) * pow(2.0, le-se);
+    else
+      res = (s/l) * pow(2.0, se-le);
+
+    if ( negative )
+      fesetround(r_mode);
+  }
+
+  return negative ? -res : res;
+}
+
+double
+mpq_to_double(mpq_t q)
+{ return mpz_fdiv(mpq_numref(q), mpq_denref(q));
+}
+
+
+/*
+ * Try to compute a "nice" rational from a float, using continued fractions.
+ * Stop when the rational converts back into the original float exactly.
+ * If the process doesn't converge (due to numeric problems), or produces
+ * a result longer than the fast bitwise conversion from the float mantissa,
+ * then fall back to the bitwise conversion.
+ */
+
+void
+mpq_set_double(mpq_t q, double f)	/* float -> nice rational */
+{ double fabs = (f < 0.0) ? -f : f;	/* get rid of the sign */
+  double x = fabs;
+  mpq_t b, c;
+  MP_INT *pna = mpq_numref(q);		/* use output q for a directly */
+  MP_INT *pda = mpq_denref(q);
+  MP_INT *pnb = mpq_numref(b);
+  MP_INT *pdb = mpq_denref(b);
+  MP_INT *pnc = mpq_numref(c);
+  MP_INT *pdc = mpq_denref(c);
+  mpz_t big_xi;
+  int half_exp;                       /* predict bitwise conversion size */
+  double fr = frexp(fabs, &half_exp);
+  int bitwise_denominator_size = DBL_MANT_DIG-(half_exp-1);
+
+  (void)fr;
+  if ( bitwise_denominator_size < 1 )
+    bitwise_denominator_size = 1;
+
+  mpz_set_ui(pna, 1L);                /* a = q = 1/0 */
+  mpz_set_ui(pda, 0L);
+  mpq_init(b);			      /* b = 0/1 */
+  mpq_init(c);			      /* auxiliary */
+  mpz_init(big_xi);                   /* auxiliary */
+
+  while ((mpz_fdiv(pna, pda)) != fabs)
+  { /* infinite x indicates failure to converge */
+    if ( !isfinite(x) )
+      goto _bitwise_conversion_;
+
+    double xi = floor(x);
+    double xf = x - xi;
+
+      /* compute a = a*xi + b for both numerator and denominator */
+    mpq_swap(q, b);
+    if ( x < (double)LONG_MAX )
+    { unsigned long int_xi = (unsigned long) xi;
+      mpz_mul_ui(pnc, pnb, int_xi);
+      mpz_mul_ui(pdc, pdb, int_xi);
+    } else
+    { mpz_set_d(big_xi, xi);
+      mpz_mul(pnc, pnb, big_xi);
+      mpz_mul(pdc, pdb, big_xi);
+    }
+    mpz_add(pna, pna, pnc);
+    mpz_add(pda, pda, pdc);
+
+    /* if it gets too long, fall back to bitwise conversion */
+    if (mpz_sizeinbase(pda, 2) > bitwise_denominator_size)
+      goto _bitwise_conversion_;
+
+    x = 1.0/xf;
+  }
+
+  if ( f < 0.0 )
+    mpq_neg(q, q);
+  mpq_canonicalize(q);                /* normally not be necessary */
+
+  goto _cleanup_;
+
+_bitwise_conversion_:
+  mpq_set_d(q, f);                    /* bitwise conversion */
+
+_cleanup_:
+  mpz_clear(big_xi);
+  mpq_clear(c);
+  mpq_clear(b);
+}
+
 
 		 /*******************************
 		 *	 PUBLIC INTERFACE	*
 		 *******************************/
 
-#ifdef O_GMP
-
 int
 PL_get_mpz(term_t t, mpz_t mpz)
 { GET_LD
