@@ -93,6 +93,8 @@ MSB64(int64_t i)
 }
 #endif
 
+#define MEMORY_ACQUIRE() MemoryBarrier()
+#define MEMORY_RELEASE() MemoryBarrier()
 #define MEMORY_BARRIER() MemoryBarrier()
 
 static inline size_t
@@ -120,6 +122,8 @@ __builtin_popcount(size_t sz)
 #endif
 
 #ifdef HAVE_GCC_ATOMIC
+#define MEMORY_ACQUIRE()	__atomic_thread_fence(__ATOMIC_ACQUIRE)
+#define MEMORY_RELEASE()	__atomic_thread_fence(__ATOMIC_RELEASE)
 #define MEMORY_BARRIER()	__atomic_thread_fence(__ATOMIC_SEQ_CST)
 #endif
 
@@ -229,6 +233,8 @@ MSB64(int64_t i)
 
 #ifndef MEMORY_BARRIER
 #define MEMORY_BARRIER() (void)0
+#define MEMORY_ACQUIRE() (void)0
+#define MEMORY_RELEASE() (void)0
 #endif
 
 		 /*******************************
@@ -603,26 +609,34 @@ normally 0). longToPointer() does the inverse operation.
 
 static inline uintptr_t
 pointerToInt(void *ptr)
-{ uintptr_t p   = (uintptr_t) ptr;
+{
+#if SIZEOF_VOIDP == 8
+  return (uintptr_t) ptr;
+#else
+  uintptr_t p   = (uintptr_t) ptr;
   uintptr_t low = p & 0x3L;
 
-  p -= GD->heap_base;
   p >>= 2;
   p |= low<<(sizeof(uintptr_t)*8-2);
 
   return p;
+#endif
 }
 
 
 static inline void *
 intToPointer(uintptr_t p)
-{ uintptr_t low = p >> (sizeof(uintptr_t)*8-2);
+{
+#if SIZEOF_VOIDP == 8
+  return (void*) p;
+#else
+  uintptr_t low = p >> (sizeof(uintptr_t)*8-2);
 
   p <<= 2;
   p |= low;
-  p += GD->heap_base;
 
   return (void *) p;
+#endif
 }
 
 #endif /*PL_INLINE_H_INCLUDED*/
