@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2015, University of Amsterdam
+    Copyright (c)  1985-2020, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -34,6 +34,7 @@
 */
 
 #include "pl-incl.h"
+#include "pl-arith.h"
 #include "pl-dict.h"
 
 #undef LD
@@ -450,12 +451,12 @@ prolog_list_to_sort_list(term_t t,		/* input list */
       return SORT_ERR;
     }
     l = valTermRef(t);			/* may be shifted */
-    deRef(l);
   }
 
   p = (list)gTop;
   *lp = p;
 
+  deRef(l);
   while(len-- > 0)
   { p->item.term = HeadList(l);
     deRef(p->item.term);
