@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2017, University of Amsterdam
+    Copyright (c)  1985-2020, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -34,6 +35,7 @@
 */
 
 #include "pl-incl.h"
+#include "pl-comp.h"
 #undef LD
 #define LD LOCAL_LD
 
@@ -84,7 +86,7 @@ _lookupModule(atom_t name ARG_LD)
 #endif
   set(m, M_CHARESCAPE);
   if ( !GD->options.traditional )
-    set(m, DBLQ_STRING|BQ_CODES);
+    set(m, DBLQ_STRING|BQ_CODES|O_RATIONAL_SYNTAX);
 
   if ( name == ATOM_user || name == ATOM_system )
     m->procedures = newHTable(PROCEDUREHASHSIZE);
@@ -155,6 +157,8 @@ temporary modules.
 
     - current_op/3 to facilitate using the Pengine operators for
       rendering results.
+    - current_predicate/1, which no longer enumerates through
+      temporary modules.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Module
@@ -186,6 +190,61 @@ releaseModule(Module m)
 }
 
 
+ModuleEnum
+newModuleEnum(int flags)
+{ ModuleEnum en = malloc(sizeof(*en));
+
+  if ( en )
+  { if ( (en->tenum = newTableEnum(GD->tables.modules)) )
+    { en->current = NULL;
+      en->flags = flags;
+    } else
+    { free(en);
+      en = NULL;
+    }
+  }
+
+  return en;
+}
+
+Module
+advanceModuleEnum(ModuleEnum en)
+{ void *v;
+  Module m = NULL;
+
+  PL_LOCK(L_MODULE);
+  for(;;)
+  { if ( advanceTableEnum(en->tenum, NULL, &v) )
+      m = v;
+    else
+      m = NULL;
+
+    if ( m && m->class == ATOM_temporary )
+    { if ( (en->flags&MENUM_TEMP) )
+      { m->references++;
+	if ( en->current )
+	  releaseModule(en->current);
+	en->current = m;
+      } else
+	continue;
+    }
+
+    break;
+  }
+  PL_UNLOCK(L_MODULE);
+
+  return m;
+}
+
+void
+freeModuleEnum(ModuleEnum en)
+{ freeTableEnum(en->tenum);
+  if ( en->current )
+    releaseModule(en->current);
+  free(en);
+}
+
+
 static void
 unallocModuleSymbol(void *name, void *value)
 { unallocModule(value);
@@ -242,7 +301,9 @@ static void
 unallocModule(Module m)
 { GET_LD
 
+#ifdef O_PLMT
   if ( LD )
+#endif
   { if ( LD->modules.source == m ) LD->modules.source = MODULE_user;
     if ( LD->modules.typein == m ) LD->modules.typein = MODULE_user;
   }
@@ -287,6 +348,7 @@ static void
 unlinkSourceFilesModule(Module m)
 { size_t i, high = highSourceFileIndex();
   struct bit_vector *vec = new_bitvector(high+1);
+  SourceFile sf;
 
   for_table(m->procedures, name, value,
 	    markSourceFilesProcedure(value, vec));
@@ -301,6 +363,11 @@ unlinkSourceFilesModule(Module m)
   }
 
   free_bitvector(vec);
+
+  if ( (sf = m->file) )
+  { m->file = NULL;
+    releaseSourceFile(sf);
+  }
 }
 
 
@@ -516,7 +583,7 @@ clearSupersModule(Module m)
 int
 setSuperModule(Module m, Module s)
 { if ( s == m )
-    cannotSetSuperModule(m, s);
+    return cannotSetSuperModule(m, s);
 
   if ( m->supers && !m->supers->next )
   { if ( (Module)m->supers->value != s )
@@ -562,11 +629,17 @@ set_module(Module m, term_t prop ARG_LD)
       { m->class = class;
 	return TRUE;
       } else if ( class == ATOM_temporary )
-      { if ( m->procedures && m->procedures->size != 0 )
-	  return PL_error(NULL, 0,
-			  "module is not empty",
+      { Table procs;
+
+	if ( m->class == ATOM_user &&
+	     !((procs=m->procedures) && procs->size != 0) )
+	{ m->class = class;
+	} else
+	{ return PL_error(NULL, 0,
+			  m->class != ATOM_user ? "Not a user module" :
+						  "module is not empty",
 			  ERR_PERMISSION, ATOM_module_property, ATOM_class, arg);
-	m->class = class;
+	}
 	return TRUE;
       } else
 	return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_module_class, arg);
@@ -854,7 +927,7 @@ is associated to multiple modules.
 static
 PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
 { PRED_LD
-  TableEnum e;
+  ModuleEnum e;
   Module m;
   atom_t name;
   SourceFile sf = NULL;
@@ -880,10 +953,10 @@ PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
 	return FALSE;			/* given, but non-existing file */
 
       if ( sf )
-      { if ( sf->modules )
-	{ int rc = FALSE;
+      { int rc = FALSE;
 
-	  PL_LOCK(L_PREDICATE);
+	if ( sf->modules )
+	{ PL_LOCK(L_PREDICATE);
 	  if ( sf->modules->next )
 	  { term_t tail = PL_copy_term_ref(module);
 	    term_t head = PL_new_term_ref();
@@ -904,20 +977,21 @@ PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
 
 	out:
 	  PL_UNLOCK(L_PREDICATE);
-	  return rc;
 	}
-	return FALSE;			/* source-file has no modules */
+	releaseSourceFile(sf);
+
+	return rc;			/* source-file has no modules */
       }
 
-      e = newTableEnum(GD->tables.modules);
+      if ( !(e = newModuleEnum(0)) )
+	return PL_no_memory();
       break;
     case FRG_REDO:
       e = CTX_PTR;
-      get_existing_source_file(file, &sf PASS_LD);
       break;
     case FRG_CUTTED:
       e = CTX_PTR;
-      freeTableEnum(e);
+      freeModuleEnum(e);
       succeed;
     default:
       assert(0);
@@ -926,7 +1000,7 @@ PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
 
 					/* mode (-,-) */
 
-  while( advanceTableEnum(e, NULL, (void**)&m) )
+  while( (m=advanceModuleEnum(e)) )
   { atom_t f = ( !m->file ? ATOM_nil : m->file->name);
 
     if ( m->class == ATOM_system && m->name != ATOM_system &&
@@ -940,7 +1014,7 @@ PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
     break;				/* must be an error */
   }
 
-  freeTableEnum(e);
+  freeModuleEnum(e);
   return FALSE;
 }
 
@@ -1210,7 +1284,7 @@ declareModule(atom_t name, atom_t class, atom_t super,
     clearHTable(module->public);
   }
   if ( super )
-    setSuperModule(module, _lookupModule(super PASS_LD));
+    rc = setSuperModule(module, _lookupModule(super PASS_LD));
 
   PL_UNLOCK(L_MODULE);
 
@@ -1218,11 +1292,12 @@ declareModule(atom_t name, atom_t class, atom_t super,
   { if ( !PL_unify_nil(rtail) )
       return FALSE;
 
-    rc = printMessage(ATOM_warning,
-		      PL_FUNCTOR_CHARS, "declare_module", 2,
-		        PL_ATOM, name,
-		        PL_FUNCTOR_CHARS, "abolish", 1,
-		          PL_TERM, rdef);
+    if ( rc )
+      rc = printMessage(ATOM_warning,
+			PL_FUNCTOR_CHARS, "declare_module", 2,
+			  PL_ATOM, name,
+			  PL_FUNCTOR_CHARS, "abolish", 1,
+			    PL_TERM, rdef);
   }
 
   return rc;
@@ -1273,7 +1348,6 @@ unify_export_list(term_t public, Module module ARG_LD)
   term_t list = PL_copy_term_ref(public);
   int rval = TRUE;
 
-  LOCKMODULE(module);
   for_table(module->public, name, value,
 	    { if ( !PL_unify_list(list, head, list) ||
 		   !unify_functor(head, (functor_t)name, GP_NAMEARITY) )
@@ -1281,7 +1355,6 @@ unify_export_list(term_t public, Module module ARG_LD)
 		break;
 	      }
 	    })
-  UNLOCKMODULE(module);
   if ( rval )
     return PL_unify_nil(list);
 
@@ -1289,6 +1362,32 @@ unify_export_list(term_t public, Module module ARG_LD)
 }
 
 
+static size_t
+sizeof_module(Module m)
+{ GET_LD
+  size_t size = sizeof(*m);
+
+  if ( m->public)     size += sizeofTable(m->public);
+  if ( m->procedures) size += sizeofTable(m->procedures);
+  if ( m->operators)  size += sizeofTable(m->operators);
+
+  for_table(m->procedures, name, value,
+	    { Procedure proc = value;
+	      Definition def = proc->definition;
+
+	      size += sizeof(*proc);
+
+	      if ( def->module == m && false(def, P_FOREIGN) )
+	      { Definition def = getProcDefinition(proc);
+		size += sizeof_predicate(def);
+	      }
+	    });
+
+  return size;
+}
+
+
+
 static
 PRED_IMPL("$module_property", 2, module_property, 0)
 { PRED_LD
@@ -1321,6 +1420,8 @@ PRED_IMPL("$module_property", 2, module_property, 0)
   { return unify_export_list(a, m PASS_LD);
   } else if ( pname == ATOM_class )
   { return PL_unify_atom(a, m->class);
+  } else if ( pname == ATOM_size )
+  { return PL_unify_int64(a, sizeof_module(m));
   } else if ( pname == ATOM_program_size )
   { return PL_unify_int64(a, m->code_size);
   } else if ( pname == ATOM_last_modified_generation )
@@ -1365,7 +1466,9 @@ export_pi1(term_t pi, Module module ARG_LD)
 
   if ( ReadingSource )
   { SourceFile sf = lookupSourceFile(source_file_name, TRUE);
-    return exportProcedureSource(sf, module, proc);
+    int rc = exportProcedureSource(sf, module, proc);
+    releaseSourceFile(sf);
+    return rc;
   } else
   { return exportProcedure(module, proc);
   }
@@ -1478,7 +1581,10 @@ fixExportModule(Module m, Definition old, Definition new)
 	      if ( proc->definition == old )
 	      { DEBUG(1, Sdprintf("Patched def of %s\n",
 				  procedureName(proc)));
+		shareDefinition(new);
 		proc->definition = new;
+		if ( unshareDefinition(old) == 0 )
+		  lingerDefinition(old);
 	      }
 	    });
 
@@ -1547,7 +1653,7 @@ retry:
       } else
       { lingerDefinition(odef);
       }
-      set(old, pflags);
+      set(old, pflags|PROC_IMPORTED);
 
       succeed;
     }
@@ -1602,8 +1708,8 @@ retry:
 
     nproc->flags = pflags;
     nproc->source_no = 0;
-    nproc->definition = proc->definition;
     shareDefinition(proc->definition);
+    nproc->definition = proc->definition;
 
     LOCKMODULE(destination);
     old = addHTable(destination->procedures,
