@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2017, University of Amsterdam
+    Copyright (c)  1985-2020, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,9 +36,12 @@
 
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
+#include "pl-comp.h"
+#include "pl-arith.h"
 #include "pl-dbref.h"
 #include "pl-termwalk.c"
 #include "pl-dict.h"
+#include "pl-event.h"
 
 #define WORDS_PER_PLINT (sizeof(int64_t)/sizeof(word))
 
@@ -246,8 +250,9 @@ typedef struct
 #define PL_REC_ALLOCVAR		(16)	/* Allocate a variable on global */
 #define PL_REC_CYCLE		(17)	/* cyclic reference */
 #define PL_REC_MPZ		(18)	/* GMP integer */
+#define PL_REC_MPQ		(19)	/* GMP rational */
 
-#define PL_TYPE_EXT_COMPOUND_V2	(19)	/* Read V2 external records */
+#define PL_TYPE_EXT_COMPOUND_V2	(20)	/* Read V2 external records */
 
 static const int v2_map[] =
 { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,		/* variable..string */
@@ -534,7 +539,7 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 
 	  info->size += wsizeofIndirect(w) + 2;
 
-	  get_integer(w, &n);
+	  get_rational(w, &n);
 	  switch(n.type)
 	  { case V_INTEGER:
 	      addOpCode(info, PL_TYPE_INTEGER);
@@ -545,6 +550,10 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 	      addOpCode(info, PL_REC_MPZ);
 	      addMPZToBuffer((Buffer)&info->code, n.value.mpz);
 	      break;
+	    case V_MPQ:
+	      addOpCode(info, PL_REC_MPQ);
+	      addMPQToBuffer((Buffer)&info->code, n.value.mpq);
+	      break;
 #endif
 	    default:
 	      assert(0);
@@ -593,7 +602,8 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 
 	  mark.term = f;
 	  mark.fdef = f->definition;
-	  pushSegStack(&LD->cycle.lstack, mark, cycle_mark);
+	  if ( !pushSegStack(&LD->cycle.lstack, mark, cycle_mark) )
+	    return FALSE;
 	  f->definition = (functor_t)consUInt(info->size);
 				  /* overflow test (should not be possible) */
 	  DEBUG(CHK_SECURE, assert(valUInt(f->definition) == (uintptr_t)info->size));
@@ -607,7 +617,8 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 		   Sdprintf("Added %s/%d\n",
 			    stringAtom(valueFunctor(functor)->name),
 			    arityFunctor(functor)));
-	pushWorkAgenda(agenda, arity, f->arguments);
+	if ( !pushWorkAgenda(agenda, arity, f->arguments) )
+	  return FALSE;
 	continue;
       }
       default:
@@ -678,6 +689,7 @@ compileTermToHeap__LD(term_t t,
   size_t size;
   size_t rsize = SIZERECORD(flags);
   term_agenda agenda;
+  int rc;
 
   DEBUG(CHK_SECURE, checkData(valTermRef(t)));
 
@@ -690,30 +702,34 @@ compileTermToHeap__LD(term_t t,
   info.lock = !(info.external || (flags&R_NOLOCK));
 
   initTermAgenda(&agenda, 1, valTermRef(t));
-  compile_term_to_heap(&agenda, &info PASS_LD);
+  rc = compile_term_to_heap(&agenda, &info PASS_LD);
   clearTermAgenda(&agenda);
   restoreVars(&info);
   unvisit(PASS_LD1);
 
-  size = rsize + sizeOfBuffer(&info.code);
-  if ( allocate )
-    record = (*allocate)(closure, size);
-  else
-    record = PL_malloc_atomic_unmanaged(size);
+  if ( rc )
+  { size = rsize + sizeOfBuffer(&info.code);
+    if ( allocate )
+      record = (*allocate)(closure, size);
+    else
+      record = PL_malloc_atomic_unmanaged(size);
 
-  if ( record )
-  {
+    if ( record )
+    {
 #ifdef REC_MAGIC
-    record->magic = REC_MAGIC;
+      record->magic = REC_MAGIC;
 #endif
-    record->gsize = (unsigned int)info.size; /* only 28-bit */
-    record->nvars = info.nvars;
-    record->size  = (int)size;
-    record->flags = flags;
-    if ( flags & R_DUPLICATE )
-    { record->references = 1;
+      record->gsize = (unsigned int)info.size; /* only 28-bit */
+      record->nvars = info.nvars;
+      record->size  = (int)size;
+      record->flags = flags;
+      if ( flags & R_DUPLICATE )
+      { record->references = 1;
+      }
+      memcpy(addPointer(record, rsize), info.code.base, sizeOfBuffer(&info.code));
     }
-    memcpy(addPointer(record, rsize), info.code.base, sizeOfBuffer(&info.code));
+  } else
+  { record = NULL;
   }
   discardBuffer(&info.code);
 
@@ -742,7 +758,6 @@ variantRecords(const Record r1, const Record r2)
 #define	REC_GROUND  0x10		/* Record is ground */
 #define	REC_VMASK   0xe0		/* Version mask */
 #define REC_VSHIFT     5		/* shift for version mask */
-#define	REC_VERSION 0x03		/* Version id */
 
 #define REC_SZMASK  (REC_32|REC_64)	/* SIZE_MASK */
 
@@ -752,7 +767,7 @@ variantRecords(const Record r1, const Record r2)
 #define REC_SZ REC_32
 #endif
 
-#define REC_HDR		(REC_SZ|(REC_VERSION<<REC_VSHIFT))
+#define REC_HDR		(REC_SZ|(PL_REC_VERSION<<REC_VSHIFT))
 #define REC_COMPAT(m)	(((m)&(REC_VMASK|REC_SZMASK)) == REC_HDR)
 
 typedef struct record_data
@@ -952,7 +967,7 @@ PRED_IMPL("fast_term_serialized", 2, fast_term_serialized, 0)
     { return FALSE;
     }
   } else if ( PL_get_nchars(string, &len, &rec,
-			    CVT_STRING|BUF_RING|REP_ISO_LATIN_1|CVT_EXCEPTION) )
+			    CVT_STRING|BUF_STACK|REP_ISO_LATIN_1|CVT_EXCEPTION) )
   { term_t tmp;
 
     return ( (tmp = PL_new_term_ref()) &&
@@ -1407,9 +1422,11 @@ copy_record(Word p, CopyInfo b ARG_LD)
       }
 #ifdef O_GMP
       case PL_REC_MPZ:
-      { b->data = loadMPZFromCharp(b->data, p, &b->gstore);
+	b->data = loadMPZFromCharp(b->data, p, &b->gstore);
+	continue;
+      case PL_REC_MPQ:
+	b->data = loadMPQFromCharp(b->data, p, &b->gstore);
 	continue;
-      }
 #endif
       case PL_TYPE_FLOAT:
       case PL_TYPE_EXT_FLOAT:
@@ -1686,6 +1703,9 @@ scanAtomsRecord(CopyInfo b, void (*func)(atom_t a))
       case PL_REC_MPZ:
 	b->data = skipMPZOnCharp(b->data);
 	continue;
+      case PL_REC_MPQ:
+	b->data = skipMPQOnCharp(b->data);
+	continue;
 #endif
       case PL_TYPE_FLOAT:
       case PL_TYPE_EXT_FLOAT:
@@ -1831,7 +1851,7 @@ PL_recorded_external(const char *rec, term_t t)
       if ( !b.version_map )
       { Sdprintf("PL_recorded_external(): "
 		 "Incompatible version (%d, current %d)\n",
-		 save_version, REC_VERSION);
+		 save_version, PL_REC_VERSION);
 	fail;
       }
     }
@@ -1917,12 +1937,12 @@ real atom.
 
 bool
 unifyKey(term_t key, word val)
-{ if ( isAtom(val) || isTaggedInt(val) )
+{ GET_LD
+
+  if ( isAtom(val) || isTaggedInt(val) )
   { return _PL_unify_atomic(key, val);
   } else
-  { GET_LD
-
-    return PL_unify_functor(key, (functor_t) val);
+  { return PL_unify_functor(key, (functor_t) val);
   }
 }
 
@@ -2302,7 +2322,8 @@ PRED_IMPL("instance", 2, instance, 0)
     return FALSE;
 
   if ( type == DB_REF_CLAUSE )
-  { Clause clause = ptr;
+  { ClauseRef cref = ptr;
+    Clause clause = cref->value.clause;
     gen_t generation = generationFrame(environment_frame);
 
     if ( true(clause, GOAL_CLAUSE) ||
@@ -2345,7 +2366,8 @@ PRED_IMPL("erase", 1, erase, 0)
     return FALSE;
 
   if ( type == DB_REF_CLAUSE )
-  { Clause clause = ptr;
+  { ClauseRef cref = ptr;
+    Clause clause = cref->value.clause;
     Definition def = clause->predicate;
 
     if ( !true(def, P_DYNAMIC) )
