@@ -316,6 +316,19 @@ raiseSignal(PL_local_data_t *ld, int sig)
 }
 
 
+int
+pendingSignal(PL_local_data_t *ld, int sig)
+{ if ( sig > 0 && sig <= MAXSIGNAL && ld )
+  { int off  = (sig-1)/32;
+    int mask = 1 << ((sig-1)%32);
+
+    return (ld->signal.pending[off] & mask) ? TRUE : FALSE;
+  }
+
+  return -1;
+}
+
+
 		 /*******************************
 		 *	   STACK-LAYOUT		*
 		 *******************************/
@@ -1389,7 +1402,7 @@ static Definition
 localDefinition(Definition def ARG_LD)
 { unsigned int tid = LD->thread.info->pl_tid;
   size_t idx = MSB(tid);
-  LocalDefinitions v = def->impl.local;
+  LocalDefinitions v = def->impl.local.local;
 
   if ( !v->blocks[idx] )
   { size_t bs = (size_t)1<<idx;
@@ -1412,7 +1425,7 @@ localDefinition(Definition def ARG_LD)
 void
 destroyLocalDefinition(Definition def, unsigned int tid)
 { size_t idx = MSB(tid);
-  LocalDefinitions v = def->impl.local;
+  LocalDefinitions v = def->impl.local.local;
   Definition local;
 
   local = v->blocks[idx][tid];
@@ -1438,7 +1451,7 @@ getProcDefinition__LD(Definition def ARG_LD)
 Definition
 getProcDefinitionForThread(Definition def, unsigned int tid)
 { size_t idx = MSB(tid);
-  LocalDefinitions v = def->impl.local;
+  LocalDefinitions v = def->impl.local.local;
 
   if ( !v->blocks[idx] )
     return NULL;
@@ -2539,9 +2552,9 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
     qf->flags_saved = (LD->prolog_flag.mask.flags & NDEBUG_SAVE_FLAGS);
     setPrologFlagMask(PLFLAG_LASTCALL);
 #ifdef O_LIMIT_DEPTH
-    qf->saved_depth_limit   = depth_limit;
-    qf->saved_depth_reached = depth_reached;
-    depth_limit = DEPTH_NO_LIMIT;
+    qf->saved_depth_limit   = LD->depth_info.limit;
+    qf->saved_depth_reached = LD->depth_info.reached;
+    LD->depth_info.limit    = DEPTH_NO_LIMIT;
 #endif
   }
   fr->predicate      = def;
@@ -2618,8 +2631,8 @@ restore_after_query(QueryFrame qf)
     LD->prolog_flag.mask.flags &= (~NDEBUG_SAVE_FLAGS);
     LD->prolog_flag.mask.flags |= qf->flags_saved;
 #ifdef O_LIMIT_DEPTH
-    depth_limit   = qf->saved_depth_limit;
-    depth_reached = qf->saved_depth_reached;
+    LD->depth_info.limit   = qf->saved_depth_limit;
+    LD->depth_info.reached = qf->saved_depth_reached;
 #endif /*O_LIMIT_DEPTH*/
   }
   updateAlerted(LD);
@@ -3297,7 +3310,7 @@ next_choice:
 	  { int action;
 
 	    SAVE_REGISTERS(qid);
-	    clearLocalVariablesFrame(fr);
+	    clearLocalVariablesFrame(FR);
 	    action = tracePort(fr, BFR, REDO_PORT, NULL PASS_LD);
 	    LOAD_REGISTERS(qid);
 	    ch = BFR;			/* can be shifted */
