@@ -447,6 +447,7 @@ static int
 close_zipper(zipper *z)
 { zipFile zf;
   unzFile uzf;
+  char *path;
   int rc = 0;
 
   if ( (zf=z->writer) )
@@ -456,15 +457,19 @@ close_zipper(zipper *z)
   { z->reader = NULL;
     rc = unzClose(uzf);
   }
-  if ( z->path )
-  { free((char*)z->path);
-    z->path = NULL;
+  if ( (path=(char*)z->path) )
+  { z->path = NULL;
+    free(path);
   }
   if ( z->input.any )
   { switch(z->input_type)
     { case ZIP_STREAM:
 	if ( true(z, ZIP_CLOSE_STREAM_ON_CLOSE) )
-	  Sclose(z->input.stream);
+	{ IOSTREAM *in = z->input.stream;
+
+	  z->input.stream = NULL;
+	  Sclose(in);
+	}
         break;
       default:
 	break;
@@ -820,6 +825,7 @@ static const opt_spec zip_new_file_options[] =
   { ATOM_time,		    OPT_DOUBLE },
   { ATOM_method,	    OPT_ATOM   },
   { ATOM_level,		    OPT_INT    },
+  { ATOM_zip64,		    OPT_BOOL   },
   { NULL_ATOM,		    0          }
 };
 
@@ -837,9 +843,10 @@ PRED_IMPL("zipper_open_new_file_in_zip", 4, zipper_open_new_file_in_zip, 0)
   int level = 6;
   atom_t method = ATOM_deflated;
   int imethod;
+  int zip64 = FALSE;
 
   if ( !scan_options(A4, 0, ATOM_zip_options, zip_new_file_options,
-		     &extra, &comment, &ftime, &method, &level) )
+		     &extra, &comment, &ftime, &method, &level, &zip64) )
     return FALSE;
 
   if ( extra )
@@ -886,7 +893,7 @@ PRED_IMPL("zipper_open_new_file_in_zip", 4, zipper_open_new_file_in_zip, 0)
 				 0,		/* crc */
 				 VERSIONMADEBY,	/* versionMadeBy */
 				 0,		/* flagBase */
-				 FALSE);	/* zip64 */
+				 zip64);	/* zip64 */
 
     if ( rc == 0 )
     { IOSTREAM *s = Snew(z, SIO_OUTPUT, &Szipfunctions);
@@ -1001,7 +1008,7 @@ PRED_IMPL("zipper_open_current", 3, zipper_open_current, 0)
   atom_t encoding = NULL_ATOM;
   int	 bom      = -1;
   int    release  = TRUE;
-  int flags       = SIO_INPUT|SIO_RECORDPOS;
+  int flags       = SIO_INPUT|SIO_RECORDPOS|SIO_FBUF;
   IOENC enc;
 
   if ( !scan_options(A3, 0, ATOM_stream_option, zipopen3_options,
